package de.lmu.ifi.pms.cep.dura.compiler.antlr.utils;

import java.io.File;

import org.antlr.runtime.tree.CommonTree;

public class DuraTree extends CommonTree {
	private final String filename;
	
	DuraTree(final CommonTree tree) {
		super(tree);
		
		this.filename = null;
	}
	
	public DuraTree(final Object tree, final String filename) {
		this((CommonTree) tree, null, filename);
	}
	
	public DuraTree(final CommonTree tree, final String filename) {
		this(tree, null, filename);
	}
	
	public DuraTree(final CommonTree tree, final DuraTree parent, final String filename) {
		super(tree);
		
		this.setChildIndex(tree.childIndex);
		this.setParent(parent);
		
		int n = tree.getChildCount();
		for (int i = 0; i < n; i++) {
			CommonTree child = (CommonTree) tree.getChild(i);
			DuraTree newSubTree = new DuraTree(child, this, filename);
			this.addChild(newSubTree);
		}

		this.filename = new File(filename).getAbsolutePath();//FIXME modified to obtain absolute path 
	}
	
	@Override
	public DuraTree getChild(final int i) {
		if (this.children==null || i>=this.children.size()) {
			return null;
		}
		
		return (DuraTree) this.children.get(i);
	}
	
	@Override
	public DuraTree getParent() {
		return (DuraTree) this.parent;
	}
	
	public String getFilename() {
		return this.filename;
	}
	
	public String getPositionDescription() {
		return getPositionDescription(this);
	}
	
	public String getPositionDescription(DuraTree tree) {
		if (tree.comprisesMultipleLines()) {
			return String.format("\tat %s:%d-%d", tree.getFilename(), tree.getLine(), tree.getLastLine());
		} else {
			return String.format("\tat %s:%d", tree.getFilename(), tree.getLine());
		}
	}
	
	public int getLastLine() {
		int n = this.getChildCount();
		
		if (n == 0) {
			return this.getLine();
		} else {
			return this.getChild(n-1).getLastLine();
		}
	}
	
	public boolean comprisesMultipleLines() {
		return this.getLine() < this.getLastLine();
	}
}
