package de.lmu.ifi.pms.cep.dura.compiler.dura.desugaring;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

import org.antlr.runtime.CommonToken;
import org.antlr.runtime.tree.CommonTree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraParser;


public class ExistentialQuantification extends Transformation {

	@SuppressWarnings("unchecked")
	@Override
	public void apply(CommonTree tree) {
		for (CommonTree eventSpecification : collectEventSpecificationNodes(tree)) {
			String temp = null;
			if (EMIT_ANNOTATIONS) {
				temp = eventSpecification.toStringTree();
			}
			
			CommonTree body = (CommonTree) eventSpecification.getChild(1);
			CommonTree composition = (CommonTree) body.getChild(0);
			
			if (composition.getType() == DuraParser.DISJUNCTION) {
				continue;
			}
			
			CommonTree supplement = (CommonTree) composition.getChild(composition.getChildCount() - 1);
				
			List<CommonTree> existsNodes = collectExistsNodes(composition);
			if (!existsNodes.isEmpty()) {
				CommonTree where = new CommonTree(new CommonToken(DuraParser.CONDITION, "CONDITION"));
				if (supplement.getChildCount() > 0) {
					List<CommonTree> children = supplement.getChildren();
					children.add(0, where);
					supplement.freshenParentAndChildIndexes();
				} else {
					supplement.addChild(where);
				}
				
				CommonTree aggregation = new CommonTree(new CommonToken(DuraParser.AGGREGATION, "AGGREGATION"));
				aggregation.addChild(new CommonTree(new CommonToken(DuraParser.GROUPING, "GROUPING")));
				List<CommonTree> children = supplement.getChildren();
				children.add(where.getChildIndex(), aggregation);
				supplement.freshenParentAndChildIndexes();
				
				for (CommonTree existsNode : existsNodes) {
					handleExistsNode(existsNode, where, aggregation);
				}
				
				
				if (EMIT_ANNOTATIONS) {
					CommonTree annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
					annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
					annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Resolved existential quantifiers in deductive rule " + temp + ".\"")));
					eventSpecification.getParent().addChild(annotation);
				}
			}
		}
	}

	private IdentityHashMap<CommonTree, HashSet<CommonTreeSetWrapper>> bindingsMemoizationMap = new IdentityHashMap<CommonTree, HashSet<CommonTreeSetWrapper>>();
	private void  handleExistsNode(CommonTree existsNode, CommonTree where, CommonTree aggregation) {
		CommonTree existsSupplement = (CommonTree) existsNode.getChild(1);
		if (existsSupplement.getChildCount() > 0) {
			for (Object c : existsSupplement.getChildren()) {
				CommonTree child = (CommonTree) c;
				
				if (child.getType() == DuraParser.CONDITION) {
					where.addChildren(child.getChildren());
				}
			}
		}
		
		CommonTree grouping = (CommonTree) aggregation.getChild(0);
		
		CommonTree parent = (CommonTree) existsNode.getParent();
		
		HashSet<CommonTreeSetWrapper> envBindings = bindingsMemoizationMap.get(parent);
		// We only need to handle (add bindings to grouping) the same environment once.
		if (grouping.getChildCount() == 0 || envBindings == null) {
			envBindings = new HashSet<CommonTreeSetWrapper>();
			
			List<CommonTree> env = getEnvironment(existsNode);
			for (CommonTree q : env) {
				// If positive...
				if (q.getType() == DuraParser.EVENT_QUERY || q.getType() == DuraParser.STATE_QUERY) {
					envBindings.add(new CommonTreeSetWrapper(identifier(q)));
					envBindings.addAll(termBindings((CommonTree) q.getChild(1).getChild(0)));
					envBindings.addAll(suppDefs((CommonTree) q.getChild(2)));
				}
			}
			
			for (CommonTreeSetWrapper b : envBindings) {
				grouping.addChild(copyTree(b.getTree()));
			}
			
			bindingsMemoizationMap.put(parent, envBindings);
		}
		
		CommonTree containedQuery = (CommonTree) existsNode.getChild(0);
		
		// Rename all bindings inside the exists that do not have a matchable positive counterpart in the environment.
		// This is similar to standardization apart...
		//
		// If everything is semantically checked this isn't necessary.
		// 
		// If not done the following could occur:
		//
		// and {
		//   event a: t{ var A }
		//   exists{ event b: t{ var X } },
		//   exists{ event c: t{ var X } },
		// } where { ... } ...
		//
		// The exists are removed (and grouping added)
		//
		// and {
		//   event a: t{ var A }
		//   event b: t{ var X },
		//   event c: t{ var X },
		// } where { ... } ...
		// 
		// Now suddenly var X and var X would match. As mentioned above semantic checking should complain
		// before that happens.
		// 
		// Renaming produces something like this:
		// 
		// and {
		//   event a: t{ var A }
		//   event b: t{ var _0_1_ignore },
		//   event c: t{ var _0_2_ignore },
		// } where { ... } ...
		// 
		// This will also lead to non-matchability in the conjunction supplement etc.
		// For testing only.
		renameNonOutsideMatchableBindings(containedQuery, envBindings);
		
		int index = existsNode.getChildIndex();
		existsNode.getParent().replaceChildren(index, index, containedQuery);
	}
	
	private void renameNonOutsideMatchableBindings(CommonTree query, HashSet<CommonTreeSetWrapper> envBindings) {	
		Queue<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(query);
		
		HashMap<String, String> oldNameToNewNameMap = new HashMap<String, String>();
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			if (node.getType() == DuraParser.VARIABLE ||
					(node.getType() == DuraParser.IDENTIFIER && node != identifier(query))
					) {
				if (!envBindings.contains(new CommonTreeSetWrapper(node))) {
					CommonTree nameNode = (CommonTree) node.getChild(0);
					String oldName = nameNode.getText();
					String newName = oldNameToNewNameMap.get(oldName);
					
					if (newName == null) {
						newName = createName("ignore"); 
						// newName = createName();
						
						oldNameToNewNameMap.put(oldName, newName);
					}
					
					nameNode.getToken().setText(newName);
				}
			} else if (node.getType() == DuraParser.TYPED_VARIABLE) {
				CommonTree unification = (CommonTree) node.getParent();
				CommonTree variable = getVariableFromUnification(unification);
				
				if (!envBindings.contains(new CommonTreeSetWrapper(variable))) {
					CommonTree nameNode = (CommonTree) node.getChild(1);
					String oldName = nameNode.getText();
					String newName = oldNameToNewNameMap.get(oldName);
					
					if (newName == null) {
						newName = createName("ignore"); 
						// newName = createName();
						
						oldNameToNewNameMap.put(oldName, newName);
					}
					
					nameNode.getToken().setText(newName);
				}
			}
			
			if (node.getChildCount() > 0) {
				for (Object c : node.getChildren()) {
					CommonTree child = (CommonTree) c;
					nodeQueue.add(child);
				}
			}
		}
	}
	
	private List<CommonTree> collectExistsNodes(CommonTree composition) {
		ArrayList<CommonTree> nodes = new ArrayList<CommonTree>();
		
		for (Object c : composition.getChildren()) {
			CommonTree child = (CommonTree) c;
			
			if (child.getType() == DuraParser.EXISTS) {
				nodes.add(child);
			}
		}
		
		return nodes;
	}
}
