package de.lmu.ifi.pms.cep.dura.compiler.dura.desugaring;

import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CommonToken;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraParser;



public class ReactiveRuleActionExecutionProvenanceInfoAugmentation extends Transformation {
	
	public ReactiveRuleActionExecutionProvenanceInfoAugmentation() {
		super();
		
		ANTLRStringStream stream = new ANTLRStringStream(compositionAugmentationTermString);
		DuraLexer lexer = new DuraLexer(stream);
		CommonTokenStream tokens = new CommonTokenStream(lexer);
		DuraParser parser = new DuraParser(tokens);
			
		try {
			actionCompositionAugmentationTerm = (CommonTree) parser.internalSchemaDefinition().getTree();
			normalizeTerm(actionCompositionAugmentationTerm);
		} catch (RecognitionException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	
	@Override
	public void apply(CommonTree tree) {
		for (CommonTree reactiveRule : collectReactiveRuleNodes(tree)) {			
			CommonTree head = (CommonTree) reactiveRule.getChild(1);
			CommonTree concurrent = (CommonTree) head.getChild(0);
			
			if (EMIT_ANNOTATIONS) {
				CommonTree annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
				reactiveRule.addChild(annotation);
				annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
				annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Addition of composition-provenance information to action executions in reactive rule " + reactiveRule.toStringTree() + ".\"")));
			}
			
			
			for (int i = 0; i < concurrent.getChildCount(); i++) {
				CommonTree child = (CommonTree) concurrent.getChild(i);
				
				CommonTree childTerm = (CommonTree) child.getChild(0);
					
				String childActionName = childTerm.getText();
					
				if (!actionCompositionAugmentedActionSet.contains(childActionName)) {
					addActionCompositionInformation(tree, childActionName);
				
					actionCompositionAugmentedActionSet.add(childActionName);
					
					if (EMIT_ANNOTATIONS) {
						CommonTree actionDefinition = (CommonTree) explicitActionSchemataMap.get(childActionName).getParent().getParent();
						CommonTree annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
						actionDefinition.addChild(annotation);
						annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
						annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Addition of composition-provenance information to schema of this action due to reactive rule " + reactiveRule.toStringTree() + ".\"")));
					}
				}
				
				
				
				CommonTree composition = copyTree(actionCompositionAugmentationTerm);
				childTerm.addChild(composition);
				
				normalizeTerm(childTerm);
				
				
				// child-index
				composition.getChild(0).setChild(0, new CommonTree(new CommonToken(DuraParser.NUMBER, Integer.toString(i - concurrent.getChildCount()))));
				
				
				
				// "Dummy" value, since in fact the no actual composition is performed!
				// The value (in this case 0, but subject to change) has to stand for an action id that will never exist. 
				
				// parent-action-id
				composition.getChild(1).setChild(0, new CommonTree(new CommonToken(DuraParser.NUMBER, Integer.toString(0))));
				
				
			}
		}
	}

}
