package de.lmu.ifi.pms.cep.dura.compiler.durac;

import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.ACTION_EXECUTION;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.CONJUNCTION;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.DETECT;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.DISJUNCTION;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.DO;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.HEAD;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.LOAD;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.MODULE_NAME;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.PROGRAM;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.SCHEMA;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.STATEFUL_OBJECT;
import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer.TERM;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.actionInitiatedSuffix;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.eventIdentifierPrefix;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.eventIntervalName;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.identifierAttribute;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.identifierClockName;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.intervalBeginAttribute;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.intervalEndAttribute;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.labelSeparator;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.payloadLabel;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.provenanceLabel;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.soCreatedSuffix;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.sytemTimeClockName;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.GREATEST;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.LEAST;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import org.antlr.runtime.ANTLRFileStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTreeNodeStream;
import org.antlr.runtime.tree.Tree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCAstValidation;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCParser;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.StatefulObjectQueryTransformation;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.DuraImplicitTree;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.DuraTree;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.TreeIterator;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.ConstDefinitions;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.DuraSchema;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.Preamble;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.StreamDefinitions;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.TypeDefinitions;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.UnambiguousQueryVariables;
import de.lmu.ifi.pms.cep.dura.compiler.durac.formula_transformations.FormulaNormalization;
import de.lmu.ifi.pms.cep.dura.compiler.durac.queries.AtomicEventQuery;
import de.lmu.ifi.pms.cep.dura.compiler.durac.queries.CompositeEventQuery;
import de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraBasicTypesMapping;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraAstValidationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCModuleNotFoundException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCompilationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraParseException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraTransformationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnspecificCompliationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.UnexpectedTokenException;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.TSA_CompositeExpression;
import de.lmu.ifi.pms.cep.tsa.TSA_Expression;
import de.lmu.ifi.pms.cep.tsa.TSA_Program;
import de.lmu.ifi.pms.cep.tsa.TSA_Query;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.TSA_TemporalStream;
import de.lmu.ifi.pms.cep.tsa.TSA_TemporalStreamExpression;
import de.lmu.ifi.pms.cep.tsa.conditions.ClockFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.CompositeTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.IdentifierAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.IdentifierValue;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TimestampAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.TupleFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.TypedAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.Variable;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ImbeddingOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ProjectionOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_RenameOperator;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchemaTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.Type;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.fixed.wrapper.FixedSetWrapper;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


public class DuraCCompiler {
	
	public static TSA_Program compile(final String namespace, final String program_name, final File file) throws IOException {
		return compile(namespace, program_name, file.getAbsolutePath());
	}
	
	public static TSA_Program compile(final String namespace, final String program_name, final String filename) throws IOException {
		Collection<DuraTree> sourceTrees = new ArrayList<DuraTree>(1);

		try {
			DuraCParser parser = new DuraCParser(new CommonTokenStream(new DuraCLexer(new ANTLRFileStream(filename))));
			
			sourceTrees.add(new DuraTree(parser.program().getTree(), filename));
		} catch (RecognitionException e) {
			throw new DuraParseException(e, filename);
		}

		return compileProgramTrees(namespace, program_name, sourceTrees);
	}

	public static TSA_Program compile(final String seed, final Collection<String> sourceFiles) throws IOException {
		if (sourceFiles == null) {
			throw new DuraCModuleNotFoundException(seed);
		}
		
		String[] moduleName = TSA_SyntaxPatterns.programFullName(seed);
		Map<String, ArrayList<DuraTree>> sourceTrees =  parseSourceFiles(sourceFiles);
		Collection<DuraTree> loadedTrees =  determineLoadedFiles(seed, sourceTrees);
		
		return compileProgramTrees(moduleName[0], moduleName[1], loadedTrees);
	}
	
	private static Map<String, ArrayList<DuraTree>> parseSourceFiles(final Collection<String> sourceFiles) throws IOException {
		Map<String, ArrayList<DuraTree>> sourceTrees = new HashMap<String, ArrayList<DuraTree>>();
		
		if (sourceFiles == null) {
			throw new IllegalArgumentException("sourceFiles must contain at least one filename.");
		}
		
		for (String filename : sourceFiles) {
			try {
				DuraCParser parser = new DuraCParser(new CommonTokenStream(new DuraCLexer(new ANTLRFileStream(filename))));
				
				DuraTree file = new DuraTree(parser.program().getTree(), filename);
				Tree module = file.getChild(0).getChild(0).getChild(0);
				String moduleName = module.getChildCount()==0 ? "" : module.getChild(0).getText();
				
				assert(module.getType() == MODULE_NAME);
				
				if (! sourceTrees.containsKey(moduleName)) {
					sourceTrees.put(moduleName, new ArrayList<DuraTree>());
				}
				
				sourceTrees.get(moduleName).add(file);
			} catch (RecognitionException e) {
				throw new DuraParseException(e, filename);
			}
		}
		
		return sourceTrees;
	}
	
	private static Collection<DuraTree> determineLoadedFiles(final String seed, final Map<String, ArrayList<DuraTree>> modules) {
		Set<DuraTree> result = new HashSet<DuraTree>();
		
		if (! modules.containsKey(seed)) {
			throw new DuraCModuleNotFoundException(seed);
		}
		
		determineLoadedFiles(seed, modules, result);

		return result;
	}
	
	private static void determineLoadedFiles(final String seed, final Map<String, ArrayList<DuraTree>> modules, final Set<DuraTree> result) {
		assert(modules.containsKey(seed));
		
		for (DuraTree file : modules.get(seed)) {
			DuraTree preamble = file.getChild(0).getChild(0);
			
			result.add(file);
			
			for (DuraTree load : new TreeIterator(preamble, LOAD)) {
				String loadedModuleName = load.getChild(0).getText();
				
				if (! modules.containsKey(loadedModuleName)) {
					throw new DuraCModuleNotFoundException(loadedModuleName, load);
				}
				
				determineLoadedFiles(loadedModuleName, modules, result);
			}
		}
	}
	
	private static TSA_Program compileProgramTrees(final String namespace, final String program_name, final Collection<DuraTree> programs) {
		List<TSA_Query> rules = new ArrayList<TSA_Query>();
		
		List<TypeDefinitions> types = new ArrayList<TypeDefinitions>();
		List<ConstDefinitions> constants = new ArrayList<ConstDefinitions>();
		List<StreamDefinitions> streams = new ArrayList<StreamDefinitions>();
		Set<String> origins = new HashSet<String>();
		
		for (DuraTree program : programs) {
			DuraTree pckg = program.getChild(0);
			DuraTree prmbl = pckg.getChild(0);
			
			Preamble preamble = new Preamble(prmbl);
			
			types.add(new TypeDefinitions(prmbl, preamble));
			constants.add(new ConstDefinitions(prmbl, preamble.package_name));
		}

		ConstDefinitions constDefs = ConstDefinitions.join(constants); 
		TypeDefinitions typeDefs = TypeDefinitions.join(types);
		
		for (DuraTree program : programs) {
			DuraTree pckg = program.getChild(0);
			DuraTree prmbl = pckg.getChild(0);
			DuraTree definitions = pckg.getChild(1);
			
			Preamble preamble = new Preamble(prmbl);
			
			streams.add(new StreamDefinitions(definitions, preamble, typeDefs));
		}
		
		DuraSchema schema = new DuraSchema(StreamDefinitions.join(streams), typeDefs, constDefs);
		
		for (DuraTree program : programs) {
			rules.addAll(compileProgramTree(schema, program));
			origins.add(program.getPositionDescription());
		}
		
		try {
			return new TSA_Program(namespace, program_name, "", schema.streams.getStreams(), new FixedArray<TSA_Query>(rules), new FixedSetWrapper<String>(origins));
		} catch (TSA_Exception e) {
			StringBuilder sb = new StringBuilder();
			
			for (String pos : e.origin) {
				sb.append(pos);
				sb.append("\n");
			}
			
			throw new DuraCompilationException(e.getMessage(), sb.toString());
		}
	}
	
	private static List<TSA_Query> compileProgramTree(final DuraSchema schema, final DuraTree program) {
		assert(program.getType() == PROGRAM);
		
		DuraTree pckg = program.getChild(0);
		DuraTree prmbl = pckg.getChild(0);
		
//		validateAstStructure(program);

		ArrayList<TSA_Query> rules = new ArrayList<TSA_Query>();
		Preamble preamble = new Preamble(prmbl);
		
		rules.addAll(compileStatefulObjectInitiationRules(schema.streams));

		for (DuraTree rule : new TreeIterator(pckg, DETECT)) {
			rules.addAll(compileDeclarativeRule(rule, preamble, schema, true));
		}
		
		for (DuraTree rule : new TreeIterator(pckg, DO)) {
			rules.addAll(compileReactiveRule(rule, preamble, schema));
		}
		
		for (DuraTree definition : new TreeIterator(program, STATEFUL_OBJECT)) {
			DuraTree statefulObjectSchema = definition.getChild(1);
			
			assert(statefulObjectSchema.getType() == SCHEMA);
			
			try {
				String streamName = String.format("%s%s%s", statefulObjectSchema.getChild(0).getText(), labelSeparator, soCreatedSuffix);
				
				if (! schema.streams.getStreamByName(TemporalStreamSchemaTypeID.EVENT, preamble, streamName).isStatic) {
					DuraTree created = new DuraImplicitTree(new StatefulObjectQueryTransformation(new CommonTreeNodeStream(statefulObjectSchema)).createdRule().getTree(), statefulObjectSchema);
					DuraTree createSucceeded = new DuraImplicitTree(new StatefulObjectQueryTransformation(new CommonTreeNodeStream(statefulObjectSchema)).createSucceededRule().getTree(), statefulObjectSchema);
					DuraTree terminated = new DuraImplicitTree(new StatefulObjectQueryTransformation(new CommonTreeNodeStream(statefulObjectSchema)).terminatedRule().getTree(), statefulObjectSchema);
					DuraTree terminateSucceeded = new DuraImplicitTree(new StatefulObjectQueryTransformation(new CommonTreeNodeStream(statefulObjectSchema)).terminateSucceededRule().getTree(), statefulObjectSchema);
					DuraTree terminateFailed = new DuraImplicitTree(new StatefulObjectQueryTransformation(new CommonTreeNodeStream(statefulObjectSchema)).terminateFailedRule().getTree(), statefulObjectSchema);
					DuraTree updateTerminated = new DuraImplicitTree(new StatefulObjectQueryTransformation(new CommonTreeNodeStream(statefulObjectSchema)).updateTerminatedRule().getTree(), statefulObjectSchema);
					DuraTree updateCreated = new DuraImplicitTree(new StatefulObjectQueryTransformation(new CommonTreeNodeStream(statefulObjectSchema)).updateCreatedRule().getTree(), statefulObjectSchema);
					DuraTree updated = new DuraImplicitTree(new StatefulObjectQueryTransformation(new CommonTreeNodeStream(statefulObjectSchema)).updatedRule().getTree(), statefulObjectSchema);
					DuraTree updateSucceeded = new DuraImplicitTree(new StatefulObjectQueryTransformation(new CommonTreeNodeStream(statefulObjectSchema)).updateSucceededRule().getTree(), statefulObjectSchema);
					DuraTree updateFailed = new DuraImplicitTree(new StatefulObjectQueryTransformation(new CommonTreeNodeStream(statefulObjectSchema)).updateFailedRule().getTree(), statefulObjectSchema);
	
					rules.addAll(compileDeclarativeRule(created, preamble, schema));
					rules.addAll(compileDeclarativeRule(createSucceeded, preamble, schema));
					rules.addAll(compileDeclarativeRule(terminated, preamble, schema));
					rules.addAll(compileDeclarativeRule(terminateSucceeded, preamble, schema));
					rules.addAll(compileDeclarativeRule(terminateFailed, preamble, schema));
					rules.addAll(compileDeclarativeRule(updateTerminated, preamble, schema));
					rules.addAll(compileDeclarativeRule(updateCreated, preamble, schema));
					rules.addAll(compileDeclarativeRule(updated, preamble, schema));
					rules.addAll(compileDeclarativeRule(updateSucceeded, preamble, schema));
					rules.addAll(compileDeclarativeRule(updateFailed, preamble, schema));
				}
			} catch (RecognitionException e) {
				throw new DuraTransformationException(e, definition);
			} catch (DuraUnspecificCompliationException e) {
				throw new DuraCompilationException(e, definition);
			}
		}
		
		return rules;
	}
	
	private static List<TSA_Query> compileDeclarativeRule(final DuraTree rule, final Preamble preamble, final DuraSchema schema) {
		return compileDeclarativeRule(rule, preamble, schema, false);
	}

		
	private static List<TSA_Query> compileDeclarativeRule(final DuraTree rule, final Preamble preamble, final DuraSchema schema, final boolean userDefinedRule) {
		assert(rule.getType() == DETECT);
		
		DuraTree head = rule.getChild(0);
		DuraTree body = rule.getChild(1);
		DuraTree queryType = body.getChild(0);
		
		String outputName = head.getChild(0).getText();
		List<TSA_Query> result = new ArrayList<TSA_Query>();
		
		if (userDefinedRule && outputName.contains(labelSeparator)) {
			throw new DuraCompilationException(String.format("Events of implicitly defined stream %s may not be derived explicitly.", outputName), head);
		}

		try {
			TSA_TemporalStream outputStream = schema.streams.getStreamByName(TemporalStreamSchemaTypeID.EVENT, preamble, TSA_SyntaxPatterns.bufferFullName(preamble.package_name, outputName));
			
			if (queryType.getType() == CONJUNCTION) {
				CompositeEventQuery duraQuery = DuraCQueryCompiler.compileConjunctiveQuery(body.getChild(0), preamble, schema);
				
				//FIXME: add .dura.query. prefix to all attributes of basic event queries
				FormulaNormalization bodyRefiner = new FormulaNormalization(duraQuery.getVariables(), outputStream.schema.extend(duraQuery.getExpression().schema().attributes), duraQuery.getIdentifierMapping(), schema.constants, preamble);
				
				TSA_Expression headImbeding = new TSA_CompositeExpression(compileHeadConstruction(head, bodyRefiner), duraQuery.getExpression());
				TSA_Expression implicitAttributes = new TSA_CompositeExpression(compileImplicitEventAttributes(rule, duraQuery.getVariables(), bodyRefiner), headImbeding);
				TSA_Expression projection = new TSA_CompositeExpression(new TSA_ProjectionOperator(rule.getPositionDescription(), outputStream.schema.basicAttributes.keys()), implicitAttributes);
				
				result.add(new TSA_Query(preamble.package_name, outputStream.full_name, outputStream.schema, DTSA_BufferType.WORKING, projection, "", rule.getPositionDescription()));
			} else if (queryType.getType() == DISJUNCTION) {
				int childCount = queryType.getChildCount();
				Tree supplement = queryType.getChild(childCount-1);
				
				assert(supplement.getChildCount() == 0);
				
				for (int i=0; i<childCount-1; i++) {
					AtomicEventQuery atomicQuery = new AtomicEventQuery(queryType.getChild(i), preamble, schema);

					FormulaNormalization bodyRefiner = new FormulaNormalization(atomicQuery.getVariables(), outputStream.schema.extend(atomicQuery.getExpression().schema().attributes), schema.constants, preamble);
					
					TSA_Expression headImbeding = new TSA_CompositeExpression(compileHeadConstruction(head, bodyRefiner), atomicQuery.getExpression());
					TSA_Expression implicitAttributes = new TSA_CompositeExpression(compileImplicitEventAttributes(rule, atomicQuery.getVariables(), bodyRefiner), headImbeding);
					TSA_Expression projection = new TSA_CompositeExpression(new TSA_ProjectionOperator(rule.getPositionDescription(), outputStream.schema.basicAttributes.keys()), implicitAttributes);
					
					result.add(new TSA_Query(preamble.package_name, outputStream.full_name, outputStream.schema, DTSA_BufferType.WORKING, projection, "", rule.getPositionDescription()));
				}
			} else {
				throw new UnexpectedTokenException(queryType);
			}
		} catch (DuraUnspecificCompliationException e) {
			throw new DuraCompilationException(e, head.getChild(0));
		} catch (TSA_Exception e) {
			throw new DuraCompilationException(e, rule);
		}
		
		return result;
	}
	
	private static List<TSA_Query> compileReactiveRule(final DuraTree rule, final Preamble preamble, final DuraSchema schema) {
		assert(rule.getType() == DO);
		
		DuraTree head = rule.getChild(1);
		DuraTree body = rule.getChild(0);
		DuraTree concurrent = head.getChild(0);
		
		CompositeEventQuery duraQuery = DuraCQueryCompiler.compileConjunctiveQuery(body.getChild(0), preamble, schema);
		
		List<TSA_Query> result = new ArrayList<TSA_Query>();
		
		for (int i=0; i<concurrent.getChildCount(); i++) {
			DuraTree action = concurrent.getChild(i);
			String outputName = action.getChild(0).getText();
			String actionInitiationName = String.format("%s%s%s", outputName, labelSeparator, actionInitiatedSuffix);

			try {
				TSA_TemporalStream outputStream = schema.streams.getStreamByName(TemporalStreamSchemaTypeID.EVENT, preamble, actionInitiationName);
				
				FormulaNormalization bodyRefiner = new FormulaNormalization(duraQuery.getVariables(), outputStream.schema, schema.constants, preamble);
				
				TSA_Expression headImbeding = new TSA_CompositeExpression(compileHeadConstruction(action, payloadLabel, bodyRefiner), duraQuery.getExpression());
				TSA_Expression implicitAttributes = new TSA_CompositeExpression(compileImplicitActionAttributes(rule, duraQuery.getVariables(), bodyRefiner), headImbeding);
				TSA_Expression projection = new TSA_CompositeExpression(new TSA_ProjectionOperator(rule.getPositionDescription(), outputStream.schema.basicAttributes.keys()), implicitAttributes);
				
				result.add(new TSA_Query(preamble.package_name, outputStream.full_name, outputStream.schema, DTSA_BufferType.WORKING, projection, "", rule.getPositionDescription()));
			} catch (TSA_Exception e) {
				throw new DuraCompilationException(e, action);
			} catch (DuraUnspecificCompliationException e) {
				throw new DuraCompilationException(String.format("Unknown stream name: %s", outputName), action);
			}
		}
		
		return result;
	}
	
	private static Collection<TSA_Query> compileStatefulObjectInitiationRules(final StreamDefinitions streams) {
		ArrayList<TSA_Query> queries = new ArrayList<TSA_Query>();
		
		for (String soName : streams.getStreamNames(TemporalStreamSchemaTypeID.STATE)) {
			try {
				TSA_TemporalStream initStream = streams.getStreamByNameUngarded(TemporalStreamSchemaTypeID.STATE, soName);
				TSA_TemporalStream outputStream = streams.getStreamByNameUngarded(TemporalStreamSchemaTypeID.EVENT, String.format("%s%s%s", soName, labelSeparator, soCreatedSuffix));
				
				TSA_RenameOperator addPayloadPrefix = new TSA_RenameOperator(initStream.origin, new FixedArrayNavigableMap<String, String>(new String[]{ "" }, new String[] { payloadLabel }));
				TSA_Expression renamedAttributes = new TSA_CompositeExpression(addPayloadPrefix, new TSA_TemporalStreamExpression(initStream.full_name, DTSA_BufferType.INPUT, initStream.schema, initStream.origin));
				
				//FIXME: dirty hack, -1 is currently not a valid identifier value 
				IdentifierAttribute provenaneIdAttribute = IdentifierAttribute.create(TSA_SyntaxPatterns.attributeName(provenanceLabel, identifierAttribute));
				IdentifierValue provenanceIdValue = new IdentifierValue(-1);
				TSA_Expression defaultProvenance = new TSA_CompositeExpression(new TSA_ImbeddingOperator(initStream.origin, new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>( new TypedAttribute<?>[] { provenaneIdAttribute }, new Term<?>[] { provenanceIdValue })), renamedAttributes);
				
				TimestampAttribute end = TimestampAttribute.create(TSA_SyntaxPatterns.attributeName(eventIntervalName, intervalEndAttribute));
				TimestampAttribute begin = TimestampAttribute.create(TSA_SyntaxPatterns.prefixAttribute(eventIntervalName, intervalBeginAttribute));
				Term<BasicType> clock = new ClockFunction(sytemTimeClockName, BasicTypeID.TIMESTAMP);
				
				TSA_Expression endTimestamp = new TSA_CompositeExpression(new TSA_ImbeddingOperator(initStream.origin, new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>( new TypedAttribute<?>[] { end }, new Term<?>[] { clock })), defaultProvenance);
				TSA_Expression beginTimestamp = new TSA_CompositeExpression(new TSA_ImbeddingOperator(initStream.origin, new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>( new TypedAttribute<?>[] { begin }, new Term<?>[] { end })), endTimestamp);

				queries.add(new TSA_Query(outputStream.prog_package, outputStream.full_name, outputStream.schema, DTSA_BufferType.WORKING, beginTimestamp, outputStream.description, outputStream.origin));
			} catch (DuraUnspecificCompliationException e) {
				throw new Error(e.getMessage());
			} catch (TSA_Exception e) {
				throw new Error(e.getMessage());
			}
		}
		
		return queries;
	}
	
	private static TSA_ImbeddingOperator compileHeadConstruction(final DuraTree head, final FormulaNormalization bodyRefiner) {
		return compileHeadConstruction(head, new Stack<String>(), bodyRefiner);
	}
	
	private static TSA_ImbeddingOperator compileHeadConstruction(final DuraTree head, final String prefix, final FormulaNormalization bodyRefiner) {
		Stack<String> path = new Stack<String>();
		path.add(prefix);
		
		return compileHeadConstruction(head, path, bodyRefiner);
	}

	
	private static TSA_ImbeddingOperator compileHeadConstruction(final DuraTree head, final Stack<String> path, final FormulaNormalization bodyRefiner) {
		assert(head.getType() == HEAD || head.getType() == ACTION_EXECUTION);
		
		DuraTree type = head.getChild(0);
		
		HashMap<TypedAttribute<?>, Term<? extends Type>> substitutions = new HashMap<TypedAttribute<?>, Term<? extends Type>>();

		for (int i=0; i<type.getChildCount(); i++) {
			compileHeadVariables(type.getChild(i), bodyRefiner, substitutions, path);
		}

		try {
			return new TSA_ImbeddingOperator(head.getPositionDescription(), new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(substitutions));
		} catch (TSA_Exception e) {
			throw new DuraCompilationException(e, head);
		}
	}
	
	private static void compileHeadVariables(final DuraTree term, final FormulaNormalization bodyRefiner, final HashMap<TypedAttribute<?>, Term<? extends Type>> substitutions, final Stack<String> path) {
		switch (term.getType()) {
		case TERM:
			path.push(term.getText());

			for (int i=0; i<term.getChildCount(); i++) {
				DuraTree child = term.getChild(i);
				
				compileHeadVariables(child, bodyRefiner, substitutions, path);
			}

			path.pop();

			break;

		default:
			try {
				Term<? extends Type> value = bodyRefiner.apply(DuraCQueryCompiler.compileTerm(term));
				TypedAttribute<? extends Type> target = (TypedAttribute<? extends Type>) bodyRefiner.apply(Term.attr(DuraCQueryCompiler.joinPath(path)));
				
				FixedArrayNavigableMap<String, BasicType> valueAttributes = value.type.basicAttributes();
				FixedArrayNavigableMap<String, BasicType> targetAttributes = target.type.basicAttributes();
				
				for (String key : valueAttributes.keys()) {
					if (!targetAttributes.has(key) || 
							!DuraBasicTypesMapping.compatible(valueAttributes.get(key), targetAttributes.get(key))) {
						throw new DuraCompilationException(String.format("expected type for attribute %s is %s, found type %s", target.name, target.type, value.type), term);
					}
				}
				
				//drop (hopefully only) attributes with (implicit) default values
				substitutions.put(Term.attr(target.name, value.type), value);
				
				break;
			} catch (DuraUnspecificCompliationException e) {
				throw new DuraCompilationException(e, term);
			}
		}
	}

	/**
	 * Computes an TSA expression that determines the values of all implicit
	 * attributes, such as, the reception-time and the id.
	 * 
	 * @param variables
	 *            the identifiers which need to be considered to determine the
	 *            value for the reception-time
	 * @param refiner
	 *            refiner for the mapping of identifier names to attributes
	 * @throws DuraUnspecificCompliationException 
	 */
	private static TSA_ImbeddingOperator compileImplicitEventAttributes(final DuraTree origin, final UnambiguousQueryVariables variables, final FormulaNormalization refiner) {
		ArrayList<TimestampAttribute> begins = new ArrayList<TimestampAttribute>();
		ArrayList<TimestampAttribute> ends = new ArrayList<TimestampAttribute>();
		
		for (Variable<?> var : variables.getVariables()) {
			if (var.isTyped && var.name.startsWith(eventIdentifierPrefix) && var.type.equals(TypeDefinitions.TIME_INTERVAL)) {
				try {					
					CompositeTerm<? extends Type> interval = (CompositeTerm<? extends Type>) refiner.apply(var);

					begins.add((TimestampAttribute) interval.arg(0));
					ends.add((TimestampAttribute) interval.arg(1));
				} catch (DuraUnspecificCompliationException e) {
					throw new DuraCompilationException(e, origin);
				}
			}
		}
		
		TimestampAttribute begin = TimestampAttribute.create(TSA_SyntaxPatterns.prefixAttribute(eventIntervalName, intervalBeginAttribute));
		TimestampAttribute end = TimestampAttribute.create(TSA_SyntaxPatterns.prefixAttribute(eventIntervalName, intervalEndAttribute));
		IdentifierAttribute ref = IdentifierAttribute.create(identifierAttribute);

		Term<?> minBegin = TupleFunction.apply(LEAST, BasicTypeID.TIMESTAMP.type, begins.toArray(new TimestampAttribute[begins.size()]));
		Term<?> maxEnd = TupleFunction.apply(GREATEST, BasicTypeID.TIMESTAMP.type, ends.toArray(new TimestampAttribute[ends.size()]));
		Term<BasicType> clock = new ClockFunction(identifierClockName, BasicTypeID.IDENTIFIER);
		
		FixedArrayNavigableMap<TypedAttribute<?>, Term<?>> implicitAttributes = new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(
				new TypedAttribute[] {    begin,    end,   ref },
					      new Term[] { minBegin, maxEnd, clock }
		);

		try {
			return new TSA_ImbeddingOperator(origin.getPositionDescription(), implicitAttributes);
		} catch (TSA_Exception e) {
			throw new DuraCompilationException(e, origin);
		}
	}
	
	private static TSA_ImbeddingOperator compileImplicitActionAttributes(final DuraTree origin, final UnambiguousQueryVariables variables, final FormulaNormalization refiner) {
		FixedArrayNavigableMap<TypedAttribute<?>, Term<?>> implicitEventAttributes = (FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>) compileImplicitEventAttributes(origin, variables, refiner).definitions;
		
		FixedArrayNavigableMap<TypedAttribute<?>, Term<?>> implicitActionAttributes = new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(
				new TypedAttribute[] { IdentifierAttribute.create(TSA_SyntaxPatterns.attributeName(payloadLabel, identifierAttribute))},
				new Term[] { new ClockFunction(identifierClockName, BasicTypeID.IDENTIFIER) }
			);
		
		try {
			return new TSA_ImbeddingOperator(origin.getPositionDescription(), FixedArrayNavigableMap.merge(implicitEventAttributes, implicitActionAttributes));
		} catch (TSA_Exception e) {
			throw new DuraCompilationException(e, origin);
		}
	}

	private static void validateAstStructure(final DuraTree tree) {
		CommonTreeNodeStream nodes = new CommonTreeNodeStream(tree);
		DuraCAstValidation walker = new DuraCAstValidation(nodes);
		
        try {
            walker.program();
        } catch (RecognitionException e) {
           throw new DuraAstValidationException(e, tree.getFilename());
        }
	}
}