package de.lmu.ifi.pms.cep.dura.compiler.durac.collections;

import java.util.ArrayList;

import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraAmbiguousMappingException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnknownVariableException;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.Variable;
import de.lmu.ifi.pms.cep.tsa.schema.Type;

public class UnambiguousQueryVariables extends QueryVariables {
	
	
	public UnambiguousQueryVariables() { }
	
	
	public UnambiguousQueryVariables(final Variable<?> variable, final Term<? extends Type> term) {
		this.addMappingUngarded(variable, term);
	}
	
	
	public UnambiguousQueryVariables(QueryVariables variables) throws DuraAmbiguousMappingException {
		for (Variable<?> var : variables.mapping.keySet()) {
			addMapping(var, variables.mapping.get(var));
		}
	}
	
	
	protected void addMapping(Variable<?> variable, Term<? extends Type> attribute) throws DuraAmbiguousMappingException {
		if (! this.mapping.containsKey(variable)) {
			this.mapping.put(variable, new ArrayList<Term<? extends Type>>());
		}
		
		if (this.mapping.get(variable).size() > 0) {
			throw new DuraAmbiguousMappingException(String.format("variable %s is already contained in this mapping", variable.name));
		}
		
		this.mapping.get(variable).add(attribute);
	}
	
	
	protected void addMapping(final Variable<?> variable, final ArrayList<Term<? extends Type>> attributes) throws DuraAmbiguousMappingException {
		if (! this.mapping.containsKey(variable)) {
			this.mapping.put(variable, new ArrayList<Term<? extends Type>>());
		}
		
		this.mapping.get(variable).addAll(attributes);
		
		if (this.mapping.get(variable).size() > 1) {
			throw new DuraAmbiguousMappingException(String.format("variable %s is either already contained in this mapping or ambiguous", variable.name));
		}
	}
	
	
	private void addMappingUngarded(Variable<?> variable, Term<? extends Type> attribute) {
		if (! this.mapping.containsKey(variable)) {
			this.mapping.put(variable, new ArrayList<Term<? extends Type>>());
		}
		
		assert this.mapping.get(variable).size()==0;
		
		this.mapping.get(variable).add(attribute);
	}

	
	public UnambiguousQueryVariables getMappingsMatchingPrefix(String prefix) {
		try {
			return new UnambiguousQueryVariables(super.getMappingsMatchingPrefix(prefix));
		} catch (DuraAmbiguousMappingException e) {
			throw new AssertionError();	//cannot happen
		}
	}
	

	@Override
	public UnambiguousQueryVariables dropMappings(String... variableNames) {
		try {
			return new UnambiguousQueryVariables(super.dropMappings(variableNames));
		} catch (DuraAmbiguousMappingException e) {
			throw new AssertionError();	//cannot happen
		}
	}
	
	
	public Term<? extends Type> getVariableAttribute(Variable<?> variable) throws DuraUnknownVariableException {
		return this.getVariableAttributes(variable).get(0);
	}

}
