package de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCLexer;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.DuraTree;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.UnexpectedTokenException;
import de.lmu.ifi.pms.java.math.Math;


public enum DuraTimeUnitMapping {
	DAY("days?", 24*60*60*1000),
	HOUR("hours?", 60*60*1000),
	MINUTE("min", 60*1000),
	SECOND("sec", 1000),
	MS("ms", 1);

	private final int factorToMs;
	private final String name;
	
	private DuraTimeUnitMapping(final String name, final int factor) {
		this.factorToMs = factor;
		this.name = name;
	}

	public static DuraTimeUnitMapping fromToken(final DuraTree token) {
		assert(token.getType() == DuraCLexer.TIME_UNIT);
		
		String name = token.getText();
		
		for (DuraTimeUnitMapping timeunit : values()) {
			if (name.matches(timeunit.name)) {
				return timeunit;
			}
		}
		
		throw new UnexpectedTokenException(token);
	}
	
	public static long tokenToDuration(final DuraTree token) {
		assert(token.getType() == DuraCLexer.DURATION);
		
		float duration = 0;
		
		for (int i=0; i<token.getChildCount(); i++) {
			DuraTree child = token.getChild(i);
			
			duration += fromToken(child).factorToMs *  Float.parseFloat(child.getChild(0).getText());
		}
		
		return Math.round(duration);
	}
}
