package de.lmu.ifi.pms.cep.event_mill.console;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.BufferedReader;
import java.io.Closeable;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.PrintWriter;
import java.nio.CharBuffer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

import de.lmu.ifi.pms.java.util.string.StringBuilder;


public class Console extends WindowAdapter implements WindowListener, ActionListener, Closeable
{
  private final JFrame frame;
  private final JTextArea textArea;
  protected final Thread out_reader;
  protected final Thread err_reader;
  protected boolean running;

  
  protected final BufferedReader out_pipe; 
  protected final BufferedReader err_pipe;
  
  
  public final PrintWriter out;
  public final PrintWriter err;
  
  final OutputStream out_stream;
  final OutputStream err_stream;

  public Console(final String  name)
  {
    // create all components and add them
    this.frame=new JFrame(name);
    Dimension screenSize=Toolkit.getDefaultToolkit().getScreenSize();
    Dimension frameSize=new Dimension((screenSize.width/2),(screenSize.height/2));
    int x=(frameSize.width/2);
    int y=(frameSize.height/2);
    this.frame.setBounds(x,y,frameSize.width,frameSize.height);

    this.textArea=new JTextArea();
    this.textArea.setEditable(false);
    JButton button=new JButton("clear");

    this.frame.getContentPane().setLayout(new BorderLayout());
    this.frame.getContentPane().add(new JScrollPane(this.textArea),BorderLayout.CENTER);
    this.frame.getContentPane().add(button,BorderLayout.SOUTH);
    this.frame.setVisible(true);		

    this.frame.addWindowListener(this);		
    button.addActionListener(this);
    
    PipedInputStream out_pipe = new PipedInputStream(); 
    PipedInputStream err_pipe = new PipedInputStream();
    this.out_pipe = new BufferedReader(new InputStreamReader(out_pipe));
    this.err_pipe = new BufferedReader(new InputStreamReader(err_pipe));

    try {
    	this.out_stream = new PipedOutputStream(out_pipe);
		this.out=new PrintWriter(this.out_stream, true);
	} catch (IOException e) {
		throw new RuntimeException(e);
		//TODO better handling
	}
    try {
        this.err_stream = new PipedOutputStream(err_pipe);
		this.err=new PrintWriter(this.err_stream, true);
	} catch (IOException e) {
		throw new RuntimeException(e);
		//TODO better handling
	}

    this.running=true; // signals the Threads that they should exit

    // Starting two seperate threads to read from the PipedInputStreams				
    //
    this.out_reader=new Thread(){
    	
    	PrintWriter writer = Console.this.out;
    	BufferedReader reader = Console.this.out_pipe;
    	
    	@Override
		public void run() {
    		boolean running = true;
			CharBuffer b = CharBuffer.allocate(1024);
			StringBuilder sb = new StringBuilder();
			Pattern lineEndPattern = Pattern.compile("\r\n|[\n\r\u2028\u2029\u0085]|$");
			synchronized(Console.this){
				do{
					this.writer.flush();
	    			try {
						this.writer.flush();
	    				while(this.reader.ready()){
						
						try {
							this.reader.read(b);
							b.flip();
							sb.append(b.toString());
							Matcher m = lineEndPattern.matcher(sb);
							int end = 0;
							while(m.find()){
								end = m.end();
							}
							if(end > 0){
						    	Console.this.textArea.append(sb.subSequence(0, end).toString());
						    	Console.this.textArea.setCaretPosition(Console.this.textArea.getDocument().getLength());
						    	sb.subString(end);
							}
						} catch (IOException e) {
							Console.this.close();
							running = false;
						}
						synchronized(Console.this){
							running = Console.this.running;
						}
						this.writer.flush();
						}
						} catch (IOException e) {
							Console.this.close();
							running = false;
						}
						try {
							Console.this.wait(100);
						} catch (InterruptedException e) {
	
						}
	    		}while(running);
			}
    	}
    	
    };
    this.out_reader.setDaemon(true);	
    this.out_reader.start();	
    //
    this.err_reader=new Thread(){
    	
    	PrintWriter writer = Console.this.err;
    	BufferedReader reader = Console.this.err_pipe;
    	
    	@Override
		public void run() {
    		boolean running = true;
			CharBuffer b = CharBuffer.allocate(1024);
			StringBuilder sb = new StringBuilder();
			Pattern lineEndPattern = Pattern.compile("\r\n|[\n\r\u2028\u2029\u0085]|$");
			synchronized(Console.this){
				do{
					this.writer.flush();
	    			try {
						this.writer.flush();
	    				while(this.reader.ready()){
						
						try {
							this.reader.read(b);
							b.flip();
							sb.append(b.toString());
							Matcher m = lineEndPattern.matcher(sb);
							int end = 0;
							while(m.find()){
								end = m.end();
							}
							if(end > 0){
						    	Console.this.textArea.append(sb.subSequence(0, end).toString());
						    	Console.this.textArea.setCaretPosition(Console.this.textArea.getDocument().getLength());
						    	sb.subString(end);
							}
						} catch (IOException e) {
							Console.this.close();
							running = false;
						}
						synchronized(Console.this){
							running = Console.this.running;
						}
						this.writer.flush();
						}
						} catch (IOException e) {
							Console.this.close();
							running = false;
						}
						try {
							Console.this.wait(100);
						} catch (InterruptedException e) {
	
						}
	    		}while(running);
			}
    	}
    	
    };	
    this.err_reader.setDaemon(true);	
    this.err_reader.start();
			
  }

  	@Override
	public synchronized void windowClosed(final WindowEvent evt)
	{
  		this.close();
	}		

	@Override
	public synchronized void windowClosing(final WindowEvent evt)
	{
		this.frame.setVisible(false); // default behaviour of JFrame	
		this.frame.dispose();
	}

	@Override
	public synchronized void actionPerformed(final ActionEvent evt)
	{
		this.textArea.setText("");
	}

  	@Override
	public synchronized void close(){	    
  		if(this.running){
  		    this.running=false;
  		    this.notifyAll(); // stop all threads
  			
		    Thread closeThread = new Thread(){
		    	@Override
				public void run() {
		    		
		    	    try { 
		    		    Console.this.out_reader.join();
		    		    Console.this.out_pipe.close();   
		    		} 
		    		catch (Exception e){
		    			
		    		}		
		    		try { 
		    			Console.this.err_reader.join();
		    			Console.this.err_pipe.close(); 
		    		} 
		    		catch (Exception e){
		    		
		    		}
		    		
		    	}
		    };
		    closeThread.setDaemon(true);
		    closeThread.start();
  		}
	}
  
  
  public static void main(final String[] args) {
	  
  }
  
}
