package de.lmu.ifi.pms.cep.event_mill.execution.program;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;

import de.lmu.ifi.pms.cep.tsa.DTSA_QueryType;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


public abstract class ExecutableQuery<ProgramType extends ExecutableProgram<?, BufferType, ?>, BufferType extends StreamBuffer<BufferType,?,ProgramType>>{
	
	public final DTSA_QueryType type;
	public final String prog_package;
	public final ProgramType program;
	
	//logging
	protected static final Logger logger = LoggerFactory.getLogger(ExecutableQuery.class);
	public final Marker marker;
	
	protected final Logger progLogger;
	
	public ExecutableQuery(final DTSA_QueryType type, final String progPackage, final ProgramType program)
	{
		super();
		this.type = type;
		this.prog_package = progPackage;
		this.program = program;
		
		//logging
		this.marker = this.program.marker;
		
		this.progLogger = LoggerFactory.getLogger(this.program.name() + ":" + this.prog_package);
	}
	
	public abstract FixedArrayNavigableMap<String, BufferType> inputs();
	
	public int InDegree(){
		return this.inputs().size();
	}
	
	public abstract FixedArray<BufferType> outputs();
	
	public int OutDegree(){
		return this.outputs().size();
	}
	
	public abstract void execute();
	
	//logging
	
	protected Logger logger(){
		return logger;
	}

	protected final void trace(final String msg) {
		this.logger().trace(this.marker, msg);
	}

	protected final void trace(final String format, final Object arg) {
		this.logger().trace(this.marker, format, arg);
	}

	protected final void trace(final String format, final Object arg1, final Object arg2) {
		this.logger().trace(this.marker, format, arg1, arg2);
	}

	protected final void trace(final String format, final Object... argArray) {
		this.logger().trace(this.marker, format, argArray);
	}

	protected final void trace(final String msg, final Throwable t) {
		this.logger().trace(this.marker, msg, t);
	}

	protected final void debug(final String msg) {
		this.logger().debug(this.marker, msg);
	}

	protected final void debug(final String format, final Object arg) {
		this.logger().debug(this.marker, format, arg);
	}

	protected final void debug(final String format, final Object arg1, final Object arg2) {
		this.logger().debug(this.marker, format, arg1, arg2);
	}

	protected final void debug(final String format, final Object... argArray) {
		this.logger().debug(this.marker, format, argArray);
	}

	protected final void debug(final String msg, final Throwable t) {
		this.logger().debug(this.marker, msg, t);
	}

	protected final void info(final String msg) {
		this.logger().info(this.marker, msg);
	}

	protected final void info(final String format, final Object arg) {
		this.logger().info(this.marker, format, arg);
	}

	protected final void info(final String format, final Object arg1, final Object arg2) {
		this.logger().info(this.marker, format, arg1, arg2);
	}

	protected final void info(final String format, final Object... argArray) {
		this.logger().info(this.marker, format, argArray);
	}

	protected final void info(final String msg, final Throwable t) {
		this.logger().info(this.marker, msg, t);
	}

	protected final void warn(final String msg) {
		this.logger().warn(this.marker, msg);
	}

	protected final void warn(final String format, final Object arg) {
		this.logger().warn(this.marker, format, arg);
	}

	protected final void warn(final String format, final Object arg1, final Object arg2) {
		this.logger().warn(this.marker, format, arg1, arg2);
	}

	protected final void warn(final String format, final Object... argArray) {
		this.logger().warn(this.marker, format, argArray);
	}

	protected final void warn(final String msg, final Throwable t) {
		this.logger().warn(this.marker, msg, t);
	}

	protected final void error(final String msg) {
		this.logger().error(this.marker, msg);
	}

	protected final void error(final String format, final Object arg) {
		this.logger().error(this.marker, format, arg);
	}

	protected final void error(final String format, final Object arg1, final Object arg2) {
		this.logger().error(this.marker, format, arg1, arg2);
	}

	protected final void error(final String format, final Object... argArray) {
		this.logger().error(this.marker, format, argArray);
	}

	protected final void error(final String msg, final Throwable t) {
		this.logger().error(this.marker, msg, t);
	}

}
