package de.lmu.ifi.pms.cep.event_mill.execution.program;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;

import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;

import de.lmu.ifi.pms.java.LoggingObject;
import de.lmu.ifi.pms.java.util.collection.array.Arrays;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArraySet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;


public abstract class StreamBuffer<BufferType extends StreamBuffer<BufferType,QueryType,ProgramType>, QueryType extends ExecutableQuery<ProgramType,BufferType>, ProgramType extends ExecutableProgram<?, BufferType, QueryType>> extends LoggingObject{
	
	public final String prog_package;
	public final String simple_name;
	public final String full_name;
	
	public final DTSA_BufferType type;
	
	public final ProgramType program;
	
	//logging
	protected static final Logger logger = LoggerFactory.getLogger(StreamBuffer.class);
	public final Marker marker;
	
	protected final Logger progLogger;
	
	public StreamBuffer(final String buffer_prog_package, final String simple_name, final DTSA_BufferType type, final ProgramType program) {
		this.prog_package = buffer_prog_package;
		this.simple_name = simple_name;
		this.full_name = TSA_SyntaxPatterns.bufferFullName(buffer_prog_package, simple_name);
		this.type = type;
		
		this.program = program;
		
		//logging
		this.marker = this.program.marker;
		
		this.progLogger = LoggerFactory.getLogger(TSA_SyntaxPatterns.bufferQualifiedName(program.name(), this.full_name));
	}
	
	public final String name(){
		return this.full_name;
	}
	
	public final String simpleName(){
		return this.simple_name;
	}
	
	public abstract boolean exists();
	
	public abstract boolean isEmpty();
	
	public abstract void clear();
	
	public abstract FixedArraySet<QueryType> inputQueries();
	
	public int InDegree(){
		return this.inputQueries().size();
	}
	
	public abstract FixedArraySet<QueryType> outputQueries();
	
	public int OutDegree(){
		return this.outputQueries().size();
	}
	
	//logging
	
	@Override
	protected Logger logger(){
		return logger;
	}
	
	//static
	
	public static String[] decode_full_name(final String full_name){
		Pattern decodePattern = Pattern.compile("(\\A(?:\\w+[.])*)(\\w+)\\z");
		Matcher decoder = decodePattern.matcher(full_name);
		if(decoder.matches()){
			return new String[]{decoder.group(1),decoder.group(2)};
		}
		else{
			throw new IllegalArgumentException("Invalid full_name: " + full_name);
		}
	}
	
	public static String[] decode_prog_package(final String prog_package){
		Pattern decodePattern = Pattern.compile("(:?\\w+[.])");
		Matcher decoder = decodePattern.matcher(prog_package);
		if(prog_package.matches("\\A(:?\\w+[.])*\\z")){
			ArrayList<String> componets = new ArrayList<String>();
			while(decoder.find()){
				componets.add(decoder.group(1));
			}
			return componets.toArray(new String[componets.size()]);
		}
		else{
			throw new IllegalArgumentException("Invalid prog_package: " + prog_package);
		}
	}
	
	public static void main(final String[] args) {
		
		System.out.println(Arrays.toString(decode_full_name("b")));
		System.out.println(Arrays.toString(decode_full_name("a.b")));
		System.out.println(Arrays.toString(decode_full_name("a.b.c")));
		System.out.println(Arrays.toString(decode_full_name("ab.bc.cd")));
		System.out.println(Arrays.toString(decode_full_name("ab1.b2c.cd")));
		
		try{
			System.out.println(Arrays.toString(decode_full_name("ab.bc.")));
		} catch(IllegalArgumentException e){
			e.printStackTrace();
		}
		
		System.out.println(Arrays.toString(decode_prog_package("")));
		System.out.println(Arrays.toString(decode_prog_package("bc.")));
		System.out.println(Arrays.toString(decode_prog_package("ab.bc.")));
		System.out.println(Arrays.toString(decode_prog_package("ab.bc.cd.de.ef.fg.gh.hi.ij.jk.")));
		
		try{
			System.out.println(Arrays.toString(decode_prog_package("ab.bc.cd")));
		} catch(IllegalArgumentException e){
			e.printStackTrace();
		}
	}

}
