package de.lmu.ifi.pms.cep.event_mill.execution.scheduler;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;

import de.lmu.ifi.pms.cep.event_mill.execution.engine.Engine;
import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableQuery;


public abstract class SchedulerThread extends Thread{
	
	final Engine engine;
	
	//logging
	protected static final Logger logger = LoggerFactory.getLogger(SchedulerThread.class);	
	public final Marker marker;
	
	public SchedulerThread(final Engine engine)
	{
		super(engine.marker.getName() + ": SchedulerThread");
		this.engine = engine;
		
		this.marker = this.engine.marker;
	}

	abstract Iterable<Iterable<? extends ExecutableQuery<?,?>>> getRounds();
	
	abstract void close();
	
	@Override
	public abstract void run();
	
	
	//logging	
	
	protected Logger logger(){
		return logger;
	}

	protected final void trace(final String msg) {
		this.logger().trace(this.marker, msg);
	}

	protected final void trace(final String format, final Object arg) {
		this.logger().trace(this.marker, format, arg);
	}

	protected final void trace(final String format, final Object arg1, final Object arg2) {
		this.logger().trace(this.marker, format, arg1, arg2);
	}

	protected final void trace(final String format, final Object[] argArray) {
		this.logger().trace(this.marker, format, argArray);
	}

	protected final void trace(final String msg, final Throwable t) {
		this.logger().trace(this.marker, msg, t);
	}

	protected final void debug(final String msg) {
		this.logger().debug(this.marker, msg);
	}

	protected final void debug(final String format, final Object arg) {
		this.logger().debug(this.marker, format, arg);
	}

	protected final void debug(final String format, final Object arg1, final Object arg2) {
		this.logger().debug(this.marker, format, arg1, arg2);
	}

	protected final void debug(final String format, final Object[] argArray) {
		this.logger().debug(this.marker, format, argArray);
	}

	protected final void debug(final String msg, final Throwable t) {
		this.logger().debug(this.marker, msg, t);
	}

	protected final void info(final String msg) {
		this.logger().info(this.marker, msg);
	}

	protected final void info(final String format, final Object arg) {
		this.logger().info(this.marker, format, arg);
	}

	protected final void info(final String format, final Object arg1, final Object arg2) {
		this.logger().info(this.marker, format, arg1, arg2);
	}

	protected final void info(final String format, final Object[] argArray) {
		this.logger().info(this.marker, format, argArray);
	}

	protected final void info(final String msg, final Throwable t) {
		this.logger().info(this.marker, msg, t);
	}

	protected final void warn(final String msg) {
		this.logger().warn(this.marker, msg);
	}

	protected final void warn(final String format, final Object arg) {
		this.logger().warn(this.marker, format, arg);
	}

	protected final void warn(final String format, final Object arg1, final Object arg2) {
		this.logger().warn(this.marker, format, arg1, arg2);
	}

	protected final void warn(final String format, final Object[] argArray) {
		this.logger().warn(this.marker, format, argArray);
	}

	protected final void warn(final String msg, final Throwable t) {
		this.logger().warn(this.marker, msg, t);
	}

	protected final void error(final String msg) {
		this.logger().error(this.marker, msg);
	}

	protected final void error(final String format, final Object arg) {
		this.logger().error(this.marker, format, arg);
	}

	protected final void error(final String format, final Object arg1, final Object arg2) {
		this.logger().error(this.marker, format, arg1, arg2);
	}

	protected final void error(final String format, final Object[] argArray) {
		this.logger().error(this.marker, format, argArray);
	}

	protected final void error(final String msg, final Throwable t) {
		this.logger().error(this.marker, msg, t);
	}

}
