package de.lmu.ifi.pms.cep.event_mill.io;

import de.lmu.ifi.pms.cep.event_mill.execution.program.StreamBufferStore;
import de.lmu.ifi.pms.cep.event_mill.io.stream.SinkStream;
import de.lmu.ifi.pms.cep.event_mill.io.stream.SourceStream;
import de.lmu.ifi.pms.cep.event_mill.io.stream.StreamIO;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.filter.AbstractFilter;
import de.lmu.ifi.pms.java.util.filter.Filter;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public abstract class EngineIO<StoreType extends StreamBufferStore<StoreType, ?>> extends StreamIO{
	
	public final Class<StoreType> type;
	public final String streamBufferStoreIdentifier;
	public final String program_full_name;
	
	public static boolean isInternalAttribute(final String attribute){
		return attribute.startsWith(".") | attribute.contains("..");
	}
	
	public static final Filter<String> relevantAttributeFilter = new AbstractFilter<String>() {
		
		@Override
		public boolean accepts(final String attribute) {
			return !isInternalAttribute(attribute);
		}
		
	};
	
	public EngineIO(final Class<StoreType> type, final String streamBufferStoreIdentifier, final String program_full_name){
		super(PATTERNS.programIdentifier(streamBufferStoreIdentifier, program_full_name));
		this.type = type;
		this.streamBufferStoreIdentifier = streamBufferStoreIdentifier;
		this.program_full_name = program_full_name;
	}
	
	public static String programInstFullName(final String namespace, final String simple_name, final String instance){
		return TSA_SyntaxPatterns.programInstFullName(namespace, simple_name, instance);
	}
	
	public static void checkBufferType(final DTSA_BufferType buffer_type){
		if(buffer_type == DTSA_BufferType.WORKING){
			throw new IllegalArgumentException("Information about internal buffers should not be accessed");
		}
	}
	
	public abstract FixedArrayNavigableSet<String> bufferNames(final DTSA_BufferType buffer_type);
	
	public FixedArrayNavigableSet<String> attributeNames(final DTSA_BufferType buffer_type, final String buffer_name){
		checkBufferType(buffer_type);
		return this.attributeTypes(buffer_type, buffer_name).keys();
	}
	
	@Override
	public FixedArrayNavigableSet<String> sinkNames() {
		return this.bufferNames(DTSA_BufferType.INPUT);
	}
	
	@Override
	public FixedArrayNavigableMap<String, BasicTypeID> sinkTypes(final String name) {
		return this.attributeTypes(DTSA_BufferType.INPUT, name);
	}
	
	@Override
	protected abstract SinkStream newSinkStream(String buffer_name);
	
	@Override
	public FixedArrayNavigableSet<String> sourceNames() {
		return this.bufferNames(DTSA_BufferType.OUTPUT);
	}
	
	
	@Override
	public FixedArrayNavigableMap<String, BasicTypeID> sourceTypes(final String name) {
		return this.attributeTypes(DTSA_BufferType.OUTPUT, name);
	}
	
	@Override
	protected abstract SourceStream newSourceStream(String buffer_name);
	
	public abstract FixedArrayNavigableMap<String, BasicTypeID> attributeTypes(final DTSA_BufferType buffer_type, final String buffer_name);
	
	public synchronized BasicTypeID attributeType(final DTSA_BufferType buffer_type, final String buffer_name, final String attribute_name){
		checkBufferType(buffer_type);
		return this.attributeTypes(buffer_type, buffer_name).get(attribute_name);
	}
	
	/*public synchronized Iterable<ResultSet> outputStream(){
		
	}*/
	
	@Override
	public synchronized void close(){

	}


}
