package de.lmu.ifi.pms.cep.event_mill.io;

import java.util.NoSuchElementException;

import de.lmu.ifi.pms.cep.event_mill.io.catalog.DataCatalog;
import de.lmu.ifi.pms.cep.event_mill.io.stream.StreamIO;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class FixedDataCatalog extends DataCatalog {
	
	final FixedArrayNavigableMap<String, FixedArrayNavigableMap<String, BasicTypeID>> catalog;
	
	public FixedDataCatalog(final StreamIO io, final DTSA_BufferType type){
		FixedArrayNavigableSet<String> messageTypes;		
		switch (type) {
			case INPUT:
				messageTypes = io.sinkNames();
				break;
			case OUTPUT:
				messageTypes = io.sourceNames();
				break;
			default: throw new RuntimeException("Only INPUT or OUTPUT supported.");
		}
		@SuppressWarnings("unchecked")
		FixedArrayNavigableMap<String, BasicTypeID>[] attributeTypes = new FixedArrayNavigableMap[messageTypes.size()];
		for(int attribute = 0; attribute < attributeTypes.length; ++attribute){
			switch (type) {
				case INPUT:
					attributeTypes[attribute] = io.sinkTypes(messageTypes.get(attribute));
					break;
				case OUTPUT:
					attributeTypes[attribute] = io.sourceTypes(messageTypes.get(attribute));
					break;
				default: throw new RuntimeException("Only INPUT or OUTPUT supported.");
			}
		}
		this.catalog = new FixedArrayNavigableMap<String, FixedArrayNavigableMap<String, BasicTypeID>>(messageTypes, new FixedArray<FixedArrayNavigableMap<String, BasicTypeID>>(false, attributeTypes));
	}
	
	public FixedDataCatalog(final FixedArrayNavigableMap<String, FixedArrayNavigableMap<String, BasicTypeID>> catalog) {
		super();
		this.catalog = catalog;
	}

	@Override
	public FixedArrayNavigableSet<String> messageTypes() {
		return this.catalog.keys();
	}

	@Override
	public FixedArrayNavigableMap<String, BasicTypeID> attributeTypes(final String messageType){
		FixedArrayNavigableMap<String, BasicTypeID> attributeTypes = this.catalog.get(messageType);
		if(attributeTypes != null){
			return attributeTypes;
		}
		else{
			//TODO better Exception
			throw new NoSuchElementException();
		}
	}

}
