package de.lmu.ifi.pms.cep.event_mill.io.connect;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.io.BufferedMessageSink;
import de.lmu.ifi.pms.cep.event_mill.io.MessageSource;
import de.lmu.ifi.pms.cep.event_mill.io.bus.MessageBus;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageFormatProvider;
import de.lmu.ifi.pms.cep.event_mill.io.stream.SinkStream;
import de.lmu.ifi.pms.cep.event_mill.io.stream.StreamIO;

import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class BusToStreamConnector<MessageType> extends Connector{
	
	public final Class<MessageType> type;
	final MessageSource<MessageType> source;
	final MessageFormatProvider<MessageType> formatter;
	final FixedArrayNavigableMap<String, BufferedMessageSink<MessageType>> messageSinks;
	final BufferedMessageSink<MessageType>[] messageSinkArray;
	
	public BusToStreamConnector(final MessageBus<MessageType> messageBus, final StreamIO io) {
		this(messageBus.formatter(), messageBus.source(), io.sinkStreams());
	}
	
	public static <MessageType> FixedArrayNavigableMap<String, BufferedMessageSink<MessageType>> messageSinks(final MessageFormatProvider<MessageType> formatter, final SinkStream... sinks){
		String[] messageTypes = new String[sinks.length];
		@SuppressWarnings("unchecked")
		BufferedMessageSink<MessageType>[] messageSinks = (BufferedMessageSink<MessageType>[]) new BufferedMessageSink<?>[sinks.length];
		for(int messageType = 0; messageType < messageTypes.length; ++messageType){
			messageTypes[messageType] = sinks[messageType].name;
			messageSinks[messageType] = formatter.messageSink(sinks[messageType]);
		}
		return new FixedArrayNavigableMap<String, BufferedMessageSink<MessageType>>(messageTypes, messageSinks);
	}

	@SuppressWarnings("unchecked")
	public BusToStreamConnector(final MessageFormatProvider<MessageType> formatter, final MessageSource<MessageType> source, final SinkStream... sinks) {
		super();
		this.type = source.type;
		this.source = source;
		this.formatter = formatter;
		this.messageSinks = messageSinks(formatter, sinks);
		this.messageSinkArray = this.messageSinks.values().toArray((BufferedMessageSink<MessageType>[]) new BufferedMessageSink<?>[this.messageSinks.size()]);
	}
	
	@Override
	public void nextBlock(){
		MessageType[] messages;
		String messageType;
		BufferedMessageSink<MessageType> messageSink;
		while(true){
			messages = this.source.readBlock();
			if(messages.length > 0){
				for(MessageType message: messages){
					messageType = this.formatter.messageType(message);
					messageSink = this.messageSinks.get(messageType);
					if(messageSink != null){
						boolean nextSucceded = messageSink.next();
						if(!nextSucceded){
							messageSink.nextBlock();
							nextSucceded = messageSink.next();
							assert nextSucceded;
						}
						messageSink.add(message);
					}
					else{
						this.debug("Ignored message of unknown type: \"{}\" message:\n{}", messageType, message);
					}
				}
				for(int messageSinkIndex = 0; messageSinkIndex < this.messageSinkArray.length; ++messageSinkIndex){
					this.messageSinkArray[messageSinkIndex].nextBlock();
				}
			}
			else{
				break;
			}
		}
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(BusToStreamConnector.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}
	
	public static <MessageType> BusToStreamConnector<MessageType> create(final MessageBus<MessageType> messageBus, final StreamIO io){
		return new BusToStreamConnector<MessageType>(messageBus, io);
	}

	@Override
	public void close(){
		this.source.close();
		for(BufferedMessageSink<MessageType> messageSink: this.messageSinkArray){
			messageSink.close();
		}
	}
}
