package de.lmu.ifi.pms.cep.event_mill.io.format;

import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.*;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.java.util.collection.array.Arrays;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


public class StringPatternReader extends MessageReader<String>{
	
	final static String defaultEscapeSequence = "\\";
	final static String defaultParameterStartSequence = "%{";
	final static String defaultParameterEndSequence = "}";
	final static ParameterFormatter defaultFormater = new ParameterFormatter();

	//final boolean relaxWhiteSpace;
	
	public final String pattern;
	String message;
	final String[] parameterValues;
	final int[] parameterReferences;
	final String[] splitedPattern;
	
	public final String escapeSequence;
	public final String escapeRegEx;
	public final String parameterStartSequence;
	public final String parameterStartRegEx;
	public final String parameterEndSequence;
	public final String parameterEndRegEx;
	
	public final String parameterRegEx;
	public final Pattern parameterPattern;
	
	public final String unquoteStringRegEx;
	public final Pattern unquoteStringPattern;
	
	public final String unquoteParameterRegEx;
	public final Pattern unquoteParameterPattern;
	
	public final ParameterFormatter formater;
	
	public String quoteString(final String s){
		return this.unquoteStringPattern.matcher(s).replaceAll(Matcher.quoteReplacement(this.escapeSequence) + "$1");
	}
	
	public String quoteParameter(final String parameter){
		return this.unquoteParameterPattern.matcher(parameter).replaceAll("$1");
	}
	
	public String unquoteString(final String s){
		return this.unquoteStringPattern.matcher(s).replaceAll("$1");
	}
	
	public String unquoteParameter(final String parameter){
		return this.unquoteParameterPattern.matcher(parameter).replaceAll("$1");
	}
	
	public StringPatternReader(final String pattern, final String[] parameters, final BasicTypeID[] types) {
		this(defaultFormater, pattern, parameters, types);
	}
	
	public StringPatternReader(final ParameterFormatter formater, final String pattern, final String[] parameters, final BasicTypeID[] types) {
		this(formater, pattern, new FixedArray<String>(parameters), new FixedArray<BasicTypeID>(types));
	}
	
	public StringPatternReader(final String pattern, final FixedArray<String> parameters, final FixedArray<BasicTypeID> types) {
		this(defaultFormater, pattern, parameters, types);
	}
	
	public StringPatternReader(final ParameterFormatter formater, final String pattern, final FixedArray<String> parameters, final FixedArray<BasicTypeID> types) {
		this(defaultEscapeSequence, defaultParameterStartSequence, defaultParameterEndSequence, formater, pattern, parameters, types);
	}
	
	public StringPatternReader(final String pattern, final FixedArrayNavigableMap<String, BasicTypeID> types) {
		this(defaultFormater, pattern, types);
	}
	
	public StringPatternReader(final ParameterFormatter formater, final String pattern, final FixedArrayNavigableMap<String, BasicTypeID> types) {
		this(formater, pattern, types.keys(), types.values());
	}
	
	public StringPatternReader(final String escapeSequence, final String parameterStartSequence, final String parameterEndSequence, final String pattern, final String[] parameters, final BasicTypeID[] types) {
		this(escapeSequence, parameterStartSequence, parameterEndSequence, defaultFormater, pattern, parameters, types);
	}
	
	public StringPatternReader(final String escapeSequence, final String parameterStartSequence, final String parameterEndSequence, final ParameterFormatter formater, final String pattern, final String[] parameters, final BasicTypeID[] types) {
		this(escapeSequence, parameterStartSequence, parameterEndSequence, defaultFormater, pattern, new FixedArray<String>(parameters), new FixedArray<BasicTypeID>(types));
	}
	
	public StringPatternReader(final String escapeSequence, final String parameterStartSequence, final String parameterEndSequence, final String pattern, final FixedArrayNavigableMap<String, BasicTypeID> types) {
		this(escapeSequence, parameterStartSequence, parameterEndSequence, defaultFormater, pattern, types);
	}
	
	public StringPatternReader(final String escapeSequence, final String parameterStartSequence, final String parameterEndSequence, final ParameterFormatter formater, final String pattern, final FixedArrayNavigableMap<String, BasicTypeID> types) {
		this(escapeSequence, parameterStartSequence, parameterEndSequence, formater, pattern, types.keys(), types.values());
	}
	
	public StringPatternReader(final String escapeSequence, final String parameterStartSequence, final String parameterEndSequence, final String pattern, final FixedArray<String> parameters, final FixedArray<BasicTypeID> types) {
		this(escapeSequence, parameterStartSequence, parameterEndSequence, defaultFormater, pattern, parameters, types);
	}

	public StringPatternReader(final String escapeSequence, final String parameterStartSequence, final String parameterEndSequence, final ParameterFormatter formater, final String pattern, final FixedArray<String> parameters, final FixedArray<BasicTypeID> types) {
		super(String.class, parameters, types);
		//this.relaxWhiteSpace = relaxWhiteSpace;
		
		this.escapeSequence = escapeSequence;
		this.escapeRegEx = Pattern.quote(this.escapeSequence);
		this.parameterStartSequence = parameterStartSequence;
		this.parameterStartRegEx = Pattern.quote(this.parameterStartSequence);
		this.parameterEndSequence = parameterEndSequence;
		this.parameterEndRegEx = Pattern.quote(this.parameterEndSequence);
		
		this.parameterRegEx = "(?<!" + this.escapeRegEx + ")" + this.parameterStartRegEx + "(.*?)" + "(?<!" + this.escapeRegEx + ")" + this.parameterEndRegEx;

		this.parameterPattern = Pattern.compile(this.parameterRegEx);
		this.unquoteStringRegEx = this.escapeRegEx + "(" + this.escapeRegEx + "|" + this.parameterStartRegEx + ")";
		this.unquoteStringPattern = Pattern.compile(this.unquoteStringRegEx);
		this.unquoteParameterRegEx = this.escapeRegEx + "(" + this.escapeRegEx + "|" + this.parameterEndRegEx + ")";
		this.unquoteParameterPattern = Pattern.compile(this.unquoteParameterRegEx);
		
		this.formater = formater;
		
		this.pattern = pattern;
		
		this.parameterValues = new String[this.parameters.size()];
		
		
		Matcher parameterMatcher = this.parameterPattern.matcher(pattern);
		ArrayList<String> splitedPattern = new ArrayList<String>();
		ArrayList<Integer> parameterReferences = new ArrayList<Integer>();
		int startIndex = 0;
		String parameter;
		int parameterIndex;
		if(parameterMatcher.find()){
			assert parameterMatcher.end() >= startIndex;
			splitedPattern.add(this.unquoteString(pattern.substring(startIndex, parameterMatcher.start())));
			startIndex = parameterMatcher.end();
			parameter = this.unquoteParameter(parameterMatcher.group(1));
			parameterIndex = this.indexOf(parameter);
			if(parameterIndex < 0){
				throw new IllegalArgumentException("Reference to undeclared parameter \"" + parameter + "\"");
			}
			assert parameterIndex >= 1;
			parameterReferences.add(parameterIndex - 1);
						
			while(parameterMatcher.find()){
				assert parameterMatcher.end() >= startIndex;
				splitedPattern.add(this.unquoteString(pattern.substring(startIndex, parameterMatcher.start())));
				startIndex = parameterMatcher.end();
				parameter = this.unquoteParameter(parameterMatcher.group(1));
				parameterIndex = this.indexOf(parameter);
				if(parameterIndex < 0){
					throw new IllegalArgumentException("Reference to undeclared parameter \"" + parameter + "\"");
				}
				assert parameterIndex >= 1;
				parameterReferences.add(parameterIndex - 1);
			}
			splitedPattern.add(this.unquoteString(pattern.substring(startIndex, pattern.length())));
			this.splitedPattern = splitedPattern.toArray(new String[splitedPattern.size()]);
		}
		else{
			this.splitedPattern = new String[]{pattern};
		}		
		
		//this.parameterReferences = parameterReferences.toArray(new int[parameterReferences.size()]);
		this.parameterReferences = new int[parameterReferences.size()];
		for(int parameterRef = 0; parameterRef < this.parameterReferences.length; ++parameterRef){
			this.parameterReferences[parameterRef] = parameterReferences.getInt(parameterRef);
		}
		
		assert this.splitedPattern.length == this.parameterReferences.length + 1;
	}
	
	public void decompose(final String message, final String[] parameterValues){
		assert parameterValues.length == this.parameterValues.length;
		if(this.message != null){
			this.clearMessage();
		}
		this.message = message;
		int start = 0;
		String nextString;
		BasicTypeID parameterType;
		int parameter;
		String parameterValue;
		for(int parameterRef = 0; parameterRef < this.parameterReferences.length; ++parameterRef){
			nextString = this.splitedPattern[parameterRef];
			if(message.startsWith(nextString, start)){
				start += nextString.length();
			}
			else{
				throw new IllegalArgumentException("Unexpected input.\nExpected:" + nextString + "\nFound:" + message.substring(start));
			}
			parameter = this.parameterReferences[parameterRef];
			parameterType = this.types[parameter];
			
			if(parameterValues[parameter] == null){//use first value
				parameterValue = this.formater.matchParameter(message, start, message.length(), parameterType);
				parameterValues[parameter] = parameterValue;
				start += parameterValue.length();
			}

		}
		nextString = this.splitedPattern[this.parameterReferences.length];
		if(message.startsWith(nextString, start)){
			start += nextString.length();
		}
		else{
			throw new IllegalArgumentException("Unexpected input.\nExpected:" + nextString + "\nFound:" + message.substring(start));
		}
		if(start < message.length()){
			throw new IllegalArgumentException("Unexpected additional, unmatched input:" + message.substring(start));
		}
	}
	
	@Override
	public void decompose(final String message) {
		this.decompose(message, this.parameterValues);
	}

	@Override
	public void clearMessage() {
		this.message = null;
		Arrays.fill(this.parameterValues, null);
	}

	@Override
	public boolean isNull(final int parameterIndex) {
		return this.parameterValues[parameterIndex - 1] == null;
	}

	@Override
	public boolean getBoolean(final int parameterIndex) {
		return this.formater.parseBoolean(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public byte getByte(final int parameterIndex) {
		return this.formater.parseByte(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public short getShort(final int parameterIndex) {
		return this.formater.parseShort(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public int getInt(final int parameterIndex) {
		return this.formater.parseInt(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public long getLong(final int parameterIndex) {
		return this.formater.parseLong(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public long getTimestamp(final int parameterIndex) {
		return this.formater.parseTimestamp(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public long getDuration(final int parameterIndex) {
		return this.formater.parseDuration(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public long getIdentifier(final int parameterIndex) {
		return this.formater.parseIdentifier(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public float getFloat(final int parameterIndex) {
		return this.formater.parseFloat(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public double getDouble(final int parameterIndex) {
		return this.formater.parseDouble(this.parameterValues[parameterIndex - 1]);
	}

	@Override
	public String getString(final int parameterIndex) {
		return this.formater.parseString(this.parameterValues[parameterIndex - 1]);
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(StringPatternReader.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}

	public static void main(final String[] args) {
		String pattern = "person{ name{%{person.%{name}}, address{ street{%{person.\\}address.street}}, number{%{person.a\\\\ddress.number}}}} }";
		
		StringPatternReader reader = new StringPatternReader(pattern, 
				new String[]{"person.%{name", "person.}address.street", "person.a\\ddress.number"},
				new BasicTypeID[]{	STRING, STRING, INT}
				);
		
		String message = "person{ name{\"nobody\"}, address{ street{\"Valeria\"}, number{15}}} }";
		
		reader.decompose(message);
		
		System.out.println(reader.getString(1));
		System.out.println(reader.getString(2));
		System.out.println(reader.getInt(3));
		
		reader.clearMessage();
	}

}
