package de.lmu.ifi.pms.cep.event_mill.io.stream;

import de.lmu.ifi.pms.cep.event_mill.io.SpecificBufferedMessageSource;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageFormatProvider;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageGenerator;
import de.lmu.ifi.pms.cep.event_mill.sql.io.SQL_EngineIO;
import de.lmu.ifi.pms.cep.event_mill.sql.io.SQL_SourceStream;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

public class MessageSourceStream<MessageType> extends SpecificBufferedMessageSource<MessageType>{
	
	final MessageGenerator<MessageType> messageGenerator;
	final SourceStream sourceStream;
	public final int numberOfAttributes;
	final BasicTypeID[] types;
	
	public MessageSourceStream(final String bufferName, final StreamIO io, final MessageFormatProvider<MessageType> formatProvider){
		this(io.sourceStream(bufferName), formatProvider);
	}
	
	public MessageSourceStream(final String bufferName, final long lastReadId, final SQL_EngineIO io, final MessageFormatProvider<MessageType> formatProvider){
		this(new SQL_SourceStream(bufferName, lastReadId, io), formatProvider);
	}
	
	public MessageSourceStream(final String bufferName, final StreamIO io, final MessageGenerator<MessageType> messageGenerator){
		this(io.sourceStream(bufferName), messageGenerator);
	}
	
	public MessageSourceStream(final String bufferName, final long lastReadId, final SQL_EngineIO io, final MessageGenerator<MessageType> messageGenerator){
		this(new SQL_SourceStream(bufferName, lastReadId, io), messageGenerator);
	}
	
	public MessageSourceStream(final SourceStream sourceStream, final MessageFormatProvider<MessageType> formatProvider){
		this(sourceStream, formatProvider.messageGenerator(sourceStream.name, sourceStream.types));
	}
	
	public MessageSourceStream(final SourceStream sourceStream, final MessageGenerator<MessageType> messageGenerator){
		super(messageGenerator.type, sourceStream.name);
		this.sourceStream = sourceStream;
		this.messageGenerator = messageGenerator;//FIXME remove attributes with "." at the beginning and void setting attributes that are not used in the pattern
		this.numberOfAttributes = sourceStream.types.size();
		this.types = sourceStream.types.values().toArray(new BasicTypeID[this.numberOfAttributes]);
	}
	
	/*
	@Override
	public Iterator<String> iterator() {
		return new AbstractFixedIterator<String>() {
			
			@Override
			public boolean hasNext() {
				return MessageStream.this.hasNext();
			}
			
			@Override
			public String next() {
				return MessageStream.this.next();
			}
			
		};
	}
	*/
	
	@Override
	public boolean nextBlock(){
		return this.sourceStream.nextBlock();
	}

	@Override
	public boolean next(){
		return this.sourceStream.next();
	}
	
	@Override
	public MessageType current(){
    	BasicTypeID attributeType;
    	
    	for (int attribute = 1; attribute <= this.numberOfAttributes; ++attribute){
    		attributeType = this.types[attribute - 1];
    		switch (attributeType) {
			case BOOLEAN:
				this.messageGenerator.setBoolean(attribute, this.sourceStream.getBoolean(attribute));
				break;
			case STRING:
				this.messageGenerator.setString(attribute, this.sourceStream.getString(attribute));
				break;
			case INT:
				this.messageGenerator.setInt(attribute, this.sourceStream.getInt(attribute));
				break;
			case LONG:
				this.messageGenerator.setLong(attribute, this.sourceStream.getLong(attribute));
				break;				
			case FLOAT:
				this.messageGenerator.setFloat(attribute, this.sourceStream.getFloat(attribute));
				break;
			case DOUBLE:
				this.messageGenerator.setDouble(attribute, this.sourceStream.getDouble(attribute));
				break;
			case TIMESTAMP:
				this.messageGenerator.setTimestamp(attribute, this.sourceStream.getTimestamp(attribute));
				break;				
			case DURATION:
				this.messageGenerator.setDuration(attribute, this.sourceStream.getDuration(attribute));
				break;				
			case IDENTIFIER:
				this.messageGenerator.setIdentifier(attribute, this.sourceStream.getIdentifier(attribute));
				break;
			case REFERENCE: throw new IllegalArgumentException("Type \"reference\" not supported.");
			default:
				break;
			}
    	}
    	MessageType message = this.messageGenerator.compose();
    	this.messageGenerator.clearParameters();
    	return message;
	}
	
	@Override
	public void close() {
		this.sourceStream.close();
	}
	
	@Override
	protected void finalize() throws Throwable {
		this.close();
	}
	
}
