package de.lmu.ifi.pms.cep.event_mill.sql;

import java.util.ArrayList;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.Root;
import org.simpleframework.xml.Serializer;
import org.simpleframework.xml.core.Persister;

import de.lmu.ifi.pms.cep.COMMON_CONSTANTS;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.composition.FixedComposedCollection;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.map.fixed.FixedEnumMap;
import de.lmu.ifi.pms.java.util.string.StringBuilder;


@Root(name="program")
public class SQL_Program_Spec {
	
	@Element(name="namespace", required=false)
	public final String namespace;
	@Element(name="simple_name")
	public final String simple_name;
	public final String fullName;
	
	@Element(name="description", required=false)
	public final String description;
	
	protected final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, SQL_Buffer_Spec>> buffer_specs;
	
	public final FixedArray<SQL_Query_Spec> query_specs;
	
	public final int bufferMaxNumberOfMinTimes;
	public final int queryMaxNumberOfOutputMinTimes;
	public final int queryMaxNumberOfInputMinTimes;

	public SQL_Program_Spec(final String namespace, final String simple_name, final String description,
							final Collection<SQL_Buffer_Spec> buffer_specs,
							final Collection<SQL_Query_Spec> query_specs) {
		super();
		assert namespace == null || namespace.matches(TSA_SyntaxPatterns.ProgramNamespaceRegEx);
		assert simple_name.matches(TSA_SyntaxPatterns.ProgramNameRegEx);
		
		this.namespace = (namespace == null ? "" : namespace);
		this.simple_name = simple_name;
		this.fullName = TSA_SyntaxPatterns.programFullName(this.namespace, this.simple_name);
		this.description = (description == null ? "" : description);
		
		@SuppressWarnings("unchecked")
		ArrayList<String>[] buffer_names = new ArrayList[DTSA_BufferType.count];
		@SuppressWarnings("unchecked")
		ArrayList<SQL_Buffer_Spec>[] buffers = new ArrayList[DTSA_BufferType.count];
		for(int ordinal = 0; ordinal < DTSA_BufferType.count; ++ordinal){
			buffer_names[ordinal] = new ArrayList<String>();
			buffers[ordinal] = new ArrayList<SQL_Buffer_Spec>();
		}
		
		int bufferMaxNumberOfMinTimes = 0;
		for(SQL_Buffer_Spec buffer_spec: buffer_specs){
			int ordinal = buffer_spec.type.ordinal();
			buffer_names[ordinal].add(buffer_spec.fullName);
			buffers[ordinal].add(buffer_spec);
			
			if(buffer_spec.min_times.size() > bufferMaxNumberOfMinTimes){
				bufferMaxNumberOfMinTimes = buffer_spec.min_times.size(); 
			}
		}
		
		@SuppressWarnings("unchecked")
		FixedArrayNavigableMap<String, SQL_Buffer_Spec>[] buffer_spec_maps = new FixedArrayNavigableMap[DTSA_BufferType.count];
		for(int ordinal = 0; ordinal < DTSA_BufferType.count; ++ordinal){
			buffer_spec_maps[ordinal] = new FixedArrayNavigableMap<String, SQL_Buffer_Spec>(new FixedArray<String>(false, buffer_names[ordinal].toArray(new String[buffer_names[ordinal].size()])), new FixedArray<SQL_Buffer_Spec>(false, buffers[ordinal].toArray(new SQL_Buffer_Spec[buffers[ordinal].size()])));
		}
		
		this.buffer_specs = new FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String,SQL_Buffer_Spec>>(DTSA_BufferType.class, buffer_spec_maps);
		this.query_specs = new FixedArray<SQL_Query_Spec>(query_specs);
		
		this.bufferMaxNumberOfMinTimes = bufferMaxNumberOfMinTimes;
		
		int queryMaxNumberOfOutputMinTimes = 0;
		int queryMaxNumberOfInputMinTimes = 0;
		for(SQL_Query_Spec query_spec: this.query_specs){
			if(query_spec.numberOfOutputMinTimes > queryMaxNumberOfOutputMinTimes){
				queryMaxNumberOfOutputMinTimes = query_spec.numberOfOutputMinTimes; 
			}
			
			int queryNumberOfInputMinTimes = 0;			
			for(String inputBuffer: query_spec.inputBuffers){
				queryNumberOfInputMinTimes += this.buffer_specs.get(query_spec.inputBuffersType).get(inputBuffer).min_times.size();
			}			
			if(queryNumberOfInputMinTimes > queryMaxNumberOfInputMinTimes){
				queryMaxNumberOfInputMinTimes = queryNumberOfInputMinTimes;
			}			
		}
		this.queryMaxNumberOfOutputMinTimes = queryMaxNumberOfOutputMinTimes;
		this.queryMaxNumberOfInputMinTimes = queryMaxNumberOfInputMinTimes;
	}
	
	public SQL_Program_Spec(final String namespace, final String simple_name, final String description,
			final String xml_buffer_specs,
			final String xml_query_specs) {
		this(namespace, simple_name, description, SQL_Buffer_Spec.deserialize(xml_buffer_specs), SQL_Query_Spec.deserialize(xml_query_specs));
	}
	
	SQL_Program_Spec(	@Element(name="namespace", required=false) final String namespace, @Element(name="simple_name") final String simple_name, 
			@Element(name="description", required=false) final String description,
			@ElementList(name="buffers", entry="buffer", inline=false) final java.util.ArrayList<SQL_Buffer_Spec> buffer_specs,
			@ElementList(name="queries", entry="query", inline=false) final java.util.ArrayList<SQL_Query_Spec> query_specs) {
		this(namespace, simple_name, description, new FixedArray<SQL_Buffer_Spec>(buffer_specs), new FixedArray<SQL_Query_Spec>(query_specs));
	}
	
	public Collection<SQL_Buffer_Spec> bufferSpecs(){
		@SuppressWarnings("unchecked")
		FixedArray<SQL_Buffer_Spec>[] bufferSpecs = new FixedArray[DTSA_BufferType.count];
		for(int ordinal = 0; ordinal < DTSA_BufferType.count; ++ordinal){
			bufferSpecs[ordinal] = this.buffer_specs.values().get(ordinal).values();
		}
		return new FixedComposedCollection<SQL_Buffer_Spec>(bufferSpecs);
	}
	
	public FixedArrayNavigableSet<String> bufferNames(){
		@SuppressWarnings("unchecked")
		FixedArrayNavigableSet<String>[] bufferNames = new FixedArrayNavigableSet[DTSA_BufferType.count];
		for(int ordinal = 0; ordinal < DTSA_BufferType.count; ++ordinal){
			bufferNames[ordinal] = this.buffer_specs.values().get(ordinal).keys();
		}
		return FixedArrayNavigableSet.merge(bufferNames);
	}
	
	public FixedArrayNavigableSet<String> attributeNames(){
		Collection<SQL_Buffer_Spec> bufferSpecs = this.bufferSpecs();
		@SuppressWarnings("unchecked")
		FixedArrayNavigableSet<String>[] attributeNames = new FixedArrayNavigableSet[bufferSpecs.size()];
		int index = 0;
		for(SQL_Buffer_Spec bufferSpec: bufferSpecs){
			attributeNames[index] = bufferSpec.attributes.keys();
			++index;
		}
		return FixedArrayNavigableSet.merge(attributeNames);
	}
	
	@ElementList(name="buffers", entry="buffer", inline=false)
	ArrayList<SQL_Buffer_Spec> getBuffer_specs(){
		return new java.util.ArrayList<SQL_Buffer_Spec>(this.bufferSpecs().javaCollection());
	}
	
	@ElementList(name="queries", entry="query", inline=false)
	ArrayList<SQL_Query_Spec> getQuery_specs(){
		return new java.util.ArrayList<SQL_Query_Spec>(this.query_specs.javaCollection());
	}
	
	@Override
	public String toString() {
		Serializer serializer = new Persister();
		StringBuilder result = new StringBuilder();
		try {
			serializer.write(this, result);
			return result.toString();
		} catch (Exception e) {
			throw new Error(e);
		}
	}
	
	public static SQL_Program_Spec deserialize(final String xml_spec){
		Serializer serializer = new Persister(COMMON_CONSTANTS.SERIALIZATION_FORMAT);
		try {
			return serializer.read(SQL_Program_Spec.class, xml_spec);
		} catch (Exception e) {
			throw new IllegalArgumentException(e);
		}
	}
	
	public static String serialize(final SQL_Program_Spec program_specs){
		Serializer serializer = new Persister(COMMON_CONSTANTS.SERIALIZATION_FORMAT);
		StringBuilder result = new StringBuilder();
		try {
			serializer.write(program_specs, result);
			return result.toString();
		} catch (Exception e) {
			throw new Error(e);
		}
	}
	
}
