package de.lmu.ifi.pms.cep.event_mill.sql;

import java.sql.Connection;
import java.sql.SQLException;

import de.lmu.ifi.pms.cep.COMMON_CONSTANTS;
import de.lmu.ifi.pms.cep.event_mill.configuration.DEFAULTS;
import de.lmu.ifi.pms.cep.event_mill.execution.program.StreamBuffer;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.time.Clock;
import de.lmu.ifi.pms.development.UnimplementedMethodError;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArraySet;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class SQL_StreamBuffer extends StreamBuffer<SQL_StreamBuffer, SQL_ExecutableQuery, SQL_ExecutableProgram>{
	
	public final int id;
	public final String schema;
	public final String table_name;
	
	public final int numberOfMinTimes;
	public final FixedArrayNavigableSet<String> min_times;
	protected final long[] writeMinTimes;
	protected final long[] readMinTimes;
	protected boolean readMinTimesHaveChanged;
	
	protected long tupleCounter;
	protected long size;
	
	protected FixedArraySet<SQL_ExecutableQuery> inputQueries;
	protected FixedArraySet<SQL_ExecutableQuery> outputQueries;
	
	protected final Clock block_Counter;
	
//	public SQL_StreamBuffer(CERA_MaterialisationPoint materialisationPoint, SQL_StreamBufferStore bufferStore) {
//		this(materialisationPoint, bufferStore, materialisationPoint.Name);
//	}

	public SQL_StreamBuffer(final SQL_ExecutableProgram program, final int id, final String prog_package, final String simple_name, final FixedArrayNavigableMap<String, BasicTypeID> attributes, final FixedArrayNavigableSet<String> min_times, final DTSA_BufferType type, final String SQL_TableName) {
		super(prog_package, simple_name, type, program);
		
		this.id = id;
		this.schema = program.bufferSchemas.get(this.type);
		this.table_name = SQL_TableName;
		
		this.min_times = min_times;
		this.numberOfMinTimes = this.min_times.size();
		this.writeMinTimes = new long[this.numberOfMinTimes];
		this.readMinTimes = new long[this.numberOfMinTimes];
		this.readMinTimesHaveChanged = true;
		
		this.tupleCounter = 0;
		//FIXME init buffer size
		this.size = 0;//Get From Database
		
		this.inputQueries = new FixedArraySet<SQL_ExecutableQuery>();
		this.outputQueries = new FixedArraySet<SQL_ExecutableQuery>();
		
		this.program.register(this);
		
		//FIXME
		this.block_Counter = program.clocks.get(COMMON_CONSTANTS.BLOCK_COUNTER_CLOCK_NAME(this.type, this.full_name));
		assert this.block_Counter != null;
	}
	
	public synchronized long[] minWriteTimes()
	{
		if(this.type == DTSA_BufferType.INPUT){
			//FIXME
			if(this.numberOfMinTimes > 0){
				assert this.min_times.get(0).equals(DEFAULTS._o_block);
				this.writeMinTimes[0] = this.block_Counter.min();
			}
		}
		else{
			this.program.erase(this.writeMinTimes);
			long[] queryOutputMinTimes;
			for(SQL_ExecutableQuery query: this.inputQueries()){
				queryOutputMinTimes = query.outputMinWriteTimes();
				for(int index = 0; index < this.numberOfMinTimes; ++index){
					if(queryOutputMinTimes[index] < this.writeMinTimes[index]){
						this.writeMinTimes[index] = queryOutputMinTimes[index];
					}
				}
			}
		}
		return this.writeMinTimes;//FIXME copy
	}
	
	public synchronized long[] minReadTimes()
	{//FIXME implementation / synchronization
	//FIXME for garbage collection
		throw new UnimplementedMethodError();
		/*if(this.readMinTimesHaveChanged){
			this.program.erase(this.readMinTimes);
			long[] queryInputMinTimes;
			for(SQL_ExecutableQuery query: this.outputQueries()){
				queryInputMinTimes = query.inputMinTimes();
				//FIXME arity of inputMinTimes of a query may be different (depending on the (all) inputs of the query; garbage collection depends on time of other inputs)
				for(int index = 0; index < this.numberOfMinTimes; ++index){
					if(queryInputMinTimes[index] < this.readMinTimes[index]){
						this.readMinTimes[index] = queryInputMinTimes[index];
					}
				}
			}
			this.readMinTimesHaveChanged = false;
		}
		return this.readMinTimes;
		*/
	}
	
	@Override
	public synchronized FixedArraySet<SQL_ExecutableQuery> inputQueries()
	{
		return this.inputQueries;
	}
	
	public synchronized void RegisterInputQuery(final SQL_ExecutableQuery query){
		SQL_ExecutableQuery[] inputQueries = new SQL_ExecutableQuery[this.inputQueries.size() + 1];
		for(int input = 0; input < this.inputQueries.size(); ++input){
			if(query == this.inputQueries.get(input)){//TODO better test?
				return;
			}
			inputQueries[input] = this.inputQueries.get(input);
		}
		inputQueries[inputQueries.length - 1] = query;
		this.inputQueries = new FixedArraySet<SQL_ExecutableQuery>(false, false, inputQueries);
	}
	
	@Override
	public synchronized FixedArraySet<SQL_ExecutableQuery> outputQueries()
	{
		return this.outputQueries;
	}
	
	public synchronized void RegisterOutputQuery(final SQL_ExecutableQuery query){
		SQL_ExecutableQuery[] outputQueries = new SQL_ExecutableQuery[this.outputQueries.size() + 1];
		for(int output = 0; output < this.outputQueries.size(); ++output){
			if(query == this.outputQueries.get(output)){//TODO better test?
				return;
			}
			outputQueries[output] = this.outputQueries.get(output);
		}
		outputQueries[outputQueries.length - 1] = query;
		this.outputQueries = new FixedArraySet<SQL_ExecutableQuery>(false, false, outputQueries);
	}
	
	public synchronized void notifyRead(){
		this.readMinTimesHaveChanged = true;
	}
	
	public synchronized void notifyWrite(final long tupleNumber){
		if(tupleNumber > 0){
			this.tupleCounter += tupleNumber;
		}
		this.size += tupleNumber;
		
		assert(true); //FIXME check whether this.size == SELECT COUNT() FROM
	}
	
	protected synchronized long tupleCount(){
		return this.tupleCounter;
	}
	
	@Override
	public boolean exists(){
		//FIXME move to SQL_Translation
		Connection connection = this.program.bufferStore.connection();
		try
		{
			return !connection.getMetaData().getTables(null, this.schema, this.table_name, null).isAfterLast();
			//TODO better implementation, particular type check, now only checks whether a table with a correct name does exist
		}
		catch (SQLException e)
		{
			throw new RuntimeException(e); //TODO better exception
		}
	}
	
	@Override
	public boolean isEmpty(){
		//FIXME move to SQL_Translation
		String createQuery = "SELECT * FROM \"" + this.schema + "\".\"" + this.table_name + "\";";
		try
		{
			return this.program.bufferStore.statement().executeQuery(createQuery).isAfterLast();
			//TODO more efficient implementation
		}
		catch (SQLException e)
		{
			throw new RuntimeException(e); //TODO better exception
		}
	}
	

	@Override
	public void clear()
	{
		//FIXME move to SQL_Translation
		try
		{
			//FIXME
			this.program.bufferStore.statement().execute("DELETE FROM \"" + this.schema + "\".\"" + this.table_name + "\";");
		}
		catch (SQLException e)
		{
			throw new RuntimeException(e);//TODO better exception
		}
	}
	
}
