package de.lmu.ifi.pms.cep.event_mill.sql;

import static de.lmu.ifi.pms.cep.tsa.conditions.Term.*;

import java.io.IOException;
import java.io.Writer;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementUnion;
import org.simpleframework.xml.Root;
import org.simpleframework.xml.Serializer;
import org.simpleframework.xml.core.Persister;

import de.lmu.ifi.pms.cep.COMMON_CONSTANTS;
import de.lmu.ifi.pms.cep.tsa.conditions.ClockFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.DurationValue;
import de.lmu.ifi.pms.cep.tsa.conditions.Greatest;
import de.lmu.ifi.pms.cep.tsa.conditions.Least;
import de.lmu.ifi.pms.cep.tsa.conditions.StreamBound;
import de.lmu.ifi.pms.cep.tsa.conditions.Minus;
import de.lmu.ifi.pms.cep.tsa.conditions.Plus;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula.BottomC;
import de.lmu.ifi.pms.java.util.adaption.CollectionAdapter;
import de.lmu.ifi.pms.java.util.map.Entry;
import de.lmu.ifi.pms.java.util.map.Map;
import de.lmu.ifi.pms.java.util.map.compatibility.TreeMap;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.string.StringBuilder;


@Root(name="time")
public class TimeSpec {
	
	@Attribute(name = "name")
	public final String name;
	
	@ElementUnion({
		@Element(name="false", type=BottomC.class),
		@Element(name="plus", type=Plus.class),
		@Element(name="minus", type=Minus.class),
		@Element(name="min", type=Least.class),
		@Element(name="max", type=Greatest.class),
		@Element(name="min-time", type=StreamBound.class),
		@Element(name="clock", type=ClockFunction.class),
		@Element(name="duration", type=DurationValue.class)
	})
	protected Term<?> definition;
	
	protected TimeSpec(@Attribute(name = "name") final String name) {
		super();
		this.name = name;
	}

	public TimeSpec(final String name, final Term<?> definition) {
		this(name);
		this.definition = definition;
	}
	
	public Term<?> definition(){
		return this.definition;
	}
	
	public static java.util.ArrayList<TimeSpec> fromMap(final Map<String, Term<?>, ?> definitions){
		return new java.util.ArrayList<TimeSpec>(new CollectionAdapter<Entry<String, Term<?>>, TimeSpec>(definitions) {
			
			@Override
			public TimeSpec adapt(final Entry<String, Term<?>> source) {
				return new TimeSpec(source.key(), source.value());
			}
			
		}.javaCollection());
	}
	
	public static FixedArrayNavigableMap<String, Term<?>> toMap(final java.util.ArrayList<TimeSpec> definitions){
		TreeMap<String, Term<?>> map = new TreeMap<String, Term<?>>();
		for(TimeSpec timeSpec: definitions){
			map.add(timeSpec.name, timeSpec.definition);
		}
		return new FixedArrayNavigableMap<String, Term<?>>(map);
	}

	public static void main(final String[] args) throws Exception{
		Serializer serializer = new Persister(COMMON_CONSTANTS.SERIALIZATION_FORMAT);
		
		Term<?> duration = new DurationValue(10000);
		Term<?> var1 = var("reception");
		Term<?> var2 = var("detection");
		Term<?> var3 = var("emission");
		Term<?> max = apply(TupleFunctionID.GREATEST, var1, var2);
		Term<?> plus = apply(TupleFunctionID.PLUS, duration, max);
		Term<?> min = apply(TupleFunctionID.GREATEST, plus, var3);
		
		TimeSpec time = new TimeSpec("reception", min);
		
		Writer writer = new StringBuilder();
		try {
			serializer.write(time, writer);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
		String serialized_term = writer.toString();
		System.out.println(serialized_term);
		
		@SuppressWarnings("unused")
		TimeSpec deserialized_term;
		
		deserialized_term =  serializer.read(TimeSpec.class, serialized_term);
		
		try {
			System.in.read();
		} catch (IOException e) {
	
		}
	}

}
