package de.lmu.ifi.pms.cep.examples.access_control;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.Iterator;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.configuration.IOConfiguration;
import de.lmu.ifi.pms.cep.event_mill.console.AbstractConsoleThread;
import de.lmu.ifi.pms.cep.event_mill.io.format.StringPatternGenerator;
import de.lmu.ifi.pms.cep.event_mill.io.stream.MessageSourceStream;
import de.lmu.ifi.pms.cep.event_mill.io.stream.StreamIO;


public class GrantAccessConsole extends AbstractConsoleThread{
	
	StreamIO io;
	
	MessageSourceStream<String> grantAccessMessages;
	MessageSourceStream<String> denyAccessMessages;
	MessageSourceStream<String> intrusionWarningMessages;
	long delay = 1000;
	
	boolean running = false;
	boolean quit = false;

	public GrantAccessConsole(final Iterator<String> immediateCommands, final BufferedReader inputStream, final PrintStream outputStream) {
		super("GrantAccessConsole", immediateCommands, inputStream, outputStream);
		
		try {
			this.io = IOConfiguration.read().streamProvider("de.lmu.ifi.pms.cep.event-mill.prototype").connect("de.lmu.ifi.pms.cep.event-mill.prototype:example.access-control:test");
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	public void run(){
		
//		try {
			
			String command;
			this.outputStream.println("Welcome,");
			this.outputStream.println("Type \"start\" when you are ready.");
			this.outputStream.println("(For leaving type \"quit\")");
			
			while(!this.quit){
			while(!this.running){
				command = this.nextCommand();
				if(command.equalsIgnoreCase("quit")){
					this.quit = true;
					
					this.outputStream.println("Bye!");
					break;
				}
				else if(command.equalsIgnoreCase("start")){
					if(this.grantAccessMessages == null){
						this.grantAccessMessages = new MessageSourceStream<String>("example.access-control.grant-access", this.io, new StringPatternGenerator("Grant access to %{person}", this.io.sourceTypes("example.access-control.grant-access")));
						this.denyAccessMessages = new MessageSourceStream<String>("example.access-control.deny-access", this.io, new StringPatternGenerator("Deny access to %{person}", this.io.sourceTypes("example.access-control.deny-access")));
						this.intrusionWarningMessages = new MessageSourceStream<String>("example.access-control.intrusion-warning", this.io, new StringPatternGenerator("Intrusion warning: %{person}", this.io.sourceTypes("example.access-control.intrusion-warning")));
					}
					this.running = true;
				}
				else{
					this.outputStream.println("What do you mean?");
		
				}
			}
			
			long earliestNewBlockRead = System.currentTimeMillis() - this.delay;
			long currentTime;
			long delay;
			while(this.running){
				try {
					if(this.inputStream.ready()){
						command = this.nextCommand();
						if(command.equalsIgnoreCase("quit")){
							this.running = false;
							this.quit = true;
							
							this.outputStream.println("Bye!");
							continue;
						}
						else if(command.equalsIgnoreCase("stop")){
							this.running = false;
							
							continue;
						}
						else{
							this.outputStream.println("What do you mean?");

						}
					}
				} catch (IOException e) {
					this.error("Problem with input stream", e);
					e.printStackTrace();
					throw new RuntimeException(e);
				}
				

				currentTime = System.currentTimeMillis();
				delay = earliestNewBlockRead - currentTime;
				while(delay > 0){
					try {
						Thread.sleep(delay);
					} catch (InterruptedException e) {

					}
					currentTime = System.currentTimeMillis();
					delay = earliestNewBlockRead - currentTime;
				}
				
				
				this.grantAccessMessages.nextBlock();
				
				earliestNewBlockRead = currentTime + this.delay;
				
				while(this.grantAccessMessages.next()){
		        	this.outputStream.println(this.grantAccessMessages.current());
		        	this.outputStream.flush();
				}
				
				
				this.denyAccessMessages.nextBlock();
				
				earliestNewBlockRead = currentTime + this.delay;
				
				while(this.denyAccessMessages.next()){
		        	this.outputStream.println(this.denyAccessMessages.current());
		        	this.outputStream.flush();
				}
				
				
				this.intrusionWarningMessages.nextBlock();
				
				earliestNewBlockRead = currentTime + this.delay;
				
				while(this.intrusionWarningMessages.next()){
		        	this.outputStream.println(this.intrusionWarningMessages.current());
		        	this.outputStream.flush();
				}
			}
			}
		
//		} catch (Exception e) {
//
//		}
		
		System.exit(0);
	}
	
	//logging
	protected static final Logger logger = LoggerFactory.getLogger(GrantAccessConsole.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}
	
	public static void main(final String[] args) {
		BufferedReader inputStream = new BufferedReader(new InputStreamReader(System.in));
		PrintStream outputStream = System.out;
		
		new GrantAccessConsole(java.util.Arrays.asList(args).iterator(), inputStream, outputStream).start();
	}
	
}
