package de.lmu.ifi.pms.cep.tsa;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementListUnion;
import org.simpleframework.xml.ElementUnion;

import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_Operator;
import de.lmu.ifi.pms.cep.tsa.schema.ActionStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.EventStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.StateStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;

import de.lmu.ifi.pms.java.util.adaption.CollectionAdapter;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;


public class DTSA_CompositeExpression extends DTSA_Expression {
	
	@Element(name="operator")
	public final TSA_Operator operator;
	
	public final FixedArray<DTSA_Expression> subExpressions;
	
	@ElementListUnion({
		@ElementList(entry="stream", type=DTSA_StreamBuffer.class, inline=true), 
		@ElementList(entry="apply", type=DTSA_CompositeExpression.class, inline=true),
		@ElementList(entry="expression", type=DTSA_Expression.class, inline=true)
	})
	java.util.ArrayList<DTSA_Expression> getSubExpressionList(){
		return new java.util.ArrayList<DTSA_Expression>(this.subExpressions.javaCollection());
	}
	
	@ElementUnion({
		@Element(name="event-schema", type=EventStreamSchema.class),
		@Element(name="state-schema", type=StateStreamSchema.class),
		@Element(name="action-schema", type=ActionStreamSchema.class),
		@Element(name="schema", type=TemporalStreamSchema.class)
	})
	protected final TemporalStreamSchema schema;
	
	public DTSA_CompositeExpression(final TSA_Operator operator, final FixedArray<DTSA_Expression> subExpressions) throws TSA_Exception {
		this(operator, subExpressions, null);
		
	}
	
	protected DTSA_CompositeExpression(@Element(name="operator") final TSA_Operator operator
			, @ElementList(entry="expression", type=DTSA_Expression.class, inline=true) final java.util.ArrayList<DTSA_Expression> subExpressions
			, @Element(name="schema", type=TemporalStreamSchema.class) final TemporalStreamSchema schema) throws TSA_Exception {
		this(operator, new FixedArray<DTSA_Expression>(subExpressions), schema);
	}
	
	protected DTSA_CompositeExpression(final TSA_Operator operator, final FixedArray<DTSA_Expression> subExpressions, final TemporalStreamSchema schema) throws TSA_Exception {
		super(allSubexpressionsStatic(subExpressions), allSubexpressionsMonotonic(subExpressions) & operator.isMonotonic());
		this.operator = operator;
		this.subExpressions = subExpressions;
		
		if(schema == null){
			this.schema = this.operator.apply(new CollectionAdapter<DTSA_Expression, TemporalStreamSchema>(this.subExpressions) {
				@Override
				public TemporalStreamSchema adapt(final DTSA_Expression source) {
					return source.schema();
				}
			});
		}
		else{
			this.schema = schema;
			assert (this.schema.isStatic == this.isStatic);
		}
	}
	
	protected static boolean allSubexpressionsStatic(final FixedArray<DTSA_Expression> subExpressions){
		for(DTSA_Expression subExpression: subExpressions){
			if(!subExpression.isStatic){
				return false;
			}
		}
		return true;
	}
	
	protected static boolean allSubexpressionsMonotonic(final FixedArray<DTSA_Expression> subExpressions){
		for(DTSA_Expression subExpression: subExpressions){
			if(!subExpression.isMonotonic){
				return false;
			}
		}
		return true;
	}
	
	@Override
	public DTSA_Expression replaceSchema(final TemporalStreamSchema newSchema) throws TSA_Exception {
		return new DTSA_CompositeExpression(this.operator, this.subExpressions, newSchema); 
	}

	public DTSA_CompositeExpression(final TSA_Operator operator, final DTSA_Expression subExpression) throws TSA_Exception {
		this(operator, new FixedArray<DTSA_Expression>(subExpression));
	}
			
	@Override
	public TemporalStreamSchema schema() {
		return this.schema;
	}
	
	@Override
	protected void inputs(final ArrayList<DTSA_StreamBuffer> inputs) {
		for(DTSA_Expression subExpression: this.subExpressions){
			subExpression.inputs(inputs);
		}
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepthFirstLine, final int indentDepth) {
		this.operator.writeTo(sb, indentDepthFirstLine, indentDepth);
		assert this.subExpressions.size() > 0;
		sb.append("(\n");
		for(DTSA_Expression subExpression: this.subExpressions){
			subExpression.writeTo(sb, indentDepth + 1);
			sb.append(",\n");
		}
		sb.setLength(sb.length() - ",\n".length());
		sb.append("\n");
		indentTab(sb, indentDepth);
		sb.append(")");
	}
	
	@Override
	public void check() throws TSA_Exception {
		for(DTSA_Expression subExpression: this.subExpressions){
			subExpression.check();
		}
		this.operator.check(new CollectionAdapter<DTSA_Expression, TemporalStreamSchema>(this.subExpressions) {
			@Override
			public TemporalStreamSchema adapt(final DTSA_Expression source) {
				return source.schema();
			}
		});
	}

}
