package de.lmu.ifi.pms.cep.tsa;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementMap;
import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.tsa.conditions.StreamBound;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_TypeException;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.Type;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


@Root(name="buffer")
public class DTSA_StreamBuffer extends DTSA_Expression{
	
	@Attribute(name="package", required=true)
	public final String prog_package;
	@Attribute(name="simple_name", required=true)
	public final String simple_name;
	public final String full_name;
	
	@Element(name="description", required=false)
	public final String description;
	
	@Element(name="type", required=true)
	public final DTSA_BufferType type;
	
	@Element(name="schema", required=true)
	public final TemporalStreamSchema schema;
	
	public final FixedArrayNavigableMap<String, Term<? extends Type>> implicitAttributeSpecifications;
	
	@ElementMap(entry="implicit-attribute-specifications", inline=true, key="min-time", attribute=true, required=false)
	final java.util.HashMap<String, Term<? extends Type>> getImplicitAttributeSpecifications(){
		return new java.util.HashMap<String, Term<? extends Type>>(this.implicitAttributeSpecifications.javaMap());
	}
	
	public final FixedArrayNavigableMap<String, StreamBound> streamBounds;
	
	@ElementList(name="stream-bounds", entry="stream-bound", inline=true, required=false, type=String.class)
	java.util.ArrayList<String> getMinTime(){
		return new java.util.ArrayList<String>(this.streamBounds.keys().javaCollection());
	}
	
	public final FixedCollection<String> origin;
	
	@ElementList(name="origins", entry="origin", inline=true, required=false, type=String.class)
	java.util.ArrayList<String> getOriginList(){
		return new java.util.ArrayList<String>(this.origin.javaCollection());
	}
	
	public DTSA_StreamBuffer(@Attribute(name="package", required=true) final String prog_package
			, @Attribute(name="simple_name", required=true) final String simple_name
			, @Element(name="description", required=false) final String description
			, @Element(name="type", required=true) final DTSA_BufferType type
			, @Element(name="schema", required=true) final TemporalStreamSchema schema
			, @ElementMap(entry="implicit-attribute-specifications", inline=true, key="min-time", attribute=true, required=false) final java.util.HashMap<String, Term<? extends Type>> implicitAttributeSpecifications
			, @ElementList(name="stream-bounds", entry="stream-bound", inline=true, required=false) final java.util.ArrayList<String> streamBounds
			, @ElementList(name="origins", entry="origin", inline=true, required=false) final java.util.ArrayList<String> origin
		) throws TSA_Exception {
		this(prog_package, simple_name, (description == null ? "" : description), type, schema, (implicitAttributeSpecifications == null ? new FixedArrayNavigableMap<String, Term<? extends Type>>() : new FixedArrayNavigableMap<String, Term<? extends Type>>(implicitAttributeSpecifications)), (streamBounds == null ? new FixedArrayNavigableSet<String>() : new FixedArrayNavigableSet<String>(streamBounds)), (origin == null ? new FixedArray<String>() : new FixedArray<String>(origin)));
	}
	
	public DTSA_StreamBuffer(final String prog_package, final String simple_name,
			final String description, final DTSA_BufferType type,
			final TemporalStreamSchema schema,
			final FixedArrayNavigableMap<String, Term<? extends Type>> implicitAttributeSpecifications,
			final FixedArrayNavigableSet<String> streamBounds, final FixedCollection<String> origin) throws TSA_Exception {
		this(new String[]{prog_package, simple_name}, description, type, schema, implicitAttributeSpecifications, streamBounds, origin);		
	}
	
	public DTSA_StreamBuffer(final String full_name, final String description, final DTSA_BufferType type,
			final TemporalStreamSchema schema,
			final FixedArrayNavigableMap<String, Term<? extends Type>> implicitAttributeSpecifications,
			final FixedArrayNavigableSet<String> streamBounds, final FixedCollection<String> origin) throws TSA_Exception {
		this(TSA_SyntaxPatterns.bufferFullName(full_name), description, type, schema, implicitAttributeSpecifications, streamBounds, origin);
	}

	protected DTSA_StreamBuffer(final String[] full_name, final String description, final DTSA_BufferType type,
			final TemporalStreamSchema schema,
			final FixedArrayNavigableMap<String, Term<? extends Type>> implicitAttributeSpecifications,
			final FixedArrayNavigableSet<String> streamBounds, final FixedCollection<String> origin) throws TSA_Exception {
		super(schema.isStatic,true);
		this.prog_package = full_name[0];
		this.simple_name = full_name[1];
		this.full_name = TSA_SyntaxPatterns.bufferFullName(this.prog_package, this.simple_name);
		this.description = description;
		this.type = type;
		this.schema = schema;
		this.implicitAttributeSpecifications = (implicitAttributeSpecifications == null ? new FixedArrayNavigableMap<String, Term<? extends Type>>() : implicitAttributeSpecifications);
		
		StreamBound[] min_time_terms = new StreamBound[streamBounds.size()];
		
		for(int min_time_index = 0; min_time_index < min_time_terms.length; ++min_time_index){
			String name = streamBounds.get(min_time_index);
			try {
				assert this.schema.typeOf(name) instanceof BasicType;
				min_time_terms[min_time_index] = new StreamBound(this.full_name, name, (BasicType) this.schema.typeOf(name));
			} catch (TSA_TypeException e) {
				throw new TSA_Exception(origin, e);
			}
		}
		
		this.streamBounds = new FixedArrayNavigableMap<String, StreamBound>(streamBounds, new FixedArray<StreamBound>(false, min_time_terms));
		
		this.origin = origin;
	}
	
	@Override
	public DTSA_Expression replaceSchema(final TemporalStreamSchema newSchema) throws TSA_Exception {
		return new DTSA_StreamBuffer(this.full_name, this.description, this.type, newSchema, this.implicitAttributeSpecifications, this.streamBounds.keys(), this.origin);
	}
	
	@Override
	public void check() throws TSA_Exception {

	}
	
	@Override
	protected void inputs(final ArrayList<DTSA_StreamBuffer> inputs) {
		inputs.add(this);
	}
	
	public void writeTo(final StringBuilder sb, final int indentDepthFirstLine, final int indentDepth, final boolean finish){
		indentTab(sb, indentDepthFirstLine);
		
		if(this.prog_package != ""){
			sb.append("PACKAGE ");
			sb.append(this.prog_package);
			sb.append("\n\n");
			indentTab(sb, indentDepth);
		}
		sb.append(this.type.toString());
		sb.append("\n");
		indentTab(sb, indentDepth);
		this.schema.writeTo(sb, indentDepth, false, this.simple_name, finish);
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepthFirstLine, final int indentDepth){
		this.writeTo(sb, indentDepthFirstLine, indentDepth, true);
	}
	
	@Override
	public TemporalStreamSchema schema() {
		return this.schema;
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepth){
		this.writeTo(sb, indentDepth, 0, true);
	}

}
