package de.lmu.ifi.pms.cep.tsa.compiler;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.compilation.Compiler;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.TSA_Program;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_RuntimeException;
import de.lmu.ifi.pms.cep.tsa.optimization.DTSA_Optimizer;
import de.lmu.ifi.pms.cep.tsa.optimization.TSA_Optimizer;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;


public class TSA_DTSA_Compiler extends Compiler<TSA_Program, DTSA_Program>{
	
	public final FixedArray<TSA_Optimizer> preDifferenciationOptimizers;
	
	public final TSA_Differencer differencer;
	
	public final FixedArray<DTSA_Optimizer> postDifferenciationOptimizers;
	
	public TSA_DTSA_Compiler(final TSA_Differencer differencer) {
		this(new FixedArray<TSA_Optimizer>(), differencer, new FixedArray<DTSA_Optimizer>());
	}
	
	public TSA_DTSA_Compiler(final TSA_Optimizer[] preDifferenciationOptimizers, final TSA_Differencer differencer, final DTSA_Optimizer[] postDifferenciationOptimizers) {
		this(new FixedArray<TSA_Optimizer>(preDifferenciationOptimizers), differencer, new FixedArray<DTSA_Optimizer>(postDifferenciationOptimizers));
	}
	
	public TSA_DTSA_Compiler(final FixedArray<TSA_Optimizer> preDifferenciationOptimizers, final TSA_Differencer differencer, final FixedArray<DTSA_Optimizer> postDifferenciationOptimizers) {
		super(TSA_Program.class, DTSA_Program.class);
		this.preDifferenciationOptimizers = preDifferenciationOptimizers;
		this.differencer = differencer;
		this.postDifferenciationOptimizers = postDifferenciationOptimizers;
	}

	public void checkReservedAttributes(final TSA_Program tsa_program){
		//FIXME prevent usage of attributes starting with .tsa
	}
	
	public void checkAllowedFunctions(final TSA_Program tsa_program){
		
	}
	
	public void checkAllowedRelations(final TSA_Program tsa_program){
		
	}
	
	public void check(final TSA_Program tsa_program){
		this.checkReservedAttributes(tsa_program);
		this.checkAllowedFunctions(tsa_program);
		this.checkAllowedRelations(tsa_program);
	}

	@Override
	public DTSA_Program compile(TSA_Program tsa_program){
		this.check(tsa_program);
		try {
			tsa_program.check();
		} catch (TSA_Exception e) {
			throw new TSA_RuntimeException(e);
		}
		//FIXME input type should not serve as output of a query (check)??
		
		for(TSA_Optimizer optimizer: this.preDifferenciationOptimizers){
			tsa_program = optimizer.optimize(tsa_program);
		}
		DTSA_Program dtsa_program = this.differencer.differenciate(tsa_program);
		for(DTSA_Optimizer optimizer: this.postDifferenciationOptimizers){
			dtsa_program = optimizer.optimize(dtsa_program);
		}
		return dtsa_program;
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(TSA_DTSA_Compiler.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}
		
}
