package de.lmu.ifi.pms.cep.tsa.compiler;

import static de.lmu.ifi.pms.cep.tsa.conditions.AtomicTermID.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.Term.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.*;
import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.*;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.COMMON_CONSTANTS;
import de.lmu.ifi.pms.cep.event_mill.configuration.DEFAULTS;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.DTSA_CompositeExpression;
import de.lmu.ifi.pms.cep.tsa.DTSA_Expression;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.DTSA_Query;
import de.lmu.ifi.pms.cep.tsa.DTSA_QueryType;
import de.lmu.ifi.pms.cep.tsa.DTSA_StreamBuffer;
import de.lmu.ifi.pms.cep.tsa.TSA_CompositeExpression;
import de.lmu.ifi.pms.cep.tsa.TSA_Expression;
import de.lmu.ifi.pms.cep.tsa.TSA_Program;
import de.lmu.ifi.pms.cep.tsa.TSA_Query;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.TSA_TemporalStream;
import de.lmu.ifi.pms.cep.tsa.TSA_TemporalStreamExpression;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicStreamBoundFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.Attribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeValue;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID;
import de.lmu.ifi.pms.cep.tsa.conditions.BooleanAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BooleanTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.ClockFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID;
import de.lmu.ifi.pms.cep.tsa.conditions.ConjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.DefaultDefinition;
import de.lmu.ifi.pms.cep.tsa.conditions.DisjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.DurationValue;
import de.lmu.ifi.pms.cep.tsa.conditions.IdentifierAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.Parameter;
import de.lmu.ifi.pms.cep.tsa.conditions.StreamBound;
import de.lmu.ifi.pms.cep.tsa.conditions.StreamBoundConstraint;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TypedAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.Variable;
import de.lmu.ifi.pms.cep.tsa.conditions.adaptor.RenameAttributes;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_RuntimeException;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_BasicOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ComplexOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ImbeddingOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ProjectionOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_RenameOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_SelectionOperator;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.Type;
import de.lmu.ifi.pms.cep.tsa.schema.TypeID;
import de.lmu.ifi.pms.cep.tsa.time.ArithmeticRelationFormulaExtractor;
import de.lmu.ifi.pms.cep.tsa.time.AtomicStreamBoundFormulaFilter;
import de.lmu.ifi.pms.cep.tsa.time.StreamBoundConstraintFilter;
import de.lmu.ifi.pms.cep.tsa.time.analysis.ArithmeticAnalyser;
import de.lmu.ifi.pms.cep.tsa.time.analysis.ArithmeticAnalyserFactory;
import de.lmu.ifi.pms.cep.tsa.time.analysis.SimpleArithmeticAnalyzerFactory;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.adaption.CollectionAdapter;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;
import de.lmu.ifi.pms.java.util.filter.AbstractFilter;
import de.lmu.ifi.pms.java.util.filter.Filter;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.map.fixed.FixedEnumMap;


public class TSA_DefaultDifferencer extends TSA_Differencer {
	
	static final BasicTypeAttribute _o_id_attribute = attr(DEFAULTS._o_id, IDENTIFIER);
	static final BasicTypeAttribute _o_block_attribute = attr(DEFAULTS._o_block, IDENTIFIER);
	//static final String id = DuraDefaultAttributeNameMapping.identifierName;
	static final ClockFunction tuple_id_seq = new ClockFunction(COMMON_CONSTANTS.TUPLE_ID_CLOCK_NAME, IDENTIFIER); 
	
	static final String isNewAttributeName = ".tsa.isNew";
	static final String isNewAttributeAliasName = ".tsa.isNew.old";
	static final BooleanAttribute isNewAttribute = BooleanAttribute.create(isNewAttributeName);
	static final BooleanAttribute isNewAliasAttribute = BooleanAttribute.create(isNewAttributeAliasName);
	
	static final AbstractFilter<ConditionFormula> streamBoundConstraintFilter = new StreamBoundConstraintFilter();
	static final AbstractFilter<ConditionFormula> streamBoundFilter = new AtomicStreamBoundFormulaFilter();
	
	final ArithmeticAnalyserFactory identifierAnalyserFactory;
	final ArithmeticAnalyserFactory temporalAnalyserFactory;
	
	public TSA_DefaultDifferencer(){
		this(new SimpleArithmeticAnalyzerFactory(IDENTIFIER), new SimpleArithmeticAnalyzerFactory(TIMESTAMP));
	}
	
	public TSA_DefaultDifferencer(final ArithmeticAnalyserFactory identifierAnalyserFactory, final ArithmeticAnalyserFactory temporalAnalyserFactory) {
		this.identifierAnalyserFactory = identifierAnalyserFactory;
		this.temporalAnalyserFactory = temporalAnalyserFactory;
	}
	
	@Override
	public DTSA_Program differenciate(final TSA_Program program) {
		try {
			return this.differenciateImpl(program);
		} catch (TSA_Exception e) {
			throw new TSA_RuntimeException(e);
		}
	}

	@SuppressWarnings("unchecked")
	public DTSA_Program differenciateImpl(final TSA_Program program) throws TSA_Exception {
		try{
			assert program.isValid();
			
			FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers = this.differentiateBuffers(program.temporalStreams.values());
			
			FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>> implicitQueries = this.implicitQueries(streamBuffers);
			
			FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>> explicitQueries = this.explicitQueries(program.queries);
			
			FixedArray<TSA_Query>[] allQueriesArray = new FixedArray[DTSA_QueryType.values().length];
			for(int type = 0; type < allQueriesArray.length; ++type){
				allQueriesArray[type] = FixedArray.concat(implicitQueries.values().get(type), explicitQueries.values().get(type));
			}
			
			FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>> queries = this.differentiateQueries(streamBuffers, new FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>>(DTSA_QueryType.class, allQueriesArray));
			
			return new DTSA_Program(program.namespace, program.simple_name, program.description, streamBuffers, queries, program.origin);
		}
		catch (TSA_Exception e) {
			throw e;
		}
		catch (RuntimeException e) {
			throw new TSA_Exception(program.origin, e);
		}
	}
	
	public FixedArrayNavigableMap<TypedAttribute<?>, Term<? extends Type>> implicitAttributeSpecifications(final FixedArrayNavigableMap<String, Term<? extends Type>> implicitAttributeSpecifications){
		TypedAttribute<?>[] attributes = new TypedAttribute<?>[implicitAttributeSpecifications.size()];
		for(int attribute = 0; attribute < attributes.length; ++attribute){
			attributes[attribute] = attr(implicitAttributeSpecifications.keys().get(attribute), implicitAttributeSpecifications.values().get(attribute).type);
		}
		return new FixedArrayNavigableMap<TypedAttribute<?>, Term<? extends Type>>(new FixedArrayNavigableSet<TypedAttribute<?>>(false, false, attributes), implicitAttributeSpecifications.values());
	}
	
	public FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> differentiateBuffers(final FixedArray<TSA_TemporalStream> temporalStreams) throws TSA_Exception{
		ArrayList<String> inputStreamBufferNames = new ArrayList<String>();
		ArrayList<DTSA_StreamBuffer> inputStreamBuffers = new ArrayList<DTSA_StreamBuffer>();
		ArrayList<String> workingStreamBufferNames = new ArrayList<String>();
		ArrayList<DTSA_StreamBuffer> workingStreamBuffers = new ArrayList<DTSA_StreamBuffer>();
		ArrayList<String> outputStreamBufferNames = new ArrayList<String>();
		ArrayList<DTSA_StreamBuffer> outputStreamBuffers = new ArrayList<DTSA_StreamBuffer>();
		ArrayList<String> logStreamBufferNames = new ArrayList<String>();
		ArrayList<DTSA_StreamBuffer> logStreamBuffers = new ArrayList<DTSA_StreamBuffer>();
		
		for(final TSA_TemporalStream temporalStream: temporalStreams){

			try{ 

				//TODO treatment of definitions for composite attributes?
				FixedArrayNavigableMap<String, Term<? extends Type>> implicitAttributeSpecifications = temporalStream.schema.implicitAttributeSpecifications;
				if(temporalStream.types.contains(DTSA_BufferType.INPUT)){
	
					final TemporalStreamSchema schema = temporalStream.schema;
					
					
					
					DTSA_StreamBuffer inputStream;
					TemporalStreamSchema newSchema;
					if(implicitAttributeSpecifications.size() > 0){
						final FixedArrayNavigableSet<String> explicitAttributeNames = FixedArrayNavigableSet.difference(schema.basicAttributes.keys(), implicitAttributeSpecifications.keys());
						final AbstractFilter<AtomicConditionFormula> explicitAttributeConstraintsFilter = new AbstractFilter<AtomicConditionFormula>(){
							@Override
							public boolean accepts(final AtomicConditionFormula atom) {
								if(explicitAttributeNames.containsAll(atom.AttributeNames())){
									return true;
								}
								return false;
							}
						};
						if(explicitAttributeNames.size() > 0){
							ConditionFormula newTupleConstraints = new ConjunctiveConditionFormula(schema.tupleConstraints()).focus(explicitAttributeConstraintsFilter).flaten();
							ConditionFormula newRelationConstraints = new ConjunctiveConditionFormula(schema.relationConstraints()).focus(AbstractFilter.notFilter(DefaultDefinition.implicitAttributeValuesSpecifications)).focus(explicitAttributeConstraintsFilter).flaten();
							try {
								newSchema = TemporalStreamSchema.create(schema.ID, schema.basicAttributes, (newTupleConstraints instanceof ConjunctiveConditionFormula ? ((ConjunctiveConditionFormula)newTupleConstraints).subformulas() : new FixedArray<ConditionFormula>(newTupleConstraints)), (newRelationConstraints instanceof ConjunctiveConditionFormula ? ((ConjunctiveConditionFormula)newRelationConstraints).subformulas() : new FixedArray<ConditionFormula>(newRelationConstraints)));
								newSchema = newSchema.retain(explicitAttributeNames);
							} catch (TSA_SchemaException e) {
								throw new TSA_Exception(temporalStream.origin, e);
							}
						}
						else{
							try {
								newSchema = TemporalStreamSchema.create(schema.ID, new FixedArrayNavigableMap<String, Type>(), new FixedArray<ConditionFormula>(), new FixedArray<ConditionFormula>());
							} catch (TSA_SchemaException e) {
								throw new TSA_Exception(temporalStream.origin, e);
							} 
						}
						
						if(!temporalStream.isStatic){
							try{
								newSchema = newSchema.extend(new FixedArrayNavigableMap<String, Type>(new String[]{DEFAULTS._o_block, DEFAULTS._o_id}, new Type[]{IDENTIFIER.type, IDENTIFIER.type})).constraint(new FixedArray<ConditionFormula>(new StreamBoundConstraint(_o_block_attribute)));
							} catch (TSA_SchemaException e) {
								throw new TSA_Exception(temporalStream.origin, e);
							} 
						}
						
						inputStream = this.differentiate(temporalStream.full_name, temporalStream.description, DTSA_BufferType.INPUT, newSchema, temporalStream.origin);
					}
					else{
						if(!temporalStream.schema.isStatic){
							try{
								newSchema = schema.extend(new FixedArrayNavigableMap<String, Type>(new String[]{DEFAULTS._o_id}, new Type[]{IDENTIFIER.type})).constraint(new FixedArray<ConditionFormula>(new StreamBoundConstraint(attr(DEFAULTS._o_id, IDENTIFIER.type))));
							} catch (TSA_SchemaException e) {
								throw new TSA_Exception(temporalStream.origin, e);
							} 
						}
						else{
							newSchema = schema;
						}
						inputStream = this.differentiate(temporalStream.full_name, temporalStream.description, DTSA_BufferType.INPUT, newSchema, temporalStream.origin);
					}
					
					inputStreamBufferNames.add(inputStream.full_name);
					inputStreamBuffers.add(inputStream);
				}
				
				if(temporalStream.types.contains(DTSA_BufferType.WORKING)){			
					DTSA_StreamBuffer workingStream = this.differentiate(temporalStream.full_name, temporalStream.description, DTSA_BufferType.WORKING, temporalStream.schema, temporalStream.origin);
					workingStreamBufferNames.add(workingStream.full_name);
					workingStreamBuffers.add(workingStream);
				}
				
				if(temporalStream.types.contains(DTSA_BufferType.OUTPUT)){
					DTSA_StreamBuffer outputStreamBuffer;
					try {
						outputStreamBuffer = this.differentiate(temporalStream.full_name, temporalStream.description, DTSA_BufferType.OUTPUT, temporalStream.schema.extend(new FixedArrayNavigableMap<String, Type>(new String[]{DEFAULTS._o_id}, new Type[]{IDENTIFIER.type})).constraint(new FixedArray<ConditionFormula>(new StreamBoundConstraint(_o_id_attribute))), temporalStream.origin);
					} catch (TSA_SchemaException e) {
						throw new TSA_Exception(temporalStream.origin, e);
					}
					outputStreamBufferNames.add(outputStreamBuffer.full_name);
					outputStreamBuffers.add(outputStreamBuffer);
				}
				
				if(temporalStream.types.contains(DTSA_BufferType.LOG)){
					DTSA_StreamBuffer logStreamBuffer;
					try {
						logStreamBuffer = this.differentiate(temporalStream.full_name, temporalStream.description, DTSA_BufferType.LOG, temporalStream.schema.extend(new FixedArrayNavigableMap<String, Type>(new String[]{DEFAULTS._o_id}, new Type[]{IDENTIFIER.type})).constraint(new FixedArray<ConditionFormula>(new StreamBoundConstraint(_o_id_attribute))), temporalStream.origin);
					} catch (TSA_SchemaException e) {
						throw new TSA_Exception(temporalStream.origin, e);
					}
					logStreamBufferNames.add(logStreamBuffer.full_name);
					logStreamBuffers.add(logStreamBuffer);
				}
			}
			catch (TSA_Exception e) {
				throw e;
			}
			catch (RuntimeException e) {
				throw new TSA_Exception(temporalStream.origin, e);
			}
		}
		
		@SuppressWarnings("unchecked")
		FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers = new FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String,DTSA_StreamBuffer>>(DTSA_BufferType.class,
				new FixedArrayNavigableMap<String,DTSA_StreamBuffer>(new FixedArrayNavigableSet<String>(false, false, inputStreamBufferNames.toArray(new String[0])), new FixedArray<DTSA_StreamBuffer>(false, inputStreamBuffers.toArray(new DTSA_StreamBuffer[0]))),		
				new FixedArrayNavigableMap<String,DTSA_StreamBuffer>(new FixedArrayNavigableSet<String>(false, false, workingStreamBufferNames.toArray(new String[0])), new FixedArray<DTSA_StreamBuffer>(false, workingStreamBuffers.toArray(new DTSA_StreamBuffer[0]))),
				new FixedArrayNavigableMap<String,DTSA_StreamBuffer>(new FixedArrayNavigableSet<String>(false, false, outputStreamBufferNames.toArray(new String[0])), new FixedArray<DTSA_StreamBuffer>(false, outputStreamBuffers.toArray(new DTSA_StreamBuffer[0]))),
				new FixedArrayNavigableMap<String,DTSA_StreamBuffer>(new FixedArrayNavigableSet<String>(false, false, logStreamBufferNames.toArray(new String[0])), new FixedArray<DTSA_StreamBuffer>(false, logStreamBuffers.toArray(new DTSA_StreamBuffer[0])))
		);
		
		return streamBuffers;
	}
	
	public TSA_Expression extendExpression(final TSA_Expression expression, final DTSA_StreamBuffer output_buffer) throws TSA_Exception{
		FixedArrayNavigableSet<String> missingAttributes = FixedArrayNavigableSet.difference(output_buffer.schema.basicAttributes.keys(), expression.schema().basicAttributes.keys());
		
		if(missingAttributes.size() > 0){
			FixedArrayNavigableMap<String, Term<? extends Type>> implicitAttributeSpecifications = output_buffer.implicitAttributeSpecifications;
			
			//FIXME what happens if both, default for attribute and subattributes?
			
			//extend to non basic attributes
			
			ArrayList<TypedAttribute<?>> implicitAttributes = new ArrayList<TypedAttribute<?>>(implicitAttributeSpecifications.size());
			ArrayList<Term<?>> definitions = new ArrayList<Term<?>>(implicitAttributeSpecifications.size());
			
			for (int specification = 0; specification < implicitAttributeSpecifications.size(); specification++) {
				String name = implicitAttributeSpecifications.keys().get(specification);
				Term<? extends Type> definition = implicitAttributeSpecifications.values().get(specification);
				assert definition.isTyped;
	
				if(!expression.schema().hasAttribute(name)){
					TypedAttribute<?> implicitAttribute = Term.attr(name, definition.type);
					implicitAttributes.add(implicitAttribute);								
					definitions.add(definition);
				}
			}
			
			TSA_ImbeddingOperator imbeddingOperator = new TSA_ImbeddingOperator(output_buffer.origin, new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(implicitAttributes.toArray(new TypedAttribute<?>[0]), definitions.toArray(new Term[0])));
			return new TSA_CompositeExpression(imbeddingOperator, expression);
		}
		else{
			return expression;
		}
	}
	
	FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>> implicitQueries(final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers) throws TSA_Exception{
		@SuppressWarnings("unchecked")
		ArrayList<TSA_Query>[] implicitQueriesListsArray = (ArrayList<TSA_Query>[]) new ArrayList<?>[DTSA_QueryType.values().length];
		for(int type = 0; type < implicitQueriesListsArray.length; ++type){
			implicitQueriesListsArray[type] = new ArrayList<TSA_Query>();
		}		
		FixedEnumMap<DTSA_QueryType, ArrayList<TSA_Query>> implicitQueriesLists = new FixedEnumMap<DTSA_QueryType, ArrayList<TSA_Query>>(DTSA_QueryType.class, implicitQueriesListsArray);
		
		for(DTSA_StreamBuffer inputBuffer: streamBuffers.get(DTSA_BufferType.INPUT).values()){
			try{
				DTSA_StreamBuffer workingBuffer = streamBuffers.get(DTSA_BufferType.WORKING).get(inputBuffer.full_name);
				
				if(workingBuffer != null){					
					TSA_Expression query_expression = this.extendExpression(new TSA_CompositeExpression(new TSA_ProjectionOperator(workingBuffer.origin, FixedArrayNavigableSet.intersect(inputBuffer.schema.basicAttributes.keys(), workingBuffer.schema.basicAttributes.keys())), new TSA_TemporalStreamExpression(inputBuffer.full_name, inputBuffer.type, inputBuffer.schema, inputBuffer.origin)), workingBuffer);
					
					if(query_expression.isStatic){
						if(workingBuffer.isStatic){
							implicitQueriesLists.get(DTSA_QueryType.STATIC).add(new TSA_Query(workingBuffer.prog_package, workingBuffer.full_name, workingBuffer.schema, DTSA_BufferType.WORKING, query_expression, "Initial query for type " + workingBuffer.full_name, workingBuffer.origin));
						}
						else{
							implicitQueriesLists.get(DTSA_QueryType.INIT).add(new TSA_Query(workingBuffer.prog_package, workingBuffer.full_name, workingBuffer.schema, DTSA_BufferType.WORKING, query_expression, "Initial query for type " + workingBuffer.full_name, workingBuffer.origin));
						}
					}
					else{
						implicitQueriesLists.get(DTSA_QueryType.INPUT).add(new TSA_Query(workingBuffer.prog_package, workingBuffer.full_name, workingBuffer.schema, DTSA_BufferType.WORKING, query_expression, "Input query for type " + workingBuffer.full_name, workingBuffer.origin));
					}
				}
			}
			catch (TSA_Exception e) {
				throw e;
			}
			catch (RuntimeException e) {
				throw new TSA_Exception(inputBuffer.origin, e);
			}
		}
		
		for(DTSA_StreamBuffer outputBuffer: streamBuffers.get(DTSA_BufferType.OUTPUT).values()){
			try{
				DTSA_StreamBuffer workingBuffer = streamBuffers.get(DTSA_BufferType.WORKING).get(outputBuffer.full_name);
				
				TSA_Expression query_expression = new TSA_TemporalStreamExpression(workingBuffer.full_name, workingBuffer.type, workingBuffer.schema, workingBuffer.origin);
				query_expression = new TSA_CompositeExpression(new TSA_ImbeddingOperator(outputBuffer.origin, new FixedArrayNavigableMap<BasicTypeAttribute, Term<?>>(new BasicTypeAttribute[]{_o_id_attribute}, new Term<?>[]{tuple_id_seq})), query_expression);
				
				implicitQueriesLists.get(DTSA_QueryType.OUTPUT).add(new TSA_Query(outputBuffer.prog_package, outputBuffer.full_name, outputBuffer.schema, DTSA_BufferType.OUTPUT, query_expression, "Output query for type " + outputBuffer.full_name, outputBuffer.origin));
			}
			catch (TSA_Exception e) {
				throw e;
			}
			catch (RuntimeException e) {
				throw new TSA_Exception(outputBuffer.origin, e);
			}
		}
		
		for(DTSA_StreamBuffer logBuffer: streamBuffers.get(DTSA_BufferType.LOG).values()){
			try{
				DTSA_StreamBuffer workingBuffer = streamBuffers.get(DTSA_BufferType.WORKING).get(logBuffer.full_name);
				
				TSA_Expression query_expression = new TSA_TemporalStreamExpression(workingBuffer.full_name, workingBuffer.type, workingBuffer.schema, workingBuffer.origin);
				query_expression = new TSA_CompositeExpression(new TSA_ImbeddingOperator(logBuffer.origin, new FixedArrayNavigableMap<BasicTypeAttribute, Term<?>>(new BasicTypeAttribute[]{_o_id_attribute}, new Term<?>[]{tuple_id_seq})), query_expression);
				
				implicitQueriesLists.get(DTSA_QueryType.LOG).add(new TSA_Query(logBuffer.prog_package, logBuffer.full_name, logBuffer.schema, DTSA_BufferType.LOG, query_expression, "Log query for type " + logBuffer.full_name, logBuffer.origin));
			}
			catch (TSA_Exception e) {
				throw e;
			}
			catch (RuntimeException e) {
				throw new TSA_Exception(logBuffer.origin, e);
			}
		}
		
		@SuppressWarnings("unchecked")
		FixedArray<TSA_Query>[] implicitQueriesArray = (FixedArray<TSA_Query>[]) new FixedArray<?>[implicitQueriesListsArray.length];
		for(int type = 0; type < implicitQueriesArray.length; ++type){
			implicitQueriesArray[type] = new FixedArray<TSA_Query>(implicitQueriesListsArray[type]);
		}
		FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>> implicitQueries = new FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>>(DTSA_QueryType.class,implicitQueriesArray);
		return implicitQueries;
	}
	
	public FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>> explicitQueries(final FixedArray<TSA_Query> queries) throws TSA_Exception{
		@SuppressWarnings("unchecked")
		ArrayList<TSA_Query>[] explicitQueriesListsArray = (ArrayList<TSA_Query>[]) new ArrayList<?>[DTSA_QueryType.values().length];
		for(int type = 0; type < explicitQueriesListsArray.length; ++type){
			explicitQueriesListsArray[type] = new ArrayList<TSA_Query>();
		}		
		FixedEnumMap<DTSA_QueryType, ArrayList<TSA_Query>> explicitQueriesLists = new FixedEnumMap<DTSA_QueryType, ArrayList<TSA_Query>>(DTSA_QueryType.class, explicitQueriesListsArray);
		
		for(TSA_Query query: queries){
			if(query.isStatic){
				if(query.output_isStatic){
					explicitQueriesLists.get(DTSA_QueryType.STATIC).add(query);
				}
				else{
					explicitQueriesLists.get(DTSA_QueryType.INIT).add(query);	
				}
			}
			else{
				explicitQueriesLists.get(DTSA_QueryType.WORKING).add(query);
			}
		}
		
		@SuppressWarnings("unchecked")
		FixedArray<TSA_Query>[] explicitQueriesArray = (FixedArray<TSA_Query>[]) new FixedArray<?>[explicitQueriesListsArray.length];
		for(int type = 0; type < explicitQueriesArray.length; ++type){
			explicitQueriesArray[type] = new FixedArray<TSA_Query>(explicitQueriesListsArray[type]);
		}
		FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>> explicitQueries = new FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>>(DTSA_QueryType.class,explicitQueriesArray);
		return explicitQueries;
	}
	
	public FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>> differentiateQueries(final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers, final FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>> queries) throws TSA_Exception{	
		@SuppressWarnings("unchecked")
		ArrayList<DTSA_Query>[] differentiatedQueriesListsArray = (ArrayList<DTSA_Query>[]) new ArrayList<?>[DTSA_QueryType.values().length];
		for(int type = 0; type < differentiatedQueriesListsArray.length; ++type){
			differentiatedQueriesListsArray[type] = new ArrayList<DTSA_Query>();
		}		
		FixedEnumMap<DTSA_QueryType, ArrayList<DTSA_Query>> differentiatedQueriesLists = new FixedEnumMap<DTSA_QueryType, ArrayList<DTSA_Query>>(DTSA_QueryType.class, differentiatedQueriesListsArray);
		
		for(DTSA_QueryType type: queries.keys()){
			ArrayList<DTSA_Query> differentiatedQueriesList = differentiatedQueriesLists.get(type);
			for(TSA_Query query: queries.get(type)){
				differentiatedQueriesList.add(this.differentiate(query, type, streamBuffers));
			}
		}
		
		@SuppressWarnings("unchecked")
		FixedArray<DTSA_Query>[] differentiatedQueriesArray = (FixedArray<DTSA_Query>[]) new FixedArray<?>[differentiatedQueriesListsArray.length];
		for(int type = 0; type < differentiatedQueriesArray.length; ++type){
			differentiatedQueriesArray[type] = new FixedArray<DTSA_Query>(differentiatedQueriesListsArray[type]);
		}
		FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>> differentiatedQueries = new FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>>(DTSA_QueryType.class,differentiatedQueriesArray);
		return differentiatedQueries;
	}
	
	public DTSA_BufferType outputType(final DTSA_QueryType query_type){
		switch (query_type) {
		case STATIC: return DTSA_BufferType.WORKING;
		case INIT: return DTSA_BufferType.WORKING;
		case INPUT: return DTSA_BufferType.WORKING;
		case WORKING: return DTSA_BufferType.WORKING;
		case OUTPUT: return DTSA_BufferType.OUTPUT;
		case LOG: return DTSA_BufferType.LOG;
		default:
			throw new IllegalArgumentException();
		}
	}
	
	protected ConditionFormula outputMinTimeDefinitionFormula(final StreamBound outputMinTime, final ConditionFormula minTimeFormula, final ArithmeticAnalyser temporalAnalyser, final FixedArrayNavigableMap<String, StreamBound> referencedAttributes, final FixedArrayNavigableMap<String, StreamBound> referencedVariables){
		assert !minTimeFormula.containsNegation();
		String outputAttributeName = outputMinTime.name;
		
		final BasicTypeID type = outputMinTime.type.ID;
		Filter<ConditionFormula> typeFilter = new AbstractFilter<ConditionFormula>() {
			
			@Override
			public boolean accepts(final ConditionFormula element) {
				assert element instanceof AtomicStreamBoundFormula;
				AtomicStreamBoundFormula sbf = (AtomicStreamBoundFormula) element;				
				assert sbf.streamBound.type.equals(sbf.atomicTerm.type);
				return sbf.streamBound.type.ID == type;
			}
			
		};
		
		final ConditionFormula typeFilteredMinTimeFormula = not(minTimeFormula).focus(typeFilter).subformula().flaten();
		
		if(typeFilteredMinTimeFormula.ID == ConditionFormulaID.TOP){
			return Top;
		}
		else{		
			for(ConjunctiveConditionFormula possibleStreamBounds: typeFilteredMinTimeFormula.DNFiteration()){
				boolean isStreamBound = true;
				ArrayList<ConditionFormula> outputMinTimeDefinitionFormulaPart = new ArrayList<ConditionFormula>();
				for(ConditionFormula atom: possibleStreamBounds.subformulas()){
					assert atom instanceof AtomicStreamBoundFormula;
					AtomicStreamBoundFormula streamBound = (AtomicStreamBoundFormula) atom;
					boolean isAttribute = (streamBound.atomicTerm instanceof Attribute<?>);
					String streamBoundAttribute;
					StreamBound streamBoundMinTime;
					if(isAttribute){
						streamBoundAttribute = ((Attribute<?>) streamBound.atomicTerm).name;
						assert referencedAttributes.has(streamBoundAttribute);
						streamBoundMinTime = referencedAttributes.get(streamBoundAttribute);
					}
					else{
						assert streamBound.atomicTerm instanceof Variable<?>;
						streamBoundAttribute = ((Variable<?>) streamBound.atomicTerm).name;
						assert referencedVariables.has(streamBoundAttribute);
						streamBoundMinTime = referencedVariables.get(streamBoundAttribute);
					}
					
					long distance = temporalAnalyser.distance(streamBoundAttribute, isAttribute, outputAttributeName, true);
					if(distance < ArithmeticAnalyser.PLUS_INFTY){
						long constant = ArithmeticAnalyser.constant(distance);
						if(constant == 0){
							outputMinTimeDefinitionFormulaPart.add(new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS_EQUAL, outputMinTime, streamBoundMinTime));
						}
						else{
							outputMinTimeDefinitionFormulaPart.add(new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS_EQUAL, outputMinTime, apply(MINUS, TIMESTAMP.type, streamBoundMinTime, new DurationValue(constant))));
						}
					}
					else{
						isStreamBound = false;
						break;
					}
				}
				if(isStreamBound){
					//FIXME alternatives! now just the first possible one
					return and(outputMinTimeDefinitionFormulaPart).flaten();
				}
			}
			return Bottom;
		}
	}
	
	@SuppressWarnings("unchecked")
	protected static Term<BasicType> outputMinTimeDefinition(final ConditionFormula outputMinTimeDefinitionFormula){
		assert outputMinTimeDefinitionFormula.isFlatened;
		switch(outputMinTimeDefinitionFormula.ID){
			case ATOMIC: assert outputMinTimeDefinitionFormula instanceof BinaryComparissonFormula;
				BinaryComparissonFormula bcf = (BinaryComparissonFormula) outputMinTimeDefinitionFormula;
				assert bcf.operator == BinaryComparissonOperatorID.LESS_EQUAL;
				assert bcf.secondTerm().isTyped & ((Term<Type>)bcf.secondTerm()).type.ID == TypeID.BASIC;
				return (Term<BasicType>) bcf.secondTerm();
			case CONJUNCTIVE:{
				ConjunctiveConditionFormula conj = (ConjunctiveConditionFormula) outputMinTimeDefinitionFormula;
				Term<BasicType>[] subDefinitions = (Term<BasicType>[])new Term<?>[conj.subformulas().size()];
				for(int subFormula = 0; subFormula < subDefinitions.length; ++subFormula){
					subDefinitions[subFormula] = outputMinTimeDefinition(conj.subformulas().get(subFormula));
				}
				assert subDefinitions.length > 0;
				BasicType type = subDefinitions[0].type;
				return apply(LEAST, type, subDefinitions);
			}
			case DISJUNCTIVE:{
				DisjunctiveConditionFormula disj = (DisjunctiveConditionFormula) outputMinTimeDefinitionFormula;
				Term<BasicType>[] subDefinitions = (Term<BasicType>[])new Term<?>[disj.subformulas().size()];
				for(int subFormula = 0; subFormula < subDefinitions.length; ++subFormula){
					subDefinitions[subFormula] = outputMinTimeDefinition(disj.subformulas().get(subFormula));
				}
				assert subDefinitions.length > 0;
				BasicType type = subDefinitions[0].type;
				return apply(GREATEST, type, subDefinitions);
			}
			case NEGATED: throw new Error("Unexpected Case" + outputMinTimeDefinitionFormula);
			case BOTTOM: return Bottom;
			case TOP: return Top;
		}
		throw new Error("Unexpected Case: " + outputMinTimeDefinitionFormula);
	}
	
	protected Term<BasicType> outputMinTimeClockIncreaseDefinition(final StreamBound outputMinTime, final ArithmeticAnalyser temporalAnalyser, final FixedArrayNavigableMap<String, ClockFunction> clockAttributes, final FixedArrayNavigableMap<String, ClockFunction> clockVariables){
		String outputAttributeName = outputMinTime.name;
		ArrayList<Term<BasicType>> increasingClocks = new ArrayList<Term<BasicType>>();
		for(String clockAttribute: clockAttributes.keys()){
			long distance = temporalAnalyser.distance(clockAttribute, true, outputAttributeName, true);
			if(distance < ArithmeticAnalyser.PLUS_INFTY){
				long constant = ArithmeticAnalyser.constant(distance);
				ClockFunction clock = clockAttributes.get(clockAttribute);
				if(constant == 0){
					increasingClocks.add(clock);
				}
				else{
					increasingClocks.add(apply(MINUS, clock.type, clock, new DurationValue(constant)));
				}
			}
		}
		for(String clockVariable: clockVariables.keys()){
			long distance = temporalAnalyser.distance(clockVariable, false, outputAttributeName, true);
			if(distance < ArithmeticAnalyser.PLUS_INFTY){
				long constant = ArithmeticAnalyser.constant(distance);
				ClockFunction clock = clockVariables.get(clockVariable);
				if(constant == 0){
					increasingClocks.add(clock);
				}
				else{
					increasingClocks.add(apply(MINUS, clock.type, clock, new DurationValue(constant)));
				}
			}
		}
		
		if(increasingClocks.size() == 0){
			return Bottom;
		}
		else if(increasingClocks.size() == 1){
			return increasingClocks.first();
		}
		else{
			return apply(GREATEST, increasingClocks.first().type, increasingClocks.toArray(new Term<?>[0]));	
		}
	}
	
	protected final static Filter<ConditionFormula> atomicFormulaFilter = new AbstractFilter<ConditionFormula>() {
		@Override
		public boolean accepts(final ConditionFormula element) {
			return element.ID == ConditionFormulaID.ATOMIC;
		}
	};
	
	public DTSA_Query differentiate(final TSA_Query tsa_query, final DTSA_QueryType type, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers) throws TSA_Exception{
		try{
			assert (tsa_query.isStatic & (type == DTSA_QueryType.STATIC | type == DTSA_QueryType.INIT)) | (!tsa_query.isStatic & type != DTSA_QueryType.STATIC & type != DTSA_QueryType.INIT); 
			
			final DTSA_BufferType output_type = this.outputType(type);
			
			final DTSA_StreamBuffer output = streamBuffers.get(output_type).get(tsa_query.output_full_name);
			
			final boolean isMonotonic = tsa_query.isMonotonic();
			
			DTSA_Expression differenciatedExpression = this.differentiate(tsa_query.expression, type, isMonotonic, streamBuffers);
			
			ConditionFormula streamBoundFormula = this.streamBoundFormula(differenciatedExpression.schema().relationConstraints(), tsa_query.origin);
			assert tsa_query.isStatic | streamBoundFormula.ID != ConditionFormulaID.TOP: "Top is only allowed for static relations";
			
			final FixedArrayNavigableMap<String, StreamBound> referencedIdentifierAttributes;
			final FixedArrayNavigableMap<String, StreamBound> referencedTimestampAttributes;
			final FixedArrayNavigableMap<String, StreamBound> referencedIdentifierVariables;
			final FixedArrayNavigableMap<String, StreamBound> referencedTimestampVariables;
			{
				AbstractCollection<ConditionFormula> streamBoundFormulas = streamBoundFormula.collectFormulas(atomicFormulaFilter);
				
				ArrayList<String> referencedIdentifierAttributeNamesList = new ArrayList<String>();
				ArrayList<String> referencedTimestampAttributeNamesList = new ArrayList<String>();
				ArrayList<StreamBound> referencedIdentifierAttributeMinTimeList = new ArrayList<StreamBound>();
				ArrayList<StreamBound> referencedTimestampAttributeMinTimeList = new ArrayList<StreamBound>();
				ArrayList<String> referencedIdentifierVariableNamesList = new ArrayList<String>();
				ArrayList<String> referencedTimestampVariableNamesList = new ArrayList<String>();
				ArrayList<StreamBound> referencedIdentifierVariableMinTimeList = new ArrayList<StreamBound>();
				ArrayList<StreamBound> referencedTimestampVariableMinTimeList = new ArrayList<StreamBound>();
				
				for(ConditionFormula formula: streamBoundFormulas){
					AtomicStreamBoundFormula streamBoundFormla = (AtomicStreamBoundFormula) formula;
					StreamBound streamBound = streamBoundFormla.streamBound;
					if(streamBoundFormla.atomicTerm instanceof BasicTypeAttribute){
						BasicTypeAttribute attribute = (BasicTypeAttribute) streamBoundFormla.atomicTerm;
						if(attribute.ID == IDENTIFIER){
							referencedIdentifierAttributeNamesList.add(attribute.name);
							referencedIdentifierAttributeMinTimeList.add(streamBound);
						}
						else{assert attribute.ID == TIMESTAMP;
							referencedTimestampAttributeNamesList.add(attribute.name);
							referencedTimestampAttributeMinTimeList.add(streamBound);
						}
					}
					else{assert streamBoundFormla.atomicTerm instanceof BasicTypeVariable;					
						BasicTypeVariable variable = (BasicTypeVariable) streamBoundFormla.atomicTerm;
						if(variable.ID == IDENTIFIER){
							referencedIdentifierVariableNamesList.add(variable.name);
							referencedIdentifierVariableMinTimeList.add(streamBound);
						}
						else{assert variable.ID == TIMESTAMP;
							referencedTimestampVariableNamesList.add(variable.name);
							referencedTimestampVariableMinTimeList.add(streamBound);
						}
					}
				}
				
				referencedIdentifierAttributes = new FixedArrayNavigableMap<String, StreamBound>(referencedIdentifierAttributeNamesList.toArray(new String[0]), referencedIdentifierAttributeMinTimeList.toArray(new StreamBound[0]));
				referencedTimestampAttributes = new FixedArrayNavigableMap<String, StreamBound>(referencedTimestampAttributeNamesList.toArray(new String[0]), referencedTimestampAttributeMinTimeList.toArray(new StreamBound[0]));
				referencedIdentifierVariables = new FixedArrayNavigableMap<String, StreamBound>(referencedIdentifierVariableNamesList.toArray(new String[0]), referencedIdentifierVariableMinTimeList.toArray(new StreamBound[0]));
				referencedTimestampVariables = new FixedArrayNavigableMap<String, StreamBound>(referencedTimestampVariableNamesList.toArray(new String[0]), referencedTimestampVariableMinTimeList.toArray(new StreamBound[0]));
			}
			
			Filter<ConditionFormula> clockFunctionsFilter = new AbstractFilter<ConditionFormula>() {
				
				@Override
				public boolean accepts(final ConditionFormula element) {
					if(element instanceof BinaryComparissonFormula){
						BinaryComparissonFormula bcf = (BinaryComparissonFormula) element;
						if(bcf.secondTerm() instanceof ClockFunction){
							return true;
						}
					}
					return false;
				}
				
			};
			
			final FixedArrayNavigableMap<String, ClockFunction> clockIdentifierAttributes;
			final FixedArrayNavigableMap<String, ClockFunction> clockTimestampAttributes;
			final FixedArrayNavigableMap<String, ClockFunction> clockIdentifierVariables;
			final FixedArrayNavigableMap<String, ClockFunction> clockTimestampVariables;
			{
				AbstractCollection<ConditionFormula> clockEquations = and(differenciatedExpression.schema().tupleConstraints()).collectFormulas(clockFunctionsFilter);
				
				ArrayList<String> clockIdentifierAttributeNamesList = new ArrayList<String>();
				ArrayList<String> clockTimestampAttributeNamesList = new ArrayList<String>();
				ArrayList<ClockFunction> clockIdentifierAttributeClockFunctionList = new ArrayList<ClockFunction>();
				ArrayList<ClockFunction> clockTimestampAttributeClockFunctionList = new ArrayList<ClockFunction>();
				ArrayList<String> clockIdentifierVariableNamesList = new ArrayList<String>();
				ArrayList<String> clockTimestampVariableNamesList = new ArrayList<String>();
				ArrayList<ClockFunction> clockIdentifierVariableClockFunctionList = new ArrayList<ClockFunction>();
				ArrayList<ClockFunction> clockTimestampVariableClockFunctionList = new ArrayList<ClockFunction>();
				
				for(ConditionFormula minTimeInequationFormula: clockEquations){
					BinaryComparissonFormula clockEquation = (BinaryComparissonFormula) minTimeInequationFormula;
					ClockFunction clock = (ClockFunction) clockEquation.secondTerm();
					if(clockEquation.firstTerm() instanceof BasicTypeAttribute){
						BasicTypeAttribute attribute = (BasicTypeAttribute) clockEquation.firstTerm();
						if(attribute.ID == IDENTIFIER){
							clockIdentifierAttributeNamesList.add(attribute.name);
							clockIdentifierAttributeClockFunctionList.add(clock);
						}
						else{assert attribute.ID == TIMESTAMP;
							clockTimestampAttributeNamesList.add(attribute.name);
							clockTimestampAttributeClockFunctionList.add(clock);
						}
					}
					else{assert clockEquation.firstTerm() instanceof BasicTypeVariable;					
						BasicTypeVariable variable = (BasicTypeVariable) clockEquation.firstTerm();
						if(variable.ID == IDENTIFIER){
							clockIdentifierVariableNamesList.add(variable.name);
							clockIdentifierVariableClockFunctionList.add(clock);
						}
						else{assert variable.ID == TIMESTAMP;
							clockTimestampVariableNamesList.add(variable.name);
							clockTimestampVariableClockFunctionList.add(clock);
						}
					}
				}
				
				clockIdentifierAttributes = new FixedArrayNavigableMap<String, ClockFunction>(clockIdentifierAttributeNamesList.toArray(new String[0]), clockIdentifierAttributeClockFunctionList.toArray(new ClockFunction[0]));
				clockTimestampAttributes = new FixedArrayNavigableMap<String, ClockFunction>(clockTimestampAttributeNamesList.toArray(new String[0]), clockTimestampAttributeClockFunctionList.toArray(new ClockFunction[0]));
				clockIdentifierVariables = new FixedArrayNavigableMap<String, ClockFunction>(clockIdentifierVariableNamesList.toArray(new String[0]), clockIdentifierVariableClockFunctionList.toArray(new ClockFunction[0]));
				clockTimestampVariables = new FixedArrayNavigableMap<String, ClockFunction>(clockTimestampVariableNamesList.toArray(new String[0]), clockTimestampVariableClockFunctionList.toArray(new ClockFunction[0]));
			}
			
			final FixedArrayNavigableMap<String, StreamBound> outputIdentifierAttributes;
			final FixedArrayNavigableMap<String, StreamBound> outputTimestampAttributes;
			final FixedArrayNavigableMap<String, StreamBound> outputMinTimes;
			
			{
				outputMinTimes = output.streamBounds;
				ArrayList<String> outputIdentifierAttributeNamesList = new ArrayList<String>();
				ArrayList<String> outputTimestampAttributeNamesList = new ArrayList<String>();
				ArrayList<StreamBound> outputIdentifierAttributeMinTimeList = new ArrayList<StreamBound>();
				ArrayList<StreamBound> outputTimestampAttributeMinTimeList = new ArrayList<StreamBound>();
				
				for(int outputMinTime = 0; outputMinTime < outputMinTimes.size(); ++outputMinTime){
					StreamBound minTime = outputMinTimes.values().get(outputMinTime);
					BasicTypeAttribute correspondingOutputAttribute = (BasicTypeAttribute) output.schema.getAttribute(outputMinTimes.keys().get(outputMinTime));
					if(correspondingOutputAttribute.ID == IDENTIFIER){
						outputIdentifierAttributeNamesList.add(correspondingOutputAttribute.name);
						outputIdentifierAttributeMinTimeList.add(minTime);
					}
					else{assert correspondingOutputAttribute.ID == TIMESTAMP;
						outputTimestampAttributeNamesList.add(correspondingOutputAttribute.name);
						outputTimestampAttributeMinTimeList.add(minTime);					
					}
				}
				
				outputIdentifierAttributes = new FixedArrayNavigableMap<String, StreamBound>(new FixedArrayNavigableSet<String>(false, outputIdentifierAttributeNamesList.toArray(new String[0])),new FixedArray<StreamBound>(false, outputIdentifierAttributeMinTimeList.toArray(new StreamBound[0])));
				outputTimestampAttributes = new FixedArrayNavigableMap<String, StreamBound>(new FixedArrayNavigableSet<String>(false, outputTimestampAttributeNamesList.toArray(new String[0])),new FixedArray<StreamBound>(false, outputTimestampAttributeMinTimeList.toArray(new StreamBound[0])));
			}
			
			@SuppressWarnings("unchecked")
			final ArithmeticAnalyser identifierAnalyser = this.identifierAnalyserFactory.make(and(differenciatedExpression.schema().tupleConstraints()), FixedArrayNavigableSet.merge(referencedIdentifierAttributes.keys(), clockIdentifierAttributes.keys(), outputIdentifierAttributes.keys()), FixedArrayNavigableSet.merge(referencedIdentifierVariables.keys(), clockIdentifierVariables.keys()));
			
			ConditionFormula[] outputIdentifierMinTimesDefinitionFormulas = new ConditionFormula[outputIdentifierAttributes.size()];
			for(int outputMinTimeIndex = 0; outputMinTimeIndex < outputIdentifierMinTimesDefinitionFormulas.length; ++outputMinTimeIndex){
				StreamBound outputMinTime = outputIdentifierAttributes.values().get(outputMinTimeIndex);
				outputIdentifierMinTimesDefinitionFormulas[outputMinTimeIndex] = this.outputMinTimeDefinitionFormula(outputMinTime, streamBoundFormula, identifierAnalyser, referencedIdentifierAttributes, referencedIdentifierVariables);
			}
			
			@SuppressWarnings("unchecked")
			final ArithmeticAnalyser temporalAnalyser = this.temporalAnalyserFactory.make(and(differenciatedExpression.schema().tupleConstraints()), FixedArrayNavigableSet.merge(referencedTimestampAttributes.keys(), clockTimestampAttributes.keys(), outputTimestampAttributes.keys()), FixedArrayNavigableSet.merge(referencedTimestampVariables.keys(), clockTimestampVariables.keys()));
			
			ConditionFormula[] outputTimestampMinTimesDefinitionFormulas = new ConditionFormula[outputTimestampAttributes.size()];
			for(int outputMinTimeIndex = 0; outputMinTimeIndex < outputTimestampMinTimesDefinitionFormulas.length; ++outputMinTimeIndex){
				StreamBound outputMinTime = outputTimestampAttributes.values().get(outputMinTimeIndex);
				outputTimestampMinTimesDefinitionFormulas[outputMinTimeIndex] = this.outputMinTimeDefinitionFormula(outputMinTime, streamBoundFormula, temporalAnalyser, referencedTimestampAttributes, referencedTimestampVariables);
			}
			
			final FixedArrayNavigableMap<String, Term<BasicType>> outputIdentifierMinTimeDefinitions;
			{
					@SuppressWarnings("unchecked")
					Term<BasicType>[] definitions = (Term<BasicType>[])new Term<?>[outputIdentifierMinTimesDefinitionFormulas.length];
					for(int outputMinTime = 0; outputMinTime < outputIdentifierMinTimesDefinitionFormulas.length; ++outputMinTime){
						definitions[outputMinTime] = outputMinTimeDefinition(outputIdentifierMinTimesDefinitionFormulas[outputMinTime]);
					}
					outputIdentifierMinTimeDefinitions = new FixedArrayNavigableMap<String, Term<BasicType>>(outputIdentifierAttributes.keys(), new FixedArray<Term<BasicType>>(false, definitions));
			}
			
			final FixedArrayNavigableMap<String, Term<BasicType>> outputTimestampMinTimeDefinitions;
			{
					@SuppressWarnings("unchecked")
					Term<BasicType>[] definitions = (Term<BasicType>[])new Term<?>[outputTimestampMinTimesDefinitionFormulas.length];
					for(int outputMinTime = 0; outputMinTime < outputTimestampMinTimesDefinitionFormulas.length; ++outputMinTime){
						definitions[outputMinTime] = outputMinTimeDefinition(outputTimestampMinTimesDefinitionFormulas[outputMinTime]);
					}
					outputTimestampMinTimeDefinitions = new FixedArrayNavigableMap<String, Term<BasicType>>(outputTimestampAttributes.keys(), new FixedArray<Term<BasicType>>(false, definitions));
			}
			
			final FixedArrayNavigableMap<String, Term<BasicType>> outputMinTimeDefinitions = FixedArrayNavigableMap.merge(outputIdentifierMinTimeDefinitions, outputTimestampMinTimeDefinitions);
			
			/*
			@SuppressWarnings("unchecked")
			ArrayList<Term>[] outputMinTimePreDefinitions = new ArrayList[outputMinTimeNames.size()];
			for(int outputMinTime = 0; outputMinTime < outputMinTimePreDefinitions.length; ++outputMinTime){
				outputMinTimePreDefinitions[outputMinTime] = new ArrayList<Term>();
			}*/
			
			final FixedArrayNavigableMap<String, Term<BasicType>> outputIdentifierMinTimeClockIncreaseDefinitions;
			{
				@SuppressWarnings("unchecked")
				Term<BasicType>[] definitions = (Term<BasicType>[])new Term<?>[outputIdentifierMinTimesDefinitionFormulas.length];
				for(int outputMinTimeIndex = 0; outputMinTimeIndex < outputIdentifierMinTimesDefinitionFormulas.length; ++outputMinTimeIndex){
					StreamBound outputMinTime = outputIdentifierAttributes.values().get(outputMinTimeIndex);
					definitions[outputMinTimeIndex] = this.outputMinTimeClockIncreaseDefinition(outputMinTime, identifierAnalyser, clockIdentifierAttributes, clockIdentifierVariables);	
				}
				outputIdentifierMinTimeClockIncreaseDefinitions = new FixedArrayNavigableMap<String, Term<BasicType>>(outputIdentifierAttributes.keys(), new FixedArray<Term<BasicType>>(false, definitions));
			}
			final FixedArrayNavigableMap<String, Term<BasicType>> outputTimestampMinTimeClockIncreaseDefinitions;
			{
				@SuppressWarnings("unchecked")
				Term<BasicType>[] definitions = (Term<BasicType>[])new Term<?>[outputTimestampMinTimesDefinitionFormulas.length];
				for(int outputMinTimeIndex = 0; outputMinTimeIndex < outputTimestampMinTimesDefinitionFormulas.length; ++outputMinTimeIndex){
					StreamBound outputMinTime = outputTimestampAttributes.values().get(outputMinTimeIndex);
					definitions[outputMinTimeIndex] = this.outputMinTimeClockIncreaseDefinition(outputMinTime, temporalAnalyser, clockTimestampAttributes, clockTimestampVariables);	
				}
				outputTimestampMinTimeClockIncreaseDefinitions = new FixedArrayNavigableMap<String, Term<BasicType>>(outputTimestampAttributes.keys(), new FixedArray<Term<BasicType>>(false, definitions));
			}
			
			final FixedArrayNavigableMap<String, Term<BasicType>> outputMinTimeClockIncreaseDefinitions = FixedArrayNavigableMap.merge(outputIdentifierMinTimeClockIncreaseDefinitions, outputTimestampMinTimeClockIncreaseDefinitions);
			
			boolean hasNonClockOutputStreamBound = false; 
			//check definitions of output minTimes
			for(int outputMinTimeIndex = 0; outputMinTimeIndex < outputMinTimeDefinitions.size(); ++outputMinTimeIndex){
				Term<BasicType> definition = outputMinTimeDefinitions.values().get(outputMinTimeIndex);
				Term<BasicType> clockIncreaseDefinition = outputMinTimeClockIncreaseDefinitions.values().get(outputMinTimeIndex);
	
				if(definition instanceof ConditionFormula && ((ConditionFormula)definition).ID == ConditionFormulaID.BOTTOM){
					if(clockIncreaseDefinition instanceof ConditionFormula && ((ConditionFormula)clockIncreaseDefinition).ID == ConditionFormulaID.BOTTOM){
						throw new TSA_Exception(tsa_query.origin , "Output stream bound " + outputMinTimeDefinitions.keys().get(outputMinTimeIndex) + " is invalid.");
					}
				}
				else{
					hasNonClockOutputStreamBound = true;
				}
			}
			
			DTSA_Expression incrementExpression;
			
			if(type == DTSA_QueryType.STATIC | type == DTSA_QueryType.INIT){
				incrementExpression = differenciatedExpression;
			}
			else if(isMonotonic){		
				
				boolean hasIsNewAttribute = differenciatedExpression.schema().hasAttribute(isNewAttributeName);
				ConditionFormula isNewFormula = this.isNewFormula(hasIsNewAttribute, referencedIdentifierAttributes);
				assert isNewFormula.ID != ConditionFormulaID.TOP; 
				final TSA_SelectionOperator isNewSelection = new TSA_SelectionOperator(tsa_query.origin, isNewFormula);
	
				if(hasIsNewAttribute){
					incrementExpression =  new DTSA_CompositeExpression(new TSA_ProjectionOperator(tsa_query.origin, tsa_query.expression.schema().basicAttributes.keys()), new DTSA_CompositeExpression(isNewSelection, differenciatedExpression));
				}
				else{
					incrementExpression = new DTSA_CompositeExpression(isNewSelection, differenciatedExpression);
				}
				
	
			}
			else{
				assert type != DTSA_QueryType.INIT;
				assert type != DTSA_QueryType.INPUT;
				
				if(!tsa_query.isStatic & !hasNonClockOutputStreamBound){
					//at least one outputMinTimeDefinition must be different from Bottom
					throw new TSA_Exception(tsa_query.origin, "The temporal relations are too weak to derive an incremental evaluation expression for the non-static query.");
					//TODO better Exception
				}
				assert !tsa_query.isStatic | !hasNonClockOutputStreamBound;//a static query may only derive stream bounds out of clocks
				
				ArrayList<ConditionFormula> incrementLowerBoundConditions = new ArrayList<ConditionFormula>();
				ArrayList<ConditionFormula> incrementUpperBoundConditions = new ArrayList<ConditionFormula>();
				
				for(int outputMinTimeIndex = 0; outputMinTimeIndex < outputMinTimeDefinitions.size(); ++outputMinTimeIndex){
					Term<BasicType> definition = outputMinTimeDefinitions.values().get(outputMinTimeIndex);
					if(!(definition instanceof ConditionFormula && ((ConditionFormula)definition).ID == ConditionFormulaID.BOTTOM)){
						StreamBound minTime = outputMinTimes.values().get(outputMinTimeIndex);
						BasicTypeAttribute correspondingAttribute = (BasicTypeAttribute) output.schema.getAttribute(minTime.name);
						
						Parameter lowerBound = new Parameter("output.last:" + minTime.qualified_name, minTime.type);
						Parameter upperBound = new Parameter("output.now:" + minTime.qualified_name, minTime.type);
						
						incrementUpperBoundConditions.add(new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS_EQUAL, correspondingAttribute, upperBound));
						incrementLowerBoundConditions.add(new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS, lowerBound, correspondingAttribute));
					}
				}			
				
				TSA_SelectionOperator incrementSelection = new TSA_SelectionOperator(tsa_query.origin, and(incrementLowerBoundConditions).flaten(), or(incrementUpperBoundConditions).flaten());
	
				incrementExpression = new DTSA_CompositeExpression(incrementSelection, differenciatedExpression);
			}
	
			return new DTSA_Query(tsa_query.prog_package, incrementExpression, output, type, outputMinTimeDefinitions, outputMinTimeClockIncreaseDefinitions, tsa_query.description, tsa_query.origin);
		}
		catch (TSA_Exception e) {
			throw e;
		}
		catch (RuntimeException e) {
			throw new TSA_Exception(tsa_query.origin, e);
		}
	}
	
	public boolean isValidBufferStreamBoundFormula(final ConditionFormula streamBoundFormula){
		if(streamBoundFormula.ID == ConditionFormulaID.TOP){
			return true;
		}
		if(streamBoundFormula.ID == ConditionFormulaID.ATOMIC){
			return true;
		}
		if(streamBoundFormula.ID == ConditionFormulaID.DISJUNCTIVE){
			DisjunctiveConditionFormula disjunction = (DisjunctiveConditionFormula) streamBoundFormula;
			for(ConditionFormula subFormula: disjunction.subformulas()){
				if(subFormula.ID != ConditionFormulaID.ATOMIC){
					return false;
				}
			}
			return true;
		}
		return false;
	}
	
	public void checkBufferStreamBoundFormula(final ConditionFormula streamBoundFormula, final FixedCollection<String> origin) throws TSA_Exception{
		if(!this.isValidBufferStreamBoundFormula(streamBoundFormula)){
			throw new TSA_Exception(origin, "Invalid stream bound formula for stream buffer.");
		}
	}
	
	public ConditionFormula streamBoundFormulaFromConstraints(final String buffer_full_name, final FixedArray<ConditionFormula> relationConstraints){
		ConditionFormula streamBoundConstraints = and(relationConstraints).focus(streamBoundConstraintFilter).flaten();
		if(streamBoundConstraints.ID == ConditionFormulaID.TOP){
			return Top;
		}
		else{
			Adapter<AtomicConditionFormula, ConditionFormula> streamBoundAdapter = new Adapter<AtomicConditionFormula, ConditionFormula>() {
				
				@Override
				public ConditionFormula adapt(final AtomicConditionFormula source) {
					if (source instanceof StreamBoundConstraint) {
						StreamBoundConstraint streamBoundConstraint = (StreamBoundConstraint) source;
						assert streamBoundConstraint.atomicTerm instanceof BasicTypeAttribute;
						return not(new AtomicStreamBoundFormula(buffer_full_name, (BasicTypeAttribute) streamBoundConstraint.atomicTerm));
					}
					return source;
				}
				
			};
			return not(streamBoundConstraints.replaceAtomicFormula(streamBoundAdapter)).flaten();
		}
	}
	
	public ConditionFormula streamBoundFormula(final FixedArray<ConditionFormula> relationConstraints, final FixedCollection<String> origin) throws TSA_Exception{
		ConditionFormula streamBoundFormula = and(relationConstraints).focus(streamBoundFilter).flaten();
		if(streamBoundFormula.containsNegation()){
			throw new TSA_Exception(origin, "Stream bound formulas must not contain negation.");
		}
		return streamBoundFormula;
	}
	
	public DTSA_StreamBuffer differentiate(final String full_name, final String description, final DTSA_BufferType type, final TemporalStreamSchema schema, final FixedCollection<String> origin) throws TSA_Exception{		
		try{			
			@SuppressWarnings("unchecked")
			FixedArray<ConditionFormula> relationConstraints = FixedArray.concat(
					and(schema.relationConstraints()).focus(AbstractFilter.notFilter(AbstractFilter.orFilter(streamBoundConstraintFilter, DefaultDefinition.implicitAttributeValuesSpecifications))).subformulas(),
					new FixedArray<ConditionFormula>(this.streamBoundFormulaFromConstraints(full_name, schema.relationConstraints()))
				);
			
			TemporalStreamSchema differentiatedSchema = TemporalStreamSchema.create(schema.ID, schema.attributes, schema.tupleConstraints(), relationConstraints, false);
			
			ConditionFormula streamBoundFormula = this.streamBoundFormula(differentiatedSchema.relationConstraints(), origin);
			this.checkBufferStreamBoundFormula(streamBoundFormula, origin);
			
			this.checkBufferStreamBoundFormula(streamBoundFormula, origin);
			
			FixedArrayNavigableSet<String> streamBounds = streamBoundFormula.AttributeNames();
			return new DTSA_StreamBuffer(full_name, description, type, differentiatedSchema, schema.implicitAttributeSpecifications, streamBounds, origin);
		}
		catch (TSA_Exception e) {
			throw e;
		}
		catch (RuntimeException e) {
			throw new TSA_Exception(origin, e);
		}
	}
	
	public DTSA_Expression differentiate(final TSA_Expression tsa_expression, final DTSA_QueryType type, final boolean monotonic, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers) throws TSA_Exception{
		if(tsa_expression instanceof TSA_TemporalStreamExpression){
			TSA_TemporalStreamExpression tsa_temporalStreamExpression = (TSA_TemporalStreamExpression) tsa_expression;
			final DTSA_StreamBuffer dtsa_streamBuffer = streamBuffers.get(tsa_temporalStreamExpression.type).get(tsa_temporalStreamExpression.full_name);		
			
			if(dtsa_streamBuffer == null){
				throw new TSA_Exception(tsa_temporalStreamExpression.origin, "The expression refers to an non-existing stream: " + tsa_temporalStreamExpression.type + ":" + tsa_temporalStreamExpression.full_name);
			}
			else{
				return dtsa_streamBuffer;
			}
		}
		else{assert tsa_expression instanceof TSA_CompositeExpression;
			return this.differentiate((TSA_CompositeExpression) tsa_expression, type, monotonic, streamBuffers);
		}
				
	}
	
	protected ConditionFormula isNewFormula(final boolean hasIsNewAttribute, final FixedArrayNavigableMap<String, StreamBound> idAttributes){
		ConditionFormula isNewFormula = Top;
		if(idAttributes.size() > 0){
			BinaryComparissonFormula[] lowerBoundConditions = new BinaryComparissonFormula[idAttributes.size()];
			BinaryComparissonFormula[] upperBoundConditions = new BinaryComparissonFormula[lowerBoundConditions.length];
			for (int discardedIdAttribute = 0; discardedIdAttribute < lowerBoundConditions.length; ++discardedIdAttribute) {
				IdentifierAttribute attribute = IdentifierAttribute.create(idAttributes.keys().get(discardedIdAttribute));
				StreamBound minTime = idAttributes.values().get(discardedIdAttribute);
				
				Parameter lowerBound = new Parameter("input.last:" + minTime.qualified_name, minTime.type);
				Parameter upperBound = new Parameter("input.now:" + minTime.qualified_name, minTime.type);
				lowerBoundConditions[discardedIdAttribute] = new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS, lowerBound, attribute); 
				upperBoundConditions[discardedIdAttribute] = new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS_EQUAL, attribute, upperBound);
			}
			isNewFormula = and(new DisjunctiveConditionFormula(lowerBoundConditions), new ConjunctiveConditionFormula(upperBoundConditions));
		}
		if(hasIsNewAttribute){
			isNewFormula = 	and(new BooleanTerm(isNewAttribute) ,isNewFormula);
		}
		return isNewFormula.flaten();
	}
	
	protected ConditionFormula incrementFunctionFormulaOutputSBFFirst(final ArithmeticAnalyser temporalAnalyser, final ConditionFormula inputSBF, final ConditionFormula outputSBF){
		switch (outputSBF.ID) {
		case TOP: return (temporalAnalyser.isConsistent() & inputSBF.ID != TOP ? Bottom : Top);
		case BOTTOM: return Top;
		case ATOMIC:
			if(inputSBF.ID == ConditionFormulaID.ATOMIC){
				if(inputSBF instanceof AtomicStreamBoundFormula & outputSBF instanceof AtomicStreamBoundFormula){
					return this.incrementFunctionFormula(temporalAnalyser, (AtomicStreamBoundFormula) inputSBF, (AtomicStreamBoundFormula) outputSBF);
				}
				else{
					throw new IllegalArgumentException("Stream bound formulas may only contain instances of " + AtomicStreamBoundFormula.class.getSimpleName() + ".");
				}
			}
			else{
				return this.incrementFunctionFormulaInputSBFFirst(temporalAnalyser, inputSBF, outputSBF);
			}
		case CONJUNCTIVE:
			return and(new CollectionAdapter<ConditionFormula, ConditionFormula>(((ConjunctiveConditionFormula)outputSBF).subformulas()) {
				@Override
				public ConditionFormula adapt(final ConditionFormula outputSBFSubFormula) {
					return TSA_DefaultDifferencer.this.incrementFunctionFormulaOutputSBFFirst(temporalAnalyser, inputSBF, outputSBFSubFormula);
				}
			});
		case DISJUNCTIVE:
			return or(new CollectionAdapter<ConditionFormula, ConditionFormula>(((DisjunctiveConditionFormula)outputSBF).subformulas()) {
				@Override
				public ConditionFormula adapt(final ConditionFormula outputSBFSubFormula) {
					return TSA_DefaultDifferencer.this.incrementFunctionFormulaOutputSBFFirst(temporalAnalyser, inputSBF, outputSBFSubFormula);
				}
			});
		case NEGATED:
			throw new IllegalArgumentException("Stream bound formulas may not contain negation.");
		default: throw new Error("Unexpected case \"" + inputSBF.ID.name() + "\"");
		}
	}
	
	protected ConditionFormula incrementFunctionFormulaInputSBFFirst(final ArithmeticAnalyser temporalAnalyser, final ConditionFormula inputSBF, final ConditionFormula outputSBF){
		switch (inputSBF.ID) {
		case TOP: return Top;
		case BOTTOM: return (temporalAnalyser.isConsistent() & outputSBF.ID != BOTTOM ? Bottom : Top);//
		case ATOMIC:
			if(outputSBF.ID == ConditionFormulaID.ATOMIC){
				if(inputSBF instanceof AtomicStreamBoundFormula & outputSBF instanceof AtomicStreamBoundFormula){
					return this.incrementFunctionFormula(temporalAnalyser, (AtomicStreamBoundFormula) inputSBF, (AtomicStreamBoundFormula) outputSBF);
				}
				else{
					throw new IllegalArgumentException("Stream bound formulas may only contain instances of " + AtomicStreamBoundFormula.class.getSimpleName() + ".");
				}
			}
			else{
				return this.incrementFunctionFormulaOutputSBFFirst(temporalAnalyser, inputSBF, outputSBF);
			}
		case CONJUNCTIVE:
			return and(new CollectionAdapter<ConditionFormula, ConditionFormula>(((ConjunctiveConditionFormula)inputSBF).subformulas()) {
				@Override
				public ConditionFormula adapt(final ConditionFormula inputSBFSubFormula) {
					return TSA_DefaultDifferencer.this.incrementFunctionFormulaOutputSBFFirst(temporalAnalyser, inputSBFSubFormula, outputSBF);
				}
			});
		case DISJUNCTIVE:
			return or(new CollectionAdapter<ConditionFormula, ConditionFormula>(((DisjunctiveConditionFormula)inputSBF).subformulas()) {
				@Override
				public ConditionFormula adapt(final ConditionFormula inputSBFSubFormula) {
					return TSA_DefaultDifferencer.this.incrementFunctionFormulaOutputSBFFirst(temporalAnalyser, inputSBFSubFormula, outputSBF);
				}
			});
		case NEGATED:
			throw new IllegalArgumentException("Stream bound formulas may not contain negation.");
		default: throw new Error("Unexpected case \"" + inputSBF.ID.name() + "\"");
		}
	}
	
	protected final ArithmeticRelationFormulaExtractor temporalFormulaExtractor = new ArithmeticRelationFormulaExtractor(TIMESTAMP.type);//FIXME
	
	@SuppressWarnings("unchecked")
	protected ConditionFormula incrementFunctionFormula(final ArithmeticAnalyser temporalAnalyser, final AtomicStreamBoundFormula inputSBF, final AtomicStreamBoundFormula outputSBF){
		if(inputSBF.atomicTerm.isTyped & outputSBF.atomicTerm.isTyped){
			assert inputSBF.atomicTerm.type instanceof BasicType;
			assert outputSBF.atomicTerm instanceof BasicTypeAttribute;
			return this.incrementFunctionFormula(temporalAnalyser, (AtomicTerm<? extends BasicType>) inputSBF.atomicTerm, inputSBF.streamBound, (BasicTypeAttribute) outputSBF.atomicTerm, outputSBF.streamBound);
		}
		else{
			throw new IllegalArgumentException();
		}
	}
	
	protected ConditionFormula incrementFunctionFormula(final ArithmeticAnalyser temporalAnalyser, final AtomicTerm<? extends BasicType> inputAttribute, final StreamBound inputStreamBound, final BasicTypeAttribute outputAttribute, final StreamBound outputStreamBound){
		assert inputAttribute.type.ID == inputStreamBound.type.ID;
		assert outputAttribute.type.ID == outputStreamBound.type.ID;
		assert inputStreamBound.type.ID == outputStreamBound.type.ID;
		boolean isAttribute;
		String inputAttributeName;
		if(inputAttribute.ID == ATTRIBUTE){
			isAttribute = true;
			inputAttributeName = ((BasicTypeAttribute)inputAttribute).name;
		}
		else if(inputAttribute.ID == VARIABLE){
			isAttribute = false;
			inputAttributeName = ((BasicTypeVariable)inputAttribute).name;
		}
		else{
			throw new IllegalArgumentException();
		}
		long distance = temporalAnalyser.distance(inputAttributeName, isAttribute, outputAttribute.name, true);
		if(distance == ArithmeticAnalyser.PLUS_INFTY){
			return Bottom;
		}
		else if(distance == ArithmeticAnalyser.MINUS_INFTY){
			return Top;
		}
		else{
			BasicTypeValue distanceValue = this.temporalFormulaExtractor.value(this.temporalFormulaExtractor.durationType(inputStreamBound.type).ID, -ArithmeticAnalyser.constant(distance) + (ArithmeticAnalyser.minusEpsilon(distance) ? 1 : 0));
			return new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS_EQUAL, outputStreamBound, apply(MINUS, inputStreamBound, distanceValue));
		}
	
		
	}
	
	public DTSA_Expression differentiate(final TSA_CompositeExpression tsa_compositeExpression, final DTSA_QueryType type, final boolean monotonic, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers) throws TSA_Exception{
		try{
			switch (tsa_compositeExpression.operator.ID) {
			case BASIC: 
			{
				switch (((TSA_BasicOperator)tsa_compositeExpression.operator).ID) {
				case CROSSPRODUCT: 
				{				
					assert tsa_compositeExpression.subExpressions.size() > 0;
					boolean containsAnyIsNewAttribute = false;
					ArrayList<BooleanTerm> isNewAttributes = new ArrayList<BooleanTerm>();
					DTSA_Expression[] differentiatedSubExpressions = new DTSA_Expression[tsa_compositeExpression.subExpressions.size()];
					for (int subExpression = 0; subExpression < differentiatedSubExpressions.length; ++subExpression) {
						differentiatedSubExpressions[subExpression] = this.differentiate(tsa_compositeExpression.subExpressions.get(subExpression), type, monotonic, streamBuffers);
						if(differentiatedSubExpressions[subExpression].schema().hasAttribute(isNewAttributeName)){
							containsAnyIsNewAttribute = true;
							String isNewAttributeNewName = isNewAttributeName + subExpression;
							isNewAttributes.add(new BooleanTerm(BooleanAttribute.create(isNewAttributeNewName)));
							differentiatedSubExpressions[subExpression] = new DTSA_CompositeExpression(new TSA_RenameOperator(tsa_compositeExpression.operator.origin, new FixedArrayNavigableMap<String, String>(new String[]{isNewAttributeName}, new String[]{isNewAttributeNewName})), differentiatedSubExpressions[subExpression]);
						}
					}
					if(containsAnyIsNewAttribute){
						TSA_ImbeddingOperator isNewImbedding = new TSA_ImbeddingOperator(tsa_compositeExpression.operator.origin, new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(new BooleanAttribute[]{isNewAttribute}, new Term<?>[]{and(isNewAttributes)}));
						TSA_ProjectionOperator isNewProjection = new TSA_ProjectionOperator(tsa_compositeExpression.operator.origin, FixedArrayNavigableSet.merge(tsa_compositeExpression.schema().basicAttributes.keys(), new FixedArrayNavigableSet<String>(isNewAttributeName)));
						return new DTSA_CompositeExpression(isNewProjection, new DTSA_CompositeExpression(isNewImbedding, new DTSA_CompositeExpression(tsa_compositeExpression.operator, new FixedArray<DTSA_Expression>(false, differentiatedSubExpressions))));
					}
					else{
						return new DTSA_CompositeExpression(tsa_compositeExpression.operator, new FixedArray<DTSA_Expression>(false, differentiatedSubExpressions));
					}
				}
				case UNION: 
				{				
					assert tsa_compositeExpression.subExpressions.size() > 0;
					DTSA_Expression[] differentiatedSubExpressions = new DTSA_Expression[tsa_compositeExpression.subExpressions.size()];
					for (int subExpression = 0; subExpression < differentiatedSubExpressions.length; ++subExpression) {
						differentiatedSubExpressions[subExpression] = this.differentiate(tsa_compositeExpression.subExpressions.get(subExpression), type, monotonic, streamBuffers);
					}
					return new DTSA_CompositeExpression(tsa_compositeExpression.operator, new FixedArray<DTSA_Expression>(false, differentiatedSubExpressions));
				}
				case GROUPING:
				{
					assert tsa_compositeExpression.subExpressions.size() == 1;
					assert monotonic == false;
					return new DTSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers));
				}
				case IMBEDDING:
				{
					assert tsa_compositeExpression.subExpressions.size() == 1;
					return new DTSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers));
				}
				case PROJECTION:
				{ 
					assert tsa_compositeExpression.subExpressions.size() == 1;	
					DTSA_Expression differenciatedSubExpression = this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers);
					TSA_ProjectionOperator projection = (TSA_ProjectionOperator) tsa_compositeExpression.operator;
					
					if(monotonic & type !=  DTSA_QueryType.STATIC & type != DTSA_QueryType.INIT){
						AbstractFilter<ConditionFormula> identifierAttributeMinTimeFilter = new AbstractFilter<ConditionFormula>() {
							
							@Override
							public boolean accepts(final ConditionFormula element) {
								if(element instanceof AtomicStreamBoundFormula){
									AtomicStreamBoundFormula sbf = (AtomicStreamBoundFormula) element;
									StreamBound minTime = sbf.streamBound;
									if(minTime.type.ID == IDENTIFIER){
										if(sbf.atomicTerm instanceof BasicTypeAttribute){
											BasicTypeAttribute attribute = (BasicTypeAttribute) sbf.atomicTerm;
											return attribute.ID == minTime.type.ID;
										}
									}
								}
								return false;
							}
							
						};
						
						final FixedArrayNavigableMap<String, StreamBound> referencedIdentifierAttributes;
						{
							ConditionFormula minTimeFormula = not(and(differenciatedSubExpression.schema().relationConstraints()).focus(streamBoundFilter)).focus(identifierAttributeMinTimeFilter).subformula().flaten();
							//FIXME assert minTimeFormula.ID != ConditionFormulaID.TOP;
							//FIXME assert minTimeFormula.ID != ConditionFormulaID.BOTTOM;
							/*FIXME couldn't this be if there is already a isNewAttribute that covers all inputs? 
							introduce monotonic stream bounds check domination by monotonic stream bounds, introduce isNew if last monotonic stream bound (or equal attribute) is discarded. 
							In this case BOTTOM may possibly occur. 
							Better: if stream bound is is used in isNew then replace streamBound constraint by TOP. Then TOP would be allowed. But Bottom would not be allowed and ould not occur if implementation is correct;
							*/
							AbstractCollection<ConditionFormula> formula = minTimeFormula.collectFormulas(atomicFormulaFilter);
							
							ArrayList<String> referencedIdentifierAttributeNamesList = new ArrayList<String>();
							ArrayList<StreamBound> referencedIdentifierAttributeMinTimeList = new ArrayList<StreamBound>();
							
							for(ConditionFormula minTimeInequationFormula: formula){
								AtomicStreamBoundFormula streamBoundFormula = (AtomicStreamBoundFormula) minTimeInequationFormula;
								StreamBound streamBound = streamBoundFormula.streamBound;
								assert streamBoundFormula.atomicTerm instanceof BasicTypeAttribute;
								BasicTypeAttribute attribute = (BasicTypeAttribute) streamBoundFormula.atomicTerm;
								assert attribute.ID == IDENTIFIER;
								referencedIdentifierAttributeNamesList.add(attribute.name);
								referencedIdentifierAttributeMinTimeList.add(streamBound);
							}
							
							referencedIdentifierAttributes = new FixedArrayNavigableMap<String, StreamBound>(referencedIdentifierAttributeNamesList.toArray(new String[0]), referencedIdentifierAttributeMinTimeList.toArray(new StreamBound[0]));
						}
		
						final FixedArrayNavigableSet<String> remainingIdentifierAttributes;
						{
							FixedArrayNavigableMap<String, BasicType> attributes = tsa_compositeExpression.schema().basicAttributes;
							ArrayList<String> remainingIdAttributesList = new ArrayList<String>();
							for(int attribute = 0; attribute < attributes.size(); ++attribute){
								if(attributes.values().get(attribute).ID == IDENTIFIER){
									remainingIdAttributesList.add(attributes.keys().get(attribute));
								}
							}
							
							remainingIdentifierAttributes = new FixedArrayNavigableSet<String>(false, remainingIdAttributesList);
						}
						
						final FixedArrayNavigableMap<String, StreamBound> discardedIdentifierAttributes;
						{
							//TODO more efficient by setDifference with FixedArrayNavigableSet ?
							ArrayList<String> discardedIdAttributesList = new ArrayList<String>();
							ArrayList<StreamBound> correspondingMinTimesList = new ArrayList<StreamBound>();
							for (int minTimeFormula = 0; minTimeFormula < referencedIdentifierAttributes.size(); ++minTimeFormula) {
								String name = referencedIdentifierAttributes.keys().get(minTimeFormula);
								if(!remainingIdentifierAttributes.contains(name)){
									discardedIdAttributesList.add(name);
									correspondingMinTimesList.add(referencedIdentifierAttributes.values().get(minTimeFormula));
								}
							}
							discardedIdentifierAttributes = new FixedArrayNavigableMap<String, StreamBound>(new FixedArrayNavigableSet<String>(false, false, discardedIdAttributesList.toArray(new String[0])), new FixedArray<StreamBound>(false, correspondingMinTimesList.toArray(new StreamBound[0])));
						}
						 
						boolean hasIsNewAttribute = differenciatedSubExpression.schema().hasAttribute(isNewAttributeName);
						
						TSA_ProjectionOperator relaxedProjection = new TSA_ProjectionOperator(projection.origin, FixedArrayNavigableSet.merge(projection.attributes, new FixedArrayNavigableSet<String>(isNewAttributeName)));
	
						if(discardedIdentifierAttributes.size() > 0){
							
							ArithmeticAnalyser identifierAnalyser = this.identifierAnalyserFactory.make(and(differenciatedSubExpression.schema().tupleConstraints()), FixedArrayNavigableSet.merge(remainingIdentifierAttributes, discardedIdentifierAttributes.keys()), new FixedArrayNavigableSet<String>());
							
							FixedArrayNavigableMap<String, String> aliasMapping;
							final FixedArrayNavigableMap<String, StreamBound> nonReplaceableDiscardedIdentifierAttributes;
							
							{
								ArrayList<String> replaceableDiscardedAttributesList = new ArrayList<String>();
								ArrayList<String> replaceableDiscardedAttributesAliasList = new ArrayList<String>();
								ArrayList<String> nonReplaceableDiscardedAttributesList = new ArrayList<String>();
								ArrayList<StreamBound> nonReplaceableDiscardedAttributesMinTimeList = new ArrayList<StreamBound>();
								
								for(int discardedIdentifierAttribute = 0; discardedIdentifierAttribute < discardedIdentifierAttributes.size(); ++discardedIdentifierAttribute){
									String discardedAttribute = discardedIdentifierAttributes.keys().get(discardedIdentifierAttribute);
									boolean replaceable = false;
									for(int remainingIdentifierAttribute = 0; remainingIdentifierAttribute <remainingIdentifierAttributes.size(); ++remainingIdentifierAttribute){
										String remainingAttribute = remainingIdentifierAttributes.get(remainingIdentifierAttribute);
										long distance1 = identifierAnalyser.distance(discardedAttribute, true, remainingAttribute, true);
										long distance2 = identifierAnalyser.distance(remainingAttribute, true, discardedAttribute, true);
										if(ArithmeticAnalyser.constant(distance1) == 0 & !ArithmeticAnalyser.minusEpsilon(distance1) & ArithmeticAnalyser.constant(distance2) == 0 & !ArithmeticAnalyser.minusEpsilon(distance2)){
											replaceable = true;
											replaceableDiscardedAttributesList.add(discardedAttribute);
											replaceableDiscardedAttributesAliasList.add(remainingAttribute);
											break;
										}
									}
									if(!replaceable){
										nonReplaceableDiscardedAttributesList.add(discardedAttribute);
										nonReplaceableDiscardedAttributesMinTimeList.add(discardedIdentifierAttributes.values().get(discardedIdentifierAttribute));
									}
								}
								
								aliasMapping = new FixedArrayNavigableMap<String, String>(new FixedArrayNavigableSet<String>(false, replaceableDiscardedAttributesList), new FixedArray<String>(replaceableDiscardedAttributesAliasList));
								nonReplaceableDiscardedIdentifierAttributes = new FixedArrayNavigableMap<String, StreamBound>(new FixedArrayNavigableSet<String>(false, nonReplaceableDiscardedAttributesList), new FixedArray<StreamBound>(nonReplaceableDiscardedAttributesMinTimeList));
							}
							
							TemporalStreamSchema modifiedSchema;
							try {
								modifiedSchema = TemporalStreamSchema.create(differenciatedSubExpression.schema().ID, differenciatedSubExpression.schema().basicAttributes, differenciatedSubExpression.schema().tupleConstraints(), and(differenciatedSubExpression.schema().relationConstraints()).replaceAtomicTerm(new RenameAttributes(aliasMapping)).subformulas());
							} catch (TSA_SchemaException e) {
								throw new TSA_Exception(projection.origin, e);
							}
							
							DTSA_Expression schemaModifiedDifferentiatedSubExpression = differenciatedSubExpression.replaceSchema(modifiedSchema);
							
							if(nonReplaceableDiscardedIdentifierAttributes.size() > 0){
								ConditionFormula isNewFormula = this.isNewFormula(hasIsNewAttribute, nonReplaceableDiscardedIdentifierAttributes);
															
								if(hasIsNewAttribute){
									FixedArrayNavigableMap<String, String> mapping = new FixedArrayNavigableMap<String, String>(new String[]{isNewAttributeName}, new String[]{isNewAttributeAliasName});
									try {
										isNewFormula = isNewFormula.replaceAtomicTerm(new RenameAttributes(mapping));
									} catch (TSA_SchemaException e) {
										throw new TSA_Exception(projection.origin, e);
									}
									TSA_ImbeddingOperator isNewImbedding = new TSA_ImbeddingOperator(projection.origin, new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(new BooleanAttribute[]{isNewAttribute}, new Term<?>[]{isNewFormula}));
									return new DTSA_CompositeExpression(relaxedProjection, new DTSA_CompositeExpression(isNewImbedding, new DTSA_CompositeExpression(new TSA_RenameOperator(projection.origin, mapping), schemaModifiedDifferentiatedSubExpression)));
								}
								else{
									TSA_ImbeddingOperator isNewImbedding = new TSA_ImbeddingOperator(projection.origin, new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(new BooleanAttribute[]{isNewAttribute}, new Term<?>[]{isNewFormula}));
									return new DTSA_CompositeExpression(relaxedProjection, new DTSA_CompositeExpression(isNewImbedding, schemaModifiedDifferentiatedSubExpression));
								}
							}
							else{
								if(hasIsNewAttribute){								
									return new DTSA_CompositeExpression(relaxedProjection, schemaModifiedDifferentiatedSubExpression);
								}
								else{
									return new DTSA_CompositeExpression(projection, schemaModifiedDifferentiatedSubExpression);
								}
							}		
						}
						else{
							if(hasIsNewAttribute){
								return new DTSA_CompositeExpression(relaxedProjection, differenciatedSubExpression);
							}
							else{
								return new DTSA_CompositeExpression(projection, differenciatedSubExpression);
							}				
						}
					}
					else{
						assert !monotonic | type == DTSA_QueryType.STATIC | type == DTSA_QueryType.INIT;
						assert !differenciatedSubExpression.schema().hasAttribute(isNewAttributeName);
						return new DTSA_CompositeExpression(projection, differenciatedSubExpression);
					}
					
				}
				case SELECTION:
				{
					assert tsa_compositeExpression.subExpressions.size() == 1;
					return new DTSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers));
				}	
				case SETDIFFERENCE:
				{
					assert tsa_compositeExpression.subExpressions.size() == 2;
					assert monotonic == false;
					return new DTSA_CompositeExpression(tsa_compositeExpression.operator, new FixedArray<DTSA_Expression>(false, this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers), this.differentiate(tsa_compositeExpression.subExpressions.get(1), type, monotonic, streamBuffers)));
				}
				default:
					throw new Error("Unexpected case");
				}
			}
			case COMPLEX: 
			{
				switch (((TSA_ComplexOperator)tsa_compositeExpression.operator).ID) {
				case RENAME: {
					assert tsa_compositeExpression.subExpressions.size() == 1;
					DTSA_Expression differenciatedSubExpression = this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers);
					TSA_RenameOperator renameOperator = (TSA_RenameOperator) tsa_compositeExpression.operator;
					boolean isPrefix = renameOperator.mapping.keys().contains("");
					
					if(isPrefix & differenciatedSubExpression.schema().hasAttribute(isNewAttributeName)){
						FixedArrayNavigableSet<String> attributeNames = differenciatedSubExpression.schema().basicAttributes.keys();
						String[] newAttributeNames = new String[attributeNames.size()];
						String prefix = renameOperator.mapping.get("");
						
						for(int attribute = 0; attribute < newAttributeNames.length; ++attribute){
							newAttributeNames[attribute] = TSA_SyntaxPatterns.prefixAttribute(prefix, attributeNames.get(attribute));
						}
						
						FixedArrayNavigableMap<String, String> modifiedMapping = new FixedArrayNavigableMap<String, String>(attributeNames, new FixedArray<String>(false, newAttributeNames)); 
						TSA_RenameOperator modifiedRenameOperator = new TSA_RenameOperator(renameOperator.origin, modifiedMapping);
						return new DTSA_CompositeExpression(modifiedRenameOperator, differenciatedSubExpression);
					}
					else{
						//rename operator is not allowed to explicitly affect any attribute starting with .tsa
						return new DTSA_CompositeExpression(tsa_compositeExpression.operator, differenciatedSubExpression);
					}
				}
				case GROUPING_AND_AGGREGATION: {
					assert tsa_compositeExpression.subExpressions.size() == 1;
					assert monotonic == false;
					return new DTSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers));
				}
				case DISTINCT_PROJECTION: {
					assert tsa_compositeExpression.subExpressions.size() == 1;
					assert monotonic == false;
					return new DTSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers));
				}
				case ANTI_SEMI_JOIN: {
					assert tsa_compositeExpression.subExpressions.size() == 2;
					assert monotonic == false;
					return new DTSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers));
				}
				case SEMI_JOIN: {
					assert tsa_compositeExpression.subExpressions.size() == 2;
					assert monotonic == false;
					return new DTSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiate(tsa_compositeExpression.subExpressions.get(0), type, monotonic, streamBuffers));
				}
				default:
					throw new Error("Unexpected case");
				}
			}
			case COMPOSITE: 
			{
				//FIXME
				throw new Error("Not implemented case");
				//break;
			}
			default:
				throw new Error("Unexpected case");
			}
		}
		catch (TSA_Exception e) {
			throw e;
		}
		catch (RuntimeException e) {
			throw new TSA_Exception(tsa_compositeExpression.operator.origin, e);
		}
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(TSA_DefaultDifferencer.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}

}
