package de.lmu.ifi.pms.cep.tsa.conditions;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Element;

import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;

public class BooleanTerm extends AtomicConditionFormula {
	
	@Element(name="term")
	public final Term<?> term; 
	
	protected BooleanTerm(@Element(name="term") final Term<?> term) {
		super(isNormalizedOrFlatened(term), isNormalizedOrFlatened(term));
		this.term = term;
	}

	public BooleanTerm(final BooleanAttribute attribute) {
		this((Term<?>)attribute);
	}
	
	public BooleanTerm(final BooleanNamedConstant constant) {
		this((Term<?>)constant);
	}
	
	public BooleanTerm(final BooleanValue value) {
		this((Term<?>)value);
	}
	
	public BooleanTerm(final ConditionFormula condition) {
		this((Term<?>)condition);
	}
	
	@Override
	public boolean containsNegation() {
		if(this.term instanceof ConditionFormula){
			return ((ConditionFormula) this.term).containsNegation();
		}
		return false;
	}
	@Override
	public boolean containsConjunction() {
		if(this.term instanceof ConditionFormula){
			return ((ConditionFormula) this.term).containsConjunction();
		}
		return false;
	}
	@Override
	public boolean containsDisjunction() {
		if(this.term instanceof ConditionFormula){
			return ((ConditionFormula) this.term).containsDisjunction();
		}
		return false;
	}
	
	@Override
	protected ConditionFormula normalizeImpl() {
		if(this.term instanceof BooleanValue){
			return (((BooleanValue) this.term).value ? Top : Bottom);
		}
		else{ 
			assert this.term instanceof ConditionFormula;
			return ((ConditionFormula) this.term).normalize();
		}
	}
	
	@Override
	protected ConditionFormula flatenImpl() {
		if(this.term instanceof BooleanValue){
			return (((BooleanValue) this.term).value ? Top : Bottom);
		}
		else{ 
			assert this.term instanceof ConditionFormula;
			return ((ConditionFormula) this.term).flaten();
		}
	}
	
	protected static boolean isNormalizedOrFlatened(final Term<?> term){
		return term instanceof BooleanAttribute | (term instanceof BasicTypeVariable && ((BasicTypeVariable)term).ID == BasicTypeID.BOOLEAN) | term instanceof BooleanNamedConstant;
	}
	
	protected static boolean isValidTerm(final Term<?> term){
		return term instanceof BooleanAttribute | (term instanceof BasicTypeVariable && ((BasicTypeVariable)term).ID == BasicTypeID.BOOLEAN) | term instanceof BooleanNamedConstant | term instanceof BooleanValue | term instanceof ConditionFormula; 
	}

	@Override
	public ConditionFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
		Term<?> newTerm = this.term.replaceAtomicTerm(replacement);
		if(newTerm == this.term){
			return this;
		}
		else if(isValidTerm(newTerm)){
			return new BooleanTerm(newTerm);
		}
		else{
			throw new IllegalArgumentException("Replacement must be an boolean attribute, variable, constant, value or function;");			
		}
	}
	
	@Override
	public ConditionFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
		Term<?> newTerm = replacement.adapt(this.term);
		if(newTerm == this.term){
			return this;
		}
		else if(isValidTerm(newTerm)){
			return new BooleanTerm(newTerm);
		}
		else{
			throw new IllegalArgumentException("Replacement must be an attribute");			
		}
	}
	
	@Override
	protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms) {
		this.term.collect(filter, terms);
	}

	@Override
	protected int compareToImpl(final AtomicConditionFormula o) {
		int cmp = super.compareToImpl(o);
		if(cmp == 0){
			BooleanTerm c = (BooleanTerm)o;
			cmp = this.term.compareTo(c.term);
		}
		return cmp;
	}
	
	@Override
	public int hashCode() {
		return super.hashCode() ^ this.term.hashCode();
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepth) {
		indentTab(sb, indentDepth);
		sb.append("( ");
		this.term.writeTo(sb, 0);
		sb.append(" )");
	}
	
}
