package de.lmu.ifi.pms.cep.tsa.conditions;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementUnion;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.AbstractFilter;
import de.lmu.ifi.pms.java.util.filter.Filter;


public class DefaultDefinition extends AtomicConditionFormula {
	
	public static final AbstractFilter<ConditionFormula> implicitAttributeValuesSpecifications = new AbstractFilter<ConditionFormula>() {
		@Override
		public boolean accepts(final ConditionFormula element) {
			if(element instanceof DefaultDefinition){
				return true;				
			}
			return false;
		}
	};
	
	@Element(name="attribute",type=Attribute.class)
	public final Attribute<?> attribute;
	@ElementUnion({
		@Element(name="term", type=Term.class)
	})
	public final Term<?> definition;

	public DefaultDefinition(@Element(name="attribute",type=Attribute.class) final Attribute<?> attribute, @Element(name="term", type=Term.class) final Term<?> definition) {
		super(true, true);
		this.attribute = attribute;
		this.definition = definition;
	}

	@Override
	public ConditionFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
		Term<?> newAttribute = replacement.adapt(this.attribute);
		Term<?> newDefinition = this.definition.replaceAtomicTerm(replacement);
		if(newAttribute instanceof Attribute<?>){
			return new DefaultDefinition((Attribute<?>)newAttribute, newDefinition);
		}
		else{
			throw new IllegalArgumentException("Replacement for attribute must be an attribute");			
		}
	}
	
	@Override
	public ConditionFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
		Term<?> newAttrinute = replacement.adapt(this.attribute);
		Term<?> newDefinition = replacement.adapt(this.definition);
		if(newAttrinute instanceof Attribute<?>){
			return new DefaultDefinition((Attribute<?>)newAttrinute, newDefinition);
		}
		else{
			throw new IllegalArgumentException("Replacement for attribute must be an attribute");			
		}
	}
	
	@Override
	protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms) {
		this.attribute.collect(filter, terms);
		this.definition.collect(filter, terms);
	}

	@Override
	protected int compareToImpl(final AtomicConditionFormula o) {
		int cmp = super.compareToImpl(o);
		if(cmp == 0){
			DefaultDefinition c = (DefaultDefinition)o;
			cmp = this.attribute.compareTo(c.attribute);
			if(cmp == 0){
				cmp = this.definition.compareTo(c.definition);
			}
		}
		return cmp;
	}
	
	@Override
	public int hashCode() {
		return super.hashCode() ^ this.attribute.hashCode() ^ this.definition.hashCode();
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepth) {
		indentTab(sb, indentDepth);
		sb.append("DEFAULT{ ");
		this.attribute.writeTo(sb, 0);
		sb.append(" = ");
		this.definition.writeTo(sb, 0);
		sb.append(" }");
	}
	
}
