package de.lmu.ifi.pms.cep.tsa.conditions;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Element;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;


public class ForeignKeyConstraint extends AtomicConditionFormula {
	
	@Element(name="temporal-stream")
	public final String referencedStream;
	@Element(name="attribute")
	public final ReferenceAttribute attribute; 

	public ForeignKeyConstraint(final String referencedStream, final ReferenceAttribute attribute) {
		super(true, true);
		this.referencedStream = referencedStream;//FIXME allow specification of attributes, in this case always references to primary key
		this.attribute = attribute;
	}

	@Override
	public ConditionFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
		Term<?> newAttrinute = replacement.adapt(this.attribute);
		if(newAttrinute instanceof ReferenceAttribute){
			return new ForeignKeyConstraint(this.referencedStream, (ReferenceAttribute)newAttrinute);
		}
		else{
			throw new IllegalArgumentException("Replacement must be a identifier attribute");			
		}
	}
	
	@Override
	public ConditionFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
		return this.replaceAtomicTerm(replacement);
	}
	
	@Override
	protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms) {
		this.attribute.collect(filter, terms);
	}
	
	@Override
	protected int compareToImpl(final AtomicConditionFormula o) {
		int cmp = super.compareToImpl(o);
		if(cmp == 0){
			ForeignKeyConstraint c = (ForeignKeyConstraint)o;
			cmp = this.referencedStream.compareTo(c.referencedStream);
			if(cmp == 0){
				cmp = this.attribute.compareTo(c.attribute);
			}
		}
		return cmp;
	}
	
	@Override
	public int hashCode() {
		return super.hashCode() ^ this.referencedStream.hashCode() ^ this.attribute.hashCode();
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepth) {
		indentTab(sb, indentDepth);
		sb.append("FOREIGN KEY {");
		this.attribute.writeTo(sb, 0);
		sb.append(" REFERENCES ");
		sb.append(this.referencedStream);
		sb.append("}");
	}

}
