package de.lmu.ifi.pms.cep.tsa.conditions;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;


@Root(name="param")
public class Parameter extends AtomicTerm<BasicType> {
	
	@org.simpleframework.xml.Attribute(name="name")
	public final String name;
	
	@Element(name="type", required=true)
	protected final BasicType getType(){
		return this.type;
	}
	
	public Parameter(final String name, final BasicTypeID typeID) {
		this(name, typeID.type);
	}

	public Parameter(@org.simpleframework.xml.Attribute(name="name") final String name, @Element(name="type", required=true) final BasicType type) {
		super(AtomicTermID.PARAMETER, type);
		if(!TSA_SyntaxPatterns.isValidParameterName(name)){
			throw new IllegalArgumentException("Invalid parameter name: " + name);
		}
		this.name = name;
		//TODO allow only IDENTIFIER and TIMESTAMP?
	}
	
	public Parameter rename(final String newName){
		return new Parameter(newName, this.type);
	}
	
	@Override
	protected int compareToImpl(final Term<?> o) {
		int cmp;
		if(o instanceof Parameter){
			cmp = this.name.compareTo(((Parameter)o).name);
			if(cmp == 0){
				cmp = super.compareToImpl(o);
			}
		}
		else{
			cmp = super.compareToImpl(o);
			assert cmp != 0;
		}
		return cmp;
	}
	
	@Override
	protected boolean equalsImpl(final Object o) {
		if(this.getClass() == o.getClass()){
			return this.name.equals(((Parameter)o).name);
		}
		else{
			return false;
		}
	}
	
	@Override
	public int hashCode() {
		return super.hashCode() ^ this.name.hashCode();
	}

	
	@Override
	public String toString() {
		return this.name;
	}
}
