package de.lmu.ifi.pms.cep.tsa.conditions;

import de.lmu.ifi.pms.cep.tsa.schema.AType;
import de.lmu.ifi.pms.cep.tsa.schema.NoType;



public abstract class TupleFunction<TypeClass extends AType> extends Function<TypeClass> {
		
	@SuppressWarnings("hiding")
	public final TupleFunctionID ID;
	
	TupleFunction(final TupleFunctionID ID, final TypeClass type, final java.util.ArrayList<Term<?>> arguments) {
		this(ID, type, arguments.toArray(new Term<?>[arguments.size()]));
	}
	
	public TupleFunction(final TupleFunctionID ID, final TypeClass type, final Term<?>... arguments) {
		this(ID, ID.toString(), type, arguments);
	}
	
	TupleFunction(final TupleFunctionID ID, final String name, final TypeClass type, final java.util.ArrayList<Term<?>> arguments) {
		this(ID, name, type, arguments.toArray(new Term<?>[arguments.size()]));
	}
	
	TupleFunction(final TupleFunctionID ID, final String name, final TypeClass type, final Term<?>... arguments) {
		super(FunctionID.TUPLE, name, type, ID.minArity, ID.maxArity, arguments);
		this.ID = ID;
	}
	
	@Override
	protected int minArity() {
		return this.ID.minArity;
	}
	
	@Override
	protected int maxArity() {
		return this.ID.maxArity;
	}

	@Override
	public abstract TupleFunction<TypeClass> newTerm(final Term<?>... arguments);
	
	public static TupleFunction<NoType> apply(final TupleFunctionID ID, final Term<?>... arguments){
		return apply(ID, NoType.instance, arguments);
	}
	
	public static <TypeClass extends AType> TupleFunction<TypeClass> apply(final TupleFunctionID ID, final TypeClass type,final Term<?>... arguments){
		switch (ID) {
		case PLUS:
			return new Plus<TypeClass>(type, arguments[0], arguments[1]);			
		case MINUS:
			return new Minus<TypeClass>(type, arguments[0], arguments[1]);
		case LEAST:
			return new Least<TypeClass>(type, arguments);
		case GREATEST:
			return new Greatest<TypeClass>(type, arguments);
		case TIMES:
			return new Times<TypeClass>(type, arguments[0], arguments[1]);
		case DIVISION:
			return new Division<TypeClass>(type, arguments[0], arguments[1]);
		case POWER:
			return new Power<TypeClass>(type, arguments[0], arguments[1]);
		case USER:
			throw new IllegalArgumentException("Does not support user functions.");
		default:
			throw new Error("Unexpected case: " + ID.toString());
		}
	}
}
