package de.lmu.ifi.pms.cep.tsa.exceptions;

import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.FromJavaCollectionWrapper;

public class TSA_ComprehensiveException extends TSA_Exception {
	
	public final FixedArray<TSA_Exception> exceptions;
	
	public TSA_ComprehensiveException(final AbstractCollection<TSA_Exception> exceptions) {
		this(new FixedArray<TSA_Exception>(exceptions));
	}

	public TSA_ComprehensiveException(final FixedArray<TSA_Exception> exceptions) {
		super();
		this.exceptions = exceptions;
	}
	
	public static <ExceptionType extends Exception> String message(final java.util.Collection<ExceptionType> exceptions){
		return message(new FromJavaCollectionWrapper<ExceptionType>(exceptions));
	}
	
	@Override
	public String getSimpleMessage() {
		return message(this.exceptions);
	}
	
	public static String message(final Collection<? extends Exception> exceptions){
		StringBuilder message = new StringBuilder();
		message.append("The following " + exceptions.size() + " exceptions occured: ");
		message.append("\n---------------------");
		int counter = 1;
		if(exceptions.size() > 0){
			for(Exception e: exceptions){
				message.append("\n");
				message.append(counter);
				message.append(") ");
				writeMessage(message, e);
				message.append("\n");
				++counter;
			}
			message.setLength(message.length() - "\n".length());
		}
		message.append("\n---------------------");
		return message.toString();
	}
	
	public static void writeMessage(final StringBuilder message, final Throwable e){
		message.append(e.toString());
		for(StackTraceElement stackTraceElement: e.getStackTrace()){
			message.append("\n at ");
			message.append(stackTraceElement.toString());
		}
		if(e.getCause() != null){
			message.append("\nCaused by: ");
			writeMessage(message, e.getCause());
		}
	}

}
