package de.lmu.ifi.pms.cep.tsa.operators;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementMap;

import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TypedAttribute;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.Type;

import de.lmu.ifi.pms.java.util.adaption.CollectionAdapter;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


public class TSA_GroupingAndAggregationOperator extends TSA_ComplexOperator
{
	public final FixedArrayNavigableSet<String> grouping_attributes;
	
	@ElementList(name="grouping-attributes",entry="attribute")
	java.util.ArrayList<String> getAttributeList(){
		return new java.util.ArrayList<String>(this.grouping_attributes.javaCollection());
	}
	
	@ElementMap(entry="definition", key="attribute", value="value", inline=true)
	java.util.HashMap<TypedAttribute<?>, Term<?>> getDefinitions(){
		return new java.util.HashMap<TypedAttribute<?>, Term<?>>(this.definitionsOfAggregationAttributes.javaMap());
	}
	
	public final FixedArrayNavigableMap<? extends TypedAttribute<?>, ? extends Term<?>> definitionsOfAggregationAttributes;
	
	public TSA_GroupingAndAggregationOperator(final String origin, final FixedArray<String> grouping_attributes, final FixedArrayNavigableMap<? extends TypedAttribute<?>, ? extends Term<?>> definitionsOfAggregationAttributes){
		this(new FixedArray<String>(origin), grouping_attributes, definitionsOfAggregationAttributes);
	}
	
	@SuppressWarnings("unchecked")
	public TSA_GroupingAndAggregationOperator(@ElementList(entry="origin", inline=true) final java.util.ArrayList<String> origin
			, @ElementList(name="grouping-attributes",entry="attribute") final java.util.ArrayList<String> grouping_attributes
			, @ElementMap(entry="definition", key="attribute", value="value", inline=true) final java.util.HashMap<? extends TypedAttribute<?>, ? extends Term<?>> definitionsOfAggregationAttributes){
		this(new FixedArray<String>(origin), new FixedArray<String>(grouping_attributes), new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>((java.util.Map<TypedAttribute<?>, Term<?>>)definitionsOfAggregationAttributes));
	}
	
	public TSA_GroupingAndAggregationOperator(final FixedCollection<String> origin, final FixedArray<String> grouping_attributes, final FixedArrayNavigableMap<? extends TypedAttribute<?>, ? extends Term<?>> definitionsOfAggregationAttributes)
	{
		super(TSA_ComplexOperatorID.GROUPING_AND_AGGREGATION, origin);
		//FIXME check attributes and selection conditions
		this.grouping_attributes = TemporalStreamSchema.normalizeforRetain(grouping_attributes);
		this.definitionsOfAggregationAttributes = definitionsOfAggregationAttributes;
	}
	
	@Override
	public int inputArity() {
		return 1;
	}
	
	
	@Override
	public boolean isMonotonic(){
		return false;
	}

	@Override
	public TemporalStreamSchema apply(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception {
		this.checkInputs(inputs);
		assert inputs.size() == 1;
		
		TemporalStreamSchema input = inputs.iterator().next();
		
		FixedArrayNavigableSet<String> newAttributeNames = new FixedArrayNavigableSet<String>(false, false, new CollectionAdapter<TypedAttribute<?>, String>(this.definitionsOfAggregationAttributes.keys()){
			@Override
			public String adapt(final TypedAttribute<?> source) {
				return source.name;
			}
		}.toArray(new String[this.definitionsOfAggregationAttributes.size()]));//Attributes sorted by name
		
		Type[] newAttributeTypes = new Type[newAttributeNames.size()];
		for(int attribute = 0; attribute < newAttributeTypes.length; ++attribute){
			newAttributeTypes[newAttributeNames.indexOf(this.definitionsOfAggregationAttributes.keys().get(attribute).name)] = this.definitionsOfAggregationAttributes.keys().get(attribute).type;
		}
		
		FixedArrayNavigableMap<String, Type> newAttributes = new FixedArrayNavigableMap<String, Type>(newAttributeNames, new FixedArray<Type>(false, newAttributeTypes));
		
		//FIXME derived constraints
		try {
			return input.retain(this.grouping_attributes).extend(newAttributes);
		} catch (TSA_SchemaException e) {
			throw new TSA_Exception(this.origin, e);
		}
	}
	
	@Override
	public void check(final Collection<TemporalStreamSchema> inputs) {
		//FIXME implement
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepthFirstLine, final int indentDepth) {
		indentTab(sb, indentDepthFirstLine);
		sb.append(this.ID.toString());
		assert this.grouping_attributes.size() > 0;
		sb.append("[");
		for(String grouping_attribute: this.grouping_attributes){
			sb.append(grouping_attribute);
			sb.append(", ");
		}
		sb.setLength(sb.length() - ", ".length());
		sb.append("]");
		if(this.definitionsOfAggregationAttributes.size() > 0){
			sb.append("[");
			for(int aggregationAttributeDefinition = 0; aggregationAttributeDefinition < this.definitionsOfAggregationAttributes.size(); ++aggregationAttributeDefinition){				
				sb.append(this.definitionsOfAggregationAttributes.keys().get(aggregationAttributeDefinition));
				sb.append(" = ");
				sb.append(this.definitionsOfAggregationAttributes.values().get(aggregationAttributeDefinition));
				sb.append(", ");
			}
			sb.setLength(sb.length() - ", ".length());
			sb.append("]");
		}
		//FIXME derived constraints
	}
	
}
