package de.lmu.ifi.pms.cep.tsa.operators;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.ElementList;

import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;

import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;


public class TSA_ProjectionOperator extends TSA_BasicOperator
{
	
	public final FixedArrayNavigableSet<String> attributes;
	
	@ElementList(entry="attribute",inline=true)
	java.util.ArrayList<String> getAttributeList(){
		return new java.util.ArrayList<String>(this.attributes.javaCollection());
	}
	
	public TSA_ProjectionOperator(final String origin, final FixedArray<String> attributes) throws TSA_Exception{
		this(new FixedArray<String>(false, origin), attributes);
	}
	
	public TSA_ProjectionOperator(@ElementList(entry="origin", inline=true) final java.util.ArrayList<String> origin
			, @ElementList(entry="attribute",inline=true) final java.util.ArrayList<String> attributes) throws TSA_Exception{
		this(new FixedArray<String>(origin), new FixedArray<String>(attributes));
	}

	public TSA_ProjectionOperator(final FixedCollection<String> origin, final FixedArray<String> attributes) throws TSA_Exception{
		super(TSA_BasicOperatorID.PROJECTION, origin);
		try{
			TemporalStreamSchema.checkRetainStatic(attributes);
		} catch (TSA_SchemaException e) {
			throw new TSA_Exception(this.origin, e);
		}
		this.attributes = TemporalStreamSchema.normalizeforRetain(attributes);
	}
	
	/*
	public TSA_ProjectionOperator(final String origin, final String... attributes){
		this(new FixedArray<String>(origin), attributes);
	}*/
	
	/*
	public TSA_ProjectionOperator(final FixedCollection<String> origin, final String... attributes)
	{
		this(origin, new FixedArray<String>(false, attributes));
	}
	*/
	
	@Override
	public int inputArity() {
		return 1;
	}
	
	@Override
	public TemporalStreamSchema apply(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception{
		this.checkInputs(inputs);
		assert inputs.size() == 1;
		TemporalStreamSchema input = inputs.iterator().next();
		
		try {
			return input.retain(this.attributes);
		} catch (TSA_SchemaException e) {
			throw new TSA_Exception(this.origin, e);
		}
	}
		
	@Override
	public boolean isMonotonic(){
		return true;
	}
	
	@Override
	public void checkInputs(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception {
		super.checkInputs(inputs);
	}
	
	@Override
	public void check(final Collection<TemporalStreamSchema> inputs) {
		//FIXME implement
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepthFirstLine, final int indentDepth) {
		indentTab(sb, indentDepthFirstLine);
		sb.append(this.ID.toString());
		assert this.attributes.size() > 0;
		sb.append("[");
		for(String attribute: this.attributes){
			sb.append(attribute);
			sb.append(", ");
		}
		sb.setLength(sb.length() - ", ".length());
		sb.append("]");
	}
}
