package de.lmu.ifi.pms.cep.tsa.schema;

import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_TypeException;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.AbstractFilter;
import de.lmu.ifi.pms.java.util.filter.Filter;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


@Root(name="type")
public abstract class Type extends AType implements Comparable<Type>{
	
	public final TypeID ID;

	public Type(final TypeID ID)
	{
		super();
		this.ID = ID;
	}
	
	public Type typeOf(final String name) throws TSA_TypeException{
		if(TSA_SyntaxPatterns.isValidAttributeName(name)){
			return this.typeOf(TSA_SyntaxPatterns.splitAttributeName(name), 0);
		}
		else{
			throw new IllegalArgumentException("Invalid attribute name: " + name);
		}
	}
	
	public Type typeOf(final String[] name)  throws TSA_TypeException{
		if(TSA_SyntaxPatterns.isValidAttributeName(name)){
			return this.typeOf(name, 0);
		}
		else{
			throw new IllegalArgumentException("Invalid attribute name: " + TSA_SyntaxPatterns.attributeName(name));
		}
	}
	
	protected abstract Type extend(String[][] newAttributeNames, Type[]newAttributeTypes, int startIndex, int endIndex, int level) throws TSA_TypeException;
	
	protected abstract Type retain(String[][] retainedAttributes, int startIndex, int endIndex, int level) throws TSA_TypeException;
	
	protected abstract Type typeOf(final String[] name, final int index) throws TSA_TypeException;
	
	public abstract boolean isEmpty();
	
	public FixedArrayNavigableMap<String, BasicType> basicAttributes(){
		try{
			return this.basicAttributes("");
		}
		catch (Exception e) {
			throw new Error("Unexpected case. Should not happen.", e);
		}
	}
	
	public FixedArrayNavigableMap<String, BasicType> basicAttributes(final String ownName) throws TSA_TypeException{
		StringBuilder namePrefix = new StringBuilder(ownName);
		if(!ownName.equals("") & !TSA_SyntaxPatterns.isValidAttributeName(ownName)){
			throw new TSA_TypeException("Invalid argument name: " + ownName);
		}
		ArrayList<String> nameAcc = new ArrayList<String>();
		ArrayList<BasicType> basicTypeAcc = new ArrayList<BasicType>();
		this.basicAttributes(namePrefix, nameAcc, basicTypeAcc);
		return new FixedArrayNavigableMap<String, BasicType>(new FixedArrayNavigableSet<String>(false, nameAcc), new FixedArray<BasicType>(basicTypeAcc));
	}
	
	protected abstract void basicAttributes(final StringBuilder namePrefix, final ArrayList<String> nameAcc, final ArrayList<BasicType> basicTypeAcc);
	
	public FixedArray<ConditionFormula> tupleConstraints(final String ownName){
		return this.tupleConstraints(ownName, AbstractFilter.<String>allFilter());		
	}
	
	public FixedArray<ConditionFormula> tupleConstraints(final String ownName, final Filter<String> attributeFilter){
		ArrayList<ConditionFormula> constraintAcc = new ArrayList<ConditionFormula>();
		this.tupleConstraints(ownName, constraintAcc, attributeFilter);
		return new FixedArray<ConditionFormula>(false, constraintAcc.toArray(new ConditionFormula[constraintAcc.size()]));
	}
	
	protected abstract void tupleConstraints(final String ownName, ArrayList<ConditionFormula> constraintAcc, final Filter<String> attributeFilter);
	
	@Override
	public final int compareTo(final Type o) {
		if(this == o){
			return 0;
		}
		else{
			return this.compareToImpl(o);
		}
	}
	
	protected int compareToImpl(final Type o) {
		return this.ID.ordinal() - o.ID.ordinal();
	}
	
	@Override
	public final boolean equals(final Object o) {
		if(this == o){
			return true;
		}
		else{
			return this.equalsImpl(o);
		}
	}
	
	protected boolean equalsImpl(final Object o){
		if(this.getClass().isInstance(o)){//succeeds only if obj is at least of type Type
			return this.compareToImpl((Type)o) == 0;
		}
		else{
			return false;
		}
	}
	
	@Override
	public int hashCode() {
		return this.ID.hashCode();
	}
	
	public final boolean structuralEquals(final Type type){
		if(this == type){
			return true;
		}
		else{
			return this.structuralEqualsImpl(type);
		}
	}
	
	protected abstract boolean structuralEqualsImpl(final Type type);

}
