package de.lmu.ifi.pms.cep.tsa.time.constraints;

import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID.*;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID;
import de.lmu.ifi.pms.java.lang.iteration.ComposedIterator;
import de.lmu.ifi.pms.java.lang.iteration.FixedCrossProductArrayIterator;
import de.lmu.ifi.pms.java.lang.iteration.Iterator;
import de.lmu.ifi.pms.java.util.adaption.FixedIteratorAdapter;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;

public class ConjunctiveTemporalRelationFormula extends CompositeTemporalRelationFormula{
	
	
	public ConjunctiveTemporalRelationFormula(final Collection<? extends TemporalRelationFormula> subformulas) {
		this(subformulas.toArray(new TemporalRelationFormula[subformulas.size()]));
	}

	public ConjunctiveTemporalRelationFormula(final TemporalRelationFormula... subformulas) {
		this(initialIsNormalized(subformulas), subformulas);
	}
	
	ConjunctiveTemporalRelationFormula(final boolean isNormalized, final TemporalRelationFormula... subformulas) {
		this(isNormalized, initialIsFlatened(isNormalized, subformulas), subformulas);
	}
	
	ConjunctiveTemporalRelationFormula(final boolean isNormalized, final boolean isFlatened, final TemporalRelationFormula... subformulas) {
		super(ConditionFormulaID.CONJUNCTIVE, isNormalized, isFlatened, subformulas);
		assert isNormalized == initialIsNormalized(subformulas);
		assert isFlatened == initialIsFlatened(isNormalized, subformulas);
	}
	
	protected static boolean initialIsNormalized(final TemporalRelationFormula... subformulas){
		if(subformulas.length < 2){
			return false;
		}
		else{
			for(TemporalRelationFormula subformula: subformulas){
				if(subformula.ID == TOP | subformula.ID == BOTTOM | !subformula.isNormalized){
					return false;
				}
			}
			return true;
		}
	}
	
	protected static boolean initialIsFlatened(final boolean isNormalized, final TemporalRelationFormula... subformulas){
		if(isNormalized){
			for(TemporalRelationFormula subformula: subformulas){
				if((!subformula.isFlatened) || subformula.ID == CONJUNCTIVE){
					return false;
				}
			}
			return true;
		}
		else{
			return false;
		}
	}
	
	@Override
	protected TemporalRelationFormula simpleForm(final boolean positivePolarity) {
		boolean changed = false;
		TemporalRelationFormula[] newSubformulas = new TemporalRelationFormula[this.subformulas.length];
		for(int subformula = 0; subformula < this.subformulas.length; ++subformula){
			newSubformulas[subformula] = this.subformulas[subformula].simpleForm(positivePolarity);
			changed = changed | newSubformulas[subformula] != this.subformulas[subformula];
		}
		
		if(positivePolarity){
			if(changed){
				return new ConjunctiveTemporalRelationFormula(newSubformulas);
			}
			else{
				return this;
			}
		}
		else{
			return new DisjunctiveTemporalRelationFormula(newSubformulas);
		}
	}

	@Override
	protected TemporalRelationFormula normalizeImpl() {
		ArrayList<TemporalRelationFormula> new_subformulas = new ArrayList<TemporalRelationFormula>(this.subformulas.length);
		for(TemporalRelationFormula subformula: this.subformulas){
			TemporalRelationFormula normalizedSubFormula = subformula.normalize();			
			if(normalizedSubFormula.ID == TOP){
				
			}
			else if(normalizedSubFormula.ID == BOTTOM){
				return Bottom;
			}
			else{
				new_subformulas.add(normalizedSubFormula);
			}
		}
		if(new_subformulas.size() == 0){
			return Top;
		}		
		else if(new_subformulas.size() == 1){
			return new_subformulas.first();
		}
		else{
			return new ConjunctiveTemporalRelationFormula(true, new_subformulas.toArray(new TemporalRelationFormula[new_subformulas.size()]));
		}
	}
	
	@Override
	public TemporalRelationFormula flatenImpl() {
		assert this.isNormalized;
		assert !this.isFlatened;
		
		ArrayList<TemporalRelationFormula> new_subformulas = new ArrayList<TemporalRelationFormula>(this.subformulas.length);
		this.flatenImpl(new_subformulas);
		
		return new ConjunctiveTemporalRelationFormula(true, true, new_subformulas.toArray(new TemporalRelationFormula[new_subformulas.size()]));
	}
	
	protected void flatenImpl(final ArrayList<TemporalRelationFormula> new_subformulas){
		for(TemporalRelationFormula c: this.subformulas){				
			if(c.ID == CONJUNCTIVE){
				((ConjunctiveTemporalRelationFormula)c).flatenImpl(new_subformulas);
			}
			else{
				TemporalRelationFormula flatened = c.flaten();
				assert flatened.ID != TOP;
				assert flatened.ID != BOTTOM;
				assert flatened.ID != CONJUNCTIVE;
				
				new_subformulas.add(flatened);		
			}
		}
	}
	
	@Override
	public int tempDNFsize() {
		int size = 1;
		for(TemporalRelationFormula c: this.subformulas){
			size *= c.tempDNFsize();
		}
		return size;
	}

	@Override
	protected AbstractCollection<TemporalRelationFormula> tempDNFiteration(final boolean positivePolarity) {
		@SuppressWarnings("unchecked")
		final AbstractCollection<TemporalRelationFormula>[] iterations = new AbstractCollection[this.subformulas.length];
		int size = 0;
		for(int i = 0; i < iterations.length; ++i){
			iterations[i] = this.subformulas[i].tempDNFiteration(positivePolarity);
			size *= iterations[i].size();
		}
		final int fixedSize = size;
		return new AbstractCollection<TemporalRelationFormula>() {

			@Override
			public Iterator<TemporalRelationFormula> iterator() {
				return new FixedIteratorAdapter<TemporalRelationFormula[], TemporalRelationFormula>(new FixedCrossProductArrayIterator<TemporalRelationFormula>(TemporalRelationFormula.class, iterations)) {

					@Override
					public TemporalRelationFormula adapt(final TemporalRelationFormula[] subformulas) {
						if(positivePolarity){
							return new ConjunctiveTemporalRelationFormula(subformulas);
						}
						else{
							return new DisjunctiveTemporalRelationFormula(subformulas);
						}
					}
					
				}; 
			}

			@Override
			public int size() {
				return fixedSize;
			}
		};
	}
	
	@Override
	public int tempCNFsize() {
		int size = 0;
		for(TemporalRelationFormula c: this.subformulas){
			size += c.tempDNFsize();
		}
		return size;
	}

	@Override
	protected AbstractCollection<TemporalRelationFormula> tempCNFiteration(final boolean positivePolarity) {
		@SuppressWarnings("unchecked")
		final AbstractCollection<TemporalRelationFormula>[] iterations = new AbstractCollection[this.subformulas.length];
		int size = 0;
		for(int i = 0; i < iterations.length; ++i){
			iterations[i] = this.subformulas[i].tempCNFiteration(true);
			size += iterations[i].size();
		}
		final int fixedSize = size;
		return new AbstractCollection<TemporalRelationFormula>() {

			@Override
			public Iterator<TemporalRelationFormula> iterator() {
				return new ComposedIterator<TemporalRelationFormula>(iterations);
			}

			@Override
			public int size() {
				return fixedSize;
			}
		};
	}

	@Override
	protected CompositeTemporalRelationFormula newFormula(final TemporalRelationFormula[] subformulas) {
		return new ConjunctiveTemporalRelationFormula(subformulas);
	}
}
