package de.lmu.ifi.pms.java.lang.iteration;

import java.util.NoSuchElementException;

import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.stream.Source;
import de.lmu.ifi.pms.java.util.stream.SourceStream;
import de.lmu.ifi.pms.java.util.stream.SourceStreamWrapper;


public abstract class AbstractIterator<Type> implements Iterator<Type> , Source<Type>{

	@Override
	public abstract boolean hasNext();

	@Override
	public abstract Type next();
	
	protected void checkNext(){
		if(!this.hasNext()){
			throw new NoSuchElementException();
		}
	}
	
	public int nextInt(){
		return AbstractCollection.toInt(this.next());
	}
	
	public long nextLong(){
		return AbstractCollection.toLong(this.next());
	}
	
	public double nextDouble(){
		return AbstractCollection.toDouble(this.next());
	}
	
	public float nextFloat(){
		return AbstractCollection.toFloat(this.next());
	}
	
	public byte nextByte(){
		return AbstractCollection.toByte(this.next());
	}
	
	public short nextShort(){
		return AbstractCollection.toShort(this.next());
	}
	
	public char nextChar(){
		return AbstractCollection.toChar(this.next());
	}
	
	public boolean nextBoolean(){
		return AbstractCollection.toBoolean(this.next());
	}

	@Override
	@Deprecated
	public void remove(){
		throw new UnsupportedOperationException();
	}
	
	@Override
	public SourceStream<Type> stream() {
		return new SourceStreamWrapper<Type>(this);
	}

}
