package de.lmu.ifi.pms.java.lang.iteration;

import java.util.Iterator;

import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.Arrays;




public class FixedCrossProductArrayIterator<Type> extends AbstractIterator<Type[]> implements FixedIterator<Type[]>{
	
	final int size;
	final int max_index;
	final Iterable<Type>[] iterables;
	final Iterator<Type>[] iterators;
	final Type[] tuple;
	
	int index;
	boolean searchNext;
	
	@SuppressWarnings("unchecked")
	public FixedCrossProductArrayIterator(final Class<Type> type, final Iterable<Type>... iterables){
		this.size = iterables.length;
		this.max_index = this.size-1;
		this.iterables = iterables;
		this.iterators = new Iterator[this.size];
		this.tuple = Arrays.getArray(type, this.size);
		
		if(this.max_index < 0){
			this.index = -1;
			this.searchNext = false;
		}
		else{
			this.index = 0;
			this.iterators[0] = iterables[0].iterator();
			this.searchNext = true;
		}
	}
	
	@SuppressWarnings("unchecked")
	public FixedCrossProductArrayIterator(final Class<Type> type, final Collection<? extends Iterable<Type>> iterables){
		this(type, iterables.toArray(new Iterable[iterables.size()]));
	}

	@Override
	public boolean hasNext() {
		Iterator<Type> currentIterator;
		while(this.searchNext && this.index >= 0){
			currentIterator = this.iterators[this.index]; 
			if(currentIterator.hasNext()){
				this.tuple[this.index] = currentIterator.next();
				if(this.index < this.max_index){
					++this.index;
					this.iterators[this.index] = this.iterables[this.index].iterator();
				}
				else{
					this.searchNext = false;
				}
			}
			else{
				this.iterators[this.index] = null;
				this.tuple[this.index] = null;
				--this.index;
			}
		}
		return this.index >= 0;
	}

	@Override
	public Type[] next() {
		this.checkNext();
		this.searchNext = true;
		return Arrays.clone(this.tuple);
	}
	
	@Override
	@Deprecated
	public final void remove() {
		super.remove();
	}

}
