/**
 * 
 */
package de.lmu.ifi.pms.java.util.collection.extendible.list;

import java.util.NoSuchElementException;

public abstract class AbstractSingleLinkedEntry<Type, ListEntryType extends AbstractSingleLinkedEntry<Type, ListEntryType>> implements SingleLinkedListEntry<Type, ListEntryType>{	
	
	protected AbstractSingleLinkedList<Type, ListEntryType> linkingList;
	
	protected ListEntryType previous;

	public AbstractSingleLinkedEntry(final AbstractSingleLinkedList<Type, ListEntryType> linkingList) {
		super();
		this.linkingList = linkingList;
	}
	
	public AbstractSingleLinkedEntry() {
		this(null);
	}
	
	@Override
	public boolean isLinked() {
		return this.linkingList != null;
	}
	
	public boolean isUnlinked(){
		return this.linkingList == null;
	}
	
	protected void checkUnlinked(){
		if(this.isLinked()){
			throw new IllegalStateException();
		}
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.extendible.list.SingleLinkedListEntry#linkingList()
	 */
	@Override
	public AbstractSingleLinkedList<Type, ListEntryType> linkingList() {
		this.checkUnlinked();
		return this.linkingList;
	}
	
	protected void unlink(){
		this.linkingList = null;
		--this.linkingList.size;
	}
	
	protected void link(final AbstractSingleLinkedList<Type, ListEntryType> list){
		assert this.isUnlinked();
		this.linkingList = list;
		++this.linkingList.size;
	}

	@Override
	public boolean hasPrevious() {
		return !this.isFoot();
	}
	
	@Override
	public boolean isFoot(){
		return this.previous == null;
	}
	
	protected void checkPrevious(){
		if(this.hasPrevious()){
			throw new NoSuchElementException();
		}
	}

	@Override
	public ListEntryType previous() {
		this.checkPrevious();
		return this.previous;
	}
	
	protected void setPrevious(final ListEntryType entry){
		this.previous = entry;
	}

	@Override
	public void insertBefore(final ListEntryType entry) {
		entry.checkUnlinked();
		entry.link(this.linkingList);
		entry.setPrevious(this.previous);
		this.setPrevious(entry);
	}

	@Override
	public void removePrevious() {
		this.previous().unlink();
		this.setPrevious(this.previous.previous);
	}

	@Override
	public void removePreceeding() {
		 while(this.hasPrevious()){
			 this.removePrevious();
		 }
	}

	@Override
	public abstract Type value();
	
	@Override
	public abstract void setValue(Type newValue);
	
}