package de.lmu.ifi.pms.java.util.collection.javaWrapper;

import java.util.Comparator;
import java.util.Iterator;

import de.lmu.ifi.pms.java.lang.iteration.ExtendibleIterator;
import de.lmu.ifi.pms.java.lang.iteration.JavaIteratorExtendibleWrapper;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.extendible.set.ExtendibleNavigableSet;
import de.lmu.ifi.pms.java.util.collection.set.AbstractSortedSet;
import de.lmu.ifi.pms.java.util.collection.set.NavigableSet;


public class FromJavaExtendibleNavigableSetWrapper<Type> extends AbstractSortedSet<Type> implements ExtendibleNavigableSet<Type>{
	
	private final java.util.NavigableSet<Type> javaCollection;
	
	public FromJavaExtendibleNavigableSetWrapper(final java.util.NavigableSet<Type> javaCollection) {
		super();
		this.javaCollection = javaCollection;
	}

	@Override
	public boolean contains(final Object o) {
		return this.javaCollection.contains(o);
	}
	
	@Override
	public boolean containsAll(final de.lmu.ifi.pms.java.util.collection.Collection<?> c) {
		return this.javaCollection.containsAll(c.javaCollection());
	}
	
	@Override
	public boolean isEmpty() {
		return this.javaCollection.isEmpty();
	}
	
	@Override
	public AbstractJavaNavigableSet<Type> javaCollection() {
		if(this.javaCollection instanceof AbstractJavaNavigableSet){
			return (AbstractJavaNavigableSet<Type>) this.javaCollection;
		}
		else{
			return new ToJavaNavigableSetWrapper<Type>(this);
		}
	}
	
	@Override
	public boolean add(final Type value) {
		return this.javaCollection.add(value);
	}
	
	@Override
	public void clear() {
		this.javaCollection.clear();
	}
	
	@Override
	public boolean remove(final Object o) {
		return this.javaCollection.remove(o);
	}
	
	@Override
	public Object[] toArray() {
		return this.javaCollection.toArray();
	}
	
	@Override
	public <T> T[] toArray(final T[] a) {
		return this.javaCollection.toArray(a);
	}

	@Override
	public ExtendibleIterator<Type> iterator() {
		return new JavaIteratorExtendibleWrapper<Type>(this.javaCollection);
	}

	@Override
	public final int size() {
		return this.javaCollection.size();
	}

	@Override
	public boolean addAll(final java.util.Collection<? extends Type> c) {
		return this.javaCollection.addAll(c);
	}

	@Override
	public boolean addAll(final de.lmu.ifi.pms.java.util.collection.Collection<? extends Type> c) {
		return this.addAll(c.javaCollection());
	}

	@Override
	public boolean removeAll(final java.util.Collection<?> c) {
		return this.javaCollection.removeAll(c);
	}
	
	@Override
	public boolean removeAll(final Collection<?> c) {
		return this.removeAll(c.javaCollection());
	}
	
	@Override
	public boolean retainAll(final java.util.Collection<?> c) {
		return this.javaCollection.retainAll(c);
	}

	@Override
	public boolean retainAll(final Collection<?> c) {
		return this.retainAll(c.javaCollection());
	}

	@Override
	public Comparator<? super Type> comparator() {
		return this.javaCollection.comparator();
	}

	@Override
	public Type first() {
		return this.javaCollection.first();
	}

	@Override
	public Type last() {
		return this.javaCollection.last();
	}
	
	@Override
	public boolean removeFirst() {
		Type value = this.javaCollection.pollFirst();
		return value != null;
	}
	
	@Override
	public boolean removeLast() {
		Type value = this.javaCollection.pollLast();
		return value != null;
	}

	@Override
	public NavigableSet<Type> headSet(final Type toElement) {
		return this.headSet(toElement, false);
	}

	@Override
	public NavigableSet<Type> tailSet(final Type fromElement) {
		return this.tailSet(fromElement, true);
	}

	@Override
	public NavigableSet<Type> subSet(final Type fromElement, final Type toElement) {
		return this.subSet(fromElement, true, toElement, false);
	}

	@Override
	public Type ceiling(final Type e) {
		return this.javaCollection.ceiling(e);
	}

	@Override
	public Type floor(final Type e) {
		return this.javaCollection.floor(e);
	}

	@Override
	public Type higher(final Type e) {
		return this.javaCollection.higher(e);
	}

	@Override
	public Type lower(final Type e) {
		return this.javaCollection.lower(e);
	}

	@Override
	public NavigableSet<Type> headSet(final Type toElement, final boolean inclusive) {
		return new FromJavaExtendibleNavigableSetWrapper<Type>(this.javaCollection.headSet(toElement, inclusive));
	}

	@Override
	public NavigableSet<Type> subSet(final Type fromElement, final boolean fromInclusive, final Type toElement, final boolean toInclusive) {
		return new FromJavaExtendibleNavigableSetWrapper<Type>(this.javaCollection.subSet(fromElement, fromInclusive, toElement, toInclusive));
	}

	@Override
	public NavigableSet<Type> tailSet(final Type fromElement, final boolean inclusive) {
		return new FromJavaExtendibleNavigableSetWrapper<Type>(this.javaCollection.tailSet(fromElement, inclusive));
	}

	@Override
	public NavigableSet<Type> descendingSet() {
		return new FromJavaExtendibleNavigableSetWrapper<Type>(this.javaCollection.descendingSet());
	}

	@Override
	public Iterator<Type> descendingIterator() {
		return this.javaCollection.descendingIterator();
	}
	
	

}
