/**
 * 
 */
package de.lmu.ifi.pms.java.util.map.javaWrapper;

import java.util.NoSuchElementException;

import de.lmu.ifi.pms.java.lang.iteration.AbstractExtendibleIterator;
import de.lmu.ifi.pms.java.lang.iteration.ExtendibleIterator;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.fixed.wrapper.FixedCollectionWrapper;
import de.lmu.ifi.pms.java.util.collection.fixed.wrapper.FixedSetWrapper;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.FromJavaCollectionWrapper;
import de.lmu.ifi.pms.java.util.collection.set.Set;
import de.lmu.ifi.pms.java.util.map.AbstractExtendibleMap;
import de.lmu.ifi.pms.java.util.map.Entry;




/**
 * @author Simon Matthias Brodt
 *
 */
public class FromJavaMapExtendibleWrapper<KeyType, ValueType>
		extends AbstractExtendibleMap<KeyType, ValueType, FromJavaEntryExtendibleWrapper<KeyType, ValueType>> {

	final java.util.Map<KeyType, ValueType> javaMap;

	public FromJavaMapExtendibleWrapper(final java.util.Map<KeyType, ValueType> javaMap) {
		super();
		this.javaMap = javaMap;
	}

	@Override
	public FromJavaEntryExtendibleWrapper<KeyType, ValueType> entry(final KeyType key) {
		for(java.util.Map.Entry<KeyType, ValueType> javaEntry: this.javaMap.entrySet()){
			if(javaEntry.getKey().equals(key)){
				return new FromJavaEntryExtendibleWrapper<KeyType, ValueType>(javaEntry, this.javaMap);
			}
		}
		throw new NoSuchElementException();//TODO better exception
	}

	@Override
	public ValueType get(final KeyType key) {
		return this.javaMap.get(key);
	}

	@Override
	public boolean has(final KeyType key) {
		return this.javaMap.containsKey(key);
	}

	@Override
	public ExtendibleIterator<FromJavaEntryExtendibleWrapper<KeyType, ValueType>> iterator() {
		final java.util.Iterator<java.util.Map.Entry<KeyType, ValueType>> javaIterator = this.javaMap.entrySet().iterator();
		return new AbstractExtendibleIterator<FromJavaEntryExtendibleWrapper<KeyType,ValueType>>() {
			
			FromJavaEntryExtendibleWrapper<KeyType, ValueType> current;

			@Override
			public boolean hasNext() {
				return javaIterator.hasNext();
			}

			@Override
			public FromJavaEntryExtendibleWrapper<KeyType, ValueType> next() {
				this.current = new FromJavaEntryExtendibleWrapper<KeyType, ValueType>(javaIterator.next(), FromJavaMapExtendibleWrapper.this.javaMap);
				return this.current;
			}

			@Override
			public void replace(final FromJavaEntryExtendibleWrapper<KeyType, ValueType> newValue) {
				if(this.current == null){
					throw new IllegalStateException();//TODO better exception
				}
				else{
					if(this.current.key().equals(newValue)){
						try{
							this.current.setValue(newValue.value());
						}
						catch (UnsupportedOperationException e) {
							FromJavaMapExtendibleWrapper.this.set(newValue.key(), newValue.value());
						}
					}
					else{
						throw new IllegalArgumentException();//TODO better exception
					}
				}
			}

			@Override
			public void remove() {
				this.current.remove();
			}

			@Override
			public void insert(final FromJavaEntryExtendibleWrapper<KeyType, ValueType> entry) {
				FromJavaMapExtendibleWrapper.this.add(entry);
			}

		};
	}

	@Override
	public Set<KeyType> keys() {
		return new FixedSetWrapper<KeyType>(this.javaMap.keySet());
	}


	@Override
	public int size() {
		return this.javaMap.size();
	}

	@Override
	public Collection<ValueType> values() {
		return new FixedCollectionWrapper<ValueType>(new FromJavaCollectionWrapper<ValueType>(this.javaMap.values()));
	}

	@Override
	public void set(final KeyType key, final ValueType value) {
		this.javaMap.put(key, value);
	}
	
	@Override
	public boolean remove(final Entry<KeyType, ValueType> entry) {
		if(entry.value().equals(this.get(entry.key()))){
			return this.removeKey(entry.key());
		}
		else{
			return false;
		}
	}

	@Override
	public void clear() {
		this.javaMap.clear();
	}
	
	@Override
	public boolean add(final KeyType key, final ValueType value) {
		this.javaMap.put(key, value);
		return true;
	}

	@Override
	public boolean removeKey(final KeyType key) {
		return this.javaMap.remove(key) != null;
	}
	
	//TODO improve addAll, removeAll, retainAll add(Map)
	
}
