package de.lmu.ifi.pms.java.util.observable;

import java.util.ArrayList;

public class ObserverManager<ArgumentType
							, ObservableType extends Observable<ArgumentType, ObservableType>>
							{
	
	private final ObservableType observable;
	
	private boolean changed = false;
	private final ArrayList<Observer<ArgumentType, ObservableType>> observers = new ArrayList<Observer<ArgumentType, ObservableType>>();	
	
	protected ObserverManager(final ObservableType observable) {
		super();
		this.observable = observable;
	}
	
	@SuppressWarnings("unchecked")
	protected ObserverManager() {
		super();
		this.observable = (ObservableType) this;
	}

	public final void addObserver(final Observer<ArgumentType, ObservableType> observer) {
		synchronized (this.observers) {
			if(!this.observers.contains(observer)){
				this.observers.add(observer);
			}
		}
	}

	public final int countObservers() {
		synchronized (this.observers) {
			return this.observers.size();
		}
	}

	public final void deleteObserver(final Observer<ArgumentType, ObservableType> observer) {
		synchronized (this.observers) {
			this.observers.remove(observer);			
		}
	}

	public final void deleteObservers() {
		synchronized (this.observers) {
			this.observers.clear();
		}
	}
	
	protected final void setChanged(){
		synchronized (this.observable) {
			this.changed = true;
		}
	}
	
	protected final void clearChanged(){
		synchronized (this.observable) {
			this.changed = false;	
		}
	}

	public final boolean hasChanged() {
		synchronized (this.observable) {
			return this.changed;
		}
	}

	public final void notifyObservers() {
		synchronized (this.observers) {
			if(this.hasChanged()){
				for(Observer<ArgumentType, ObservableType> observer: this.observers){
					observer.update(this.observable);
				}
			}
		}		
	}

	public final void notifyObservers(final ArgumentType argument) {
		synchronized (this.observers) {
			boolean hasChanged;
			synchronized (this) {
				hasChanged = this.hasChanged();
			}
			if(hasChanged){
				for(Observer<ArgumentType, ObservableType> observer: this.observers){
					observer.update(this.observable, argument);
				}
			}
		}
	}

}
