package de.lmu.ifi.pms.cep.unitTests.dura;

import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;

import nl.cwi.monetdb.event_mill.MonetDriver;

import org.junit.BeforeClass;
import org.junit.Test;

import de.lmu.ifi.pms.cep.compilation.Compiler;
import de.lmu.ifi.pms.cep.compilation.StackCompiler;
import de.lmu.ifi.pms.cep.dura.compiler.durac.DuraCCompiler;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCModuleNotFoundException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCompilationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraParseException;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_Program_Spec;
import de.lmu.ifi.pms.cep.event_mill.sql.compiler.DTSA_SQL_Compiler;
import de.lmu.ifi.pms.cep.event_mill.sql.compiler.DTSA_SQL_WriterImpl;
import de.lmu.ifi.pms.cep.tsa.TSA_Program;
import de.lmu.ifi.pms.cep.tsa.compiler.TSA_DTSA_Compiler;
import de.lmu.ifi.pms.cep.tsa.compiler.TSA_DefaultDifferencer;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_ComprehensiveException;


public class DuraCTests {
	
	String testFileDirectory = "src/de/lmu/ifi/pms/cep/unitTests/dura/files/";
	String exampleFileDirectory = "examples/";
	
	static Compiler<TSA_Program, SQL_Program_Spec> tsaSQLCompiler;
	
	@BeforeClass
	public static void init() throws ClassNotFoundException {
		tsaSQLCompiler = new StackCompiler<TSA_Program, SQL_Program_Spec>(TSA_Program.class, SQL_Program_Spec.class, new TSA_DTSA_Compiler(new TSA_DefaultDifferencer()), new DTSA_SQL_Compiler(new DTSA_SQL_WriterImpl(new MonetDriver().translation())));
	}
	
	@Test
	public void playground() throws IOException {
		File file = new File(this.testFileDirectory + "Playground.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void compileExamples() {
		File dir = new File(this.exampleFileDirectory);
		
		FileFilter duraFilter = new FileFilter() {
			@Override
		    public boolean accept(final File file) {
		        return file.isFile() && file.getName().endsWith(".dura");
		    }
		};

		ArrayList<Exception> exceptions = new ArrayList<Exception>();
		for (File file : dir.listFiles(duraFilter)) {
			try{
				this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
			}
			catch (Exception e) {
				exceptions.add(new RuntimeException("\nException in file \n" + file.getName() , e));
			}
		}
		if(exceptions.size() > 0){
			throw new RuntimeException(TSA_ComprehensiveException.message(exceptions));
		}
	}
	
	@Test
	public void foreignKeys() throws IOException {
		File file = new File(this.testFileDirectory + "ForeignKeys.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}

	
	@Test
	public void reactiveRule() throws IOException {
		File file = new File(this.testFileDirectory + "ReactiveRules.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void implicitActionStreams() throws IOException {
		File file = new File(this.testFileDirectory + "ImplicitActionStreams.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void implicitStatefulObjectStreams() throws IOException {
		File file = new File(this.testFileDirectory + "ImplicitStatefulObjectStreams.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void streamConstraints() throws IOException {
		File file = new File(this.testFileDirectory + "StreamConstraints.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}

	@Test
	public void timeIntervalFunctionsAndRelations() throws IOException {
		File file = new File(this.testFileDirectory + "TimeIntervalFunctionsAndRelations.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void grouping() throws IOException {
		File file = new File(this.testFileDirectory + "Grouping.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void literals() throws IOException {
		File file = new File(this.testFileDirectory + "Literals.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void simpleModule() throws IOException {
		File file = new File(this.testFileDirectory + "SimpleModule.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void module() throws IOException {
		Collection<String> classfiles = new ArrayList<String>();
		
		String moduleDirectory = this.testFileDirectory + "/modules/";
		
		classfiles.add(moduleDirectory + "a.dura");
		classfiles.add(moduleDirectory + "b.dura");
		classfiles.add(moduleDirectory + "c.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("a", classfiles));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void moduleVisibility1() throws IOException {
		Collection<String> classfiles = new ArrayList<String>();
		
		String moduleDirectory = this.testFileDirectory + "/module_visibility/";
		
		classfiles.add(moduleDirectory + "a1.dura");
		classfiles.add(moduleDirectory + "b.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("a", classfiles));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void moduleVisibility2() throws IOException {
		Collection<String> classfiles = new ArrayList<String>();
		
		String moduleDirectory = this.testFileDirectory + "/module_visibility/";
		
		classfiles.add(moduleDirectory + "a2.dura");
		classfiles.add(moduleDirectory + "b.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("a", classfiles));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void moduleVisibility3() throws IOException {
		Collection<String> classfiles = new ArrayList<String>();
		
		String moduleDirectory = this.testFileDirectory + "/module_visibility/";
		
		classfiles.add(moduleDirectory + "a3.dura");
		classfiles.add(moduleDirectory + "b.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("a", classfiles));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void moduleVisibility4() throws IOException {
		Collection<String> classfiles = new ArrayList<String>();
		
		String moduleDirectory = this.testFileDirectory + "/module_visibility/";
		
		classfiles.add(moduleDirectory + "a4.dura");
		classfiles.add(moduleDirectory + "b.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("a", classfiles));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void incompleteClassfiles() throws IOException {
		Collection<String> classfiles = new ArrayList<String>();
		
		String moduleDirectory = this.testFileDirectory + "/modules/";
		
		classfiles.add(moduleDirectory + "a.dura");
		classfiles.add(moduleDirectory + "b.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("a", classfiles));
	}
	
	@Test(expected=DuraCModuleNotFoundException.class)
	public void emptyClassfiles() throws IOException {
		Collection<String> classfiles = new ArrayList<String>();
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("a", classfiles));
	}
	
	
	@Test
	public void negation() throws IOException {
		File file = new File(this.testFileDirectory + "Negation.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void statefulObject() throws IOException {
		File file = new File(this.testFileDirectory + "StatefulObject.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void atomicQuerySupplement() throws IOException {
		File file = new File(this.testFileDirectory + "AtomicQuerySupplement.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void pathConditions() throws IOException {
		File file = new File(this.testFileDirectory + "PathConditions.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void dynamicStatefulObjects() throws IOException {
		File file = new File(this.testFileDirectory + "DynamicStatefulObjects.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void dynamicStatefulObjectsMinimal() throws IOException {
		File file = new File(this.testFileDirectory + "DynamicStatefulObjectsMinimal.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void syntaxRestrictionsStreamName() throws IOException {
		File file = new File(this.testFileDirectory + "SyntaxRestrictionsStreamName.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void syntaxRestrictionsAttributeName() throws IOException {
		File file = new File(this.testFileDirectory + "SyntaxRestrictionsAttributeName.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraParseException.class)
	public void parseException() throws IOException {
		File file = new File(this.testFileDirectory + "ParseException.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void staticStatefulObjects() throws IOException {
		File file = new File(this.testFileDirectory + "StaticStatefulObjects.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void filteredAttributes2() throws IOException {
		File file = new File(this.testFileDirectory + "FilteredAttributes2.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void filteredAttributes3() throws IOException {
		File file = new File(this.testFileDirectory + "FilteredAttributes3.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void filteredAttributes4() throws IOException {
		File file = new File(this.testFileDirectory + "FilteredAttributes4.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void filteredAttributes() throws IOException {
		File file = new File(this.testFileDirectory + "FilteredAttributes.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void rangeRestrictionNegation() throws IOException {
		File file = new File(this.testFileDirectory + "RangeRestrictionNegation.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void negatedStaticStatefulObject() throws IOException {
		File file = new File(this.testFileDirectory + "NegatedStaticStatefulObject.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void statefulObjectGrouping() throws IOException {
		File file = new File(this.testFileDirectory + "StatefulObjectGrouping.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void incompatibleTypes() throws IOException {
		File file = new File(this.testFileDirectory + "IncompatibleTypes.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void incompatibleTypes2() throws IOException {
		File file = new File(this.testFileDirectory + "IncompatibleTypes2.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test(expected=DuraCompilationException.class)
	public void incompatibleTypes3() throws IOException {
		File file = new File(this.testFileDirectory + "IncompatibleTypes3.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void termTypes() throws IOException {
		File file = new File(this.testFileDirectory + "TermTypes.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void timeIntervalBinding() throws IOException {
		File file = new File(this.testFileDirectory + "TimeIntervalBinding.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	@Test
	public void exists() throws IOException {
		File file = new File(this.testFileDirectory + "Exists.dura");
		
		this.tsaSQLCompiler.compile(DuraCCompiler.compile("", "test", file));
	}
	
	public static void main(final String[] args) {
		new DuraCTests().compileExamples();
	}
}
