package de.lmu.ifi.pms.cep.unitTests.time.constraints.tests;

import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula.*;
import static de.lmu.ifi.pms.cep.tsa.time.constraints.enums.BinaryComparisonOperator.*;
import static org.junit.Assert.*;

import org.junit.Test;

import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.TimestampAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.analysis.GenericFormulaAnalyser;
import de.lmu.ifi.pms.cep.tsa.time.analysis.ArithmeticAnalyser;
import de.lmu.ifi.pms.cep.tsa.time.analysis.TemporalSimpleFocusedFormulaAnalyser;
import de.lmu.ifi.pms.cep.tsa.time.constraints.SimpleTemporalFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.TemporalRelationFormula;


public class FormulaAnalyserTest {
	
	static final String a = "a";
	static final String b = "b";
	static final String c = "c";
	
	static final TemporalRelationFormula f1 = new SimpleTemporalFormula(a, LESS_EQUAL, b);
	static final TemporalRelationFormula f2 = new SimpleTemporalFormula(a, GREATER, b);
	static final TemporalRelationFormula f3 = new SimpleTemporalFormula(a, LESS_EQUAL, b, 2);
	static final TemporalRelationFormula f4 = new SimpleTemporalFormula(a, LESS_EQUAL, b, -3);
	static final TemporalRelationFormula f5 = new SimpleTemporalFormula(b, LESS_EQUAL, a, 2);
	static final TemporalRelationFormula f6 = new SimpleTemporalFormula(b, LESS_EQUAL, a, -1);
	static final TemporalRelationFormula f7 = new SimpleTemporalFormula(b, LESS_EQUAL, a, 4);
	
	static final TemporalRelationFormula f8 = new SimpleTemporalFormula(b, LESS_EQUAL, c, 2);
	static final TemporalRelationFormula f9 = new SimpleTemporalFormula(a, LESS_EQUAL, c, 2);
	static final TemporalRelationFormula f10 = new SimpleTemporalFormula(b, LESS_EQUAL, c, -2);
	static final TemporalRelationFormula f11 = new SimpleTemporalFormula(a, LESS_EQUAL, c);
	
	static final GenericFormulaAnalyser genericFormulaAnalyser = new GenericFormulaAnalyser(new TemporalSimpleFocusedFormulaAnalyser());
	
	@Test
	public void test1(){
		assertTrue(ArithmeticAnalyser.encodedPlus(ArithmeticAnalyser.MINUS_INFTY + 5, -20) == ArithmeticAnalyser.MINUS_INFTY);
		assertTrue(genericFormulaAnalyser.isConsitent(f1));
		assertFalse(genericFormulaAnalyser.isConsitent(and(f1, f2)));
		assertTrue(genericFormulaAnalyser.isConsitent(and(f3, f4)));
		assertTrue(genericFormulaAnalyser.isConsitent(and(f3, f5)));
		assertFalse(genericFormulaAnalyser.isConsitent(and(f4, f5)));
		assertTrue(genericFormulaAnalyser.isConsitent(or(and(f4, f5), and(f3, f5))));
		assertFalse(genericFormulaAnalyser.isConsitent(or(and(f4, f5), and(f1, f2))));
		assertFalse(genericFormulaAnalyser.isConsitent(and(f4, or(f5, f6))));
		assertTrue(genericFormulaAnalyser.isConsitent(and(f4, or(f5, f7))));
		assertTrue(genericFormulaAnalyser.isConsitent(not(and(f4, or(f5, f7)))));
	}
	
	@Test
	public void test2(){
		assertTrue(genericFormulaAnalyser.implies(f1, f3));
		assertFalse(genericFormulaAnalyser.implies(f3, f1));
		assertTrue(genericFormulaAnalyser.implies(and(f1, f8), f9));
		assertFalse(genericFormulaAnalyser.implies(and(f3, f8), f9));
		assertTrue(genericFormulaAnalyser.implies(and(f4, f8), f9));
		assertTrue(genericFormulaAnalyser.implies(and(or(f1,f4), f8), f9));
		assertTrue(genericFormulaAnalyser.implies(not(f9), not(and(or(f1,f4), f8))));
		assertFalse(genericFormulaAnalyser.implies(and(or(f3,f4), f8), f9));
		assertFalse(genericFormulaAnalyser.implies(not(f9), not(and(or(f3,f4), f8))));
		assertTrue(genericFormulaAnalyser.implies(and(f3, f10), f11));
	}
	
	static final TimestampAttribute A = TimestampAttribute.create(a);
	static final TimestampAttribute B = TimestampAttribute.create(b);
	static final TimestampAttribute C = TimestampAttribute.create(c);
	
	static final ConditionFormula G1 = new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS, A, B);
	static final ConditionFormula G2 = new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS_EQUAL, A, B);
	static final ConditionFormula G3 = new BinaryComparissonFormula(BinaryComparissonOperatorID.EQUAL, A, B);
	static final ConditionFormula G4 = new BinaryComparissonFormula(BinaryComparissonOperatorID.GREATER_EQUAL, A, B);
	static final ConditionFormula G5 = new BinaryComparissonFormula(BinaryComparissonOperatorID.GREATER, A, B);
	static final ConditionFormula G6 = new BinaryComparissonFormula(BinaryComparissonOperatorID.NOT_EQUAL, A, B);
	
	@Test
	public void test3(){
		assertTrue(genericFormulaAnalyser.isConsitent(G1));
		assertFalse(genericFormulaAnalyser.isConsitent(and(G1, G3)));
		assertTrue(genericFormulaAnalyser.isConsitent(and(G1, G6)));
		assertTrue(genericFormulaAnalyser.isConsitent(and(G2, G4)));
		assertFalse(genericFormulaAnalyser.isConsitent(and(G2, G5)));
	}

}
